//
// Copyright (c) 2007-2008, by YaPB Development Team. All rights reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
// Version: $Id:$
//

#include <core.h>

int Bot::GetMessageQueue (void)
{
   // this function get the current message from the bots message queue

   int message = m_messageQueue[m_actMessageIndex++];
   m_actMessageIndex &= 0x1f; // wraparound

   return message;
}

void Bot::PushMessageQueue (int message)
{
   // this function put a message into the bot message queue

   if (message == GameMessage_SayMsg)
   {
      // notify other bots of the spoken text otherwise, bots won't respond to other bots (network messages aren't sent from bots)
      int entityIndex = GetIndex ();

      for (int i = 0; i < GetMaxClients (); i++)
      {
         Bot *otherBot = g_botManager->GetBot (i);

         if (otherBot != NULL && otherBot->pev != pev)
         {
            if (IsAlive (GetEntity ()) == IsAlive (otherBot->GetEntity ()))
            {
               otherBot->m_sayTextBuffer.entityIndex = entityIndex;
               strcpy (otherBot->m_sayTextBuffer.sayText, m_tempStrings);
            }
            otherBot->m_sayTextBuffer.timeNextChat = GetWorldTime () + otherBot->m_sayTextBuffer.chatDelay;
         }
      }
   }
   m_messageQueue[m_pushMessageIndex++] = message;
   m_pushMessageIndex &= 0x1f; // wraparound
}

int Bot::InFieldOfView (Vector destination)
{
   float entityAngle = AngleMod (destination.ToYaw ()); // find yaw angle from source to destination...
   float viewAngle = AngleMod (pev->v_angle.y); // get bot's current view angle...

   // return the absolute value of angle to destination entity
   // zero degrees means straight ahead, 45 degrees to the left or
   // 45 degrees to the right is the limit of the normal view angle
   int absoluteAngle = abs (static_cast <int> (viewAngle) - static_cast <int> (entityAngle));

   if (absoluteAngle > 180)
      absoluteAngle = 360 - absoluteAngle;

   return absoluteAngle;
}

bool Bot::IsInViewCone (Vector origin)
{
   // this function returns true if the spatial vector location origin is located inside
   // the field of view cone of the bot entity, false otherwise. It is assumed that entities
   // have a human-like field of view, that is, about 90 degrees.

   return ::IsInViewCone (origin, GetEntity ());
}

bool Bot::CheckVisibility (entvars_t *targetEntity, Vector *origin, byte *bodyPart)
{
   // this function checks visibility of a bot target.

   Vector botHead = EyePosition ();
   TraceResult tr;

   *bodyPart = 0;

   // check for the body
   TraceLine (botHead, targetEntity->origin, true, true, GetEntity (), &tr);

   if (tr.flFraction >= 1.0)
   {
      *bodyPart |= Visibility_Body;
      *origin = tr.vecEndPos + Vector (0, 0, 6);
   }

   // check for the head
   TraceLine (botHead, targetEntity->origin + targetEntity->view_ofs, true, true, GetEntity (), &tr);

   if (tr.flFraction >= 1.0)
   {
      *bodyPart |= Visibility_Head;
      *origin = tr.vecEndPos;
   }

   if (*bodyPart != 0)
      return true;

#if 0
   // dimension table
   const int8 dimensionTab[8][3] =
   {
      {1,  1,  1}, { 1,  1, -1},
      {1, -1,  1}, {-1,  1,  1},
      {1, -1, -1}, {-1, -1,  1},
      {-1, 1, -1}, {-1, -1, -1}
   };

   // check the box borders
   for (int i = 7; i >= 0; i--)
   {
      Vector targetOrigin = petargetOrigin->origin + Vector (dimensionTab[i][0] * 24.0, dimensionTab[i][1] * 24.0, dimensionTab[i][2] * 24.0);

      // check direct line to random part of the player body
      TraceLine (botHead, targetOrigin, true, true, GetEntity (), &tr);

      // check if we hit something
      if (tr.flFraction >= 1.0)
      {
         *origin = tr.vecEndPos;
         *bodyPart |= Visibility_Other;

         return true;
      }
   }
#else
   // worst case, choose random position in enemy body
   for (int i = 0; i < 5; i++)
   {
      Vector targetOrigin = targetEntity->origin; // get the player origin

      // find the vector beetwen mins and maxs of the player body
      targetOrigin.x += g_randGen.Float (targetEntity->mins.x, targetEntity->maxs.x);
      targetOrigin.y += g_randGen.Float (targetEntity->mins.y, targetEntity->maxs.y);
      targetOrigin.z += g_randGen.Float (targetEntity->mins.z, targetEntity->maxs.z);

      // check direct line to random part of the player body
      TraceLine (botHead, targetOrigin, true, true, GetEntity (), &tr);

      // check if we hit something
      if (tr.flFraction >= 1.0)
      {
         *origin = tr.vecEndPos;
         *bodyPart |= Visibility_Other;

         return true;
      }
   }
#endif
   return false;
}

bool Bot::IsEnemyViewable (edict_t *player)
{
   if (FNullEnt (player))
      return false;

   bool forceTrueIfVisible = false;

   if (IsValidPlayer (pev->dmg_inflictor) && GetTeam (pev->dmg_inflictor) != GetTeam (GetEntity ()) && ::IsInViewCone (EyePosition (), pev->dmg_inflictor))
      forceTrueIfVisible = true;

   if (CheckVisibility (VARS (player), &m_enemyOrigin, &m_visibility) && (IsInViewCone (player->v.origin + Vector (0, 0, 14)) || forceTrueIfVisible))
   {
      m_seeEnemyTime = GetWorldTime ();
      m_lastEnemy = player;
      m_lastEnemyOrigin = player->v.origin;

      return true;
   }
   return false;
}

bool Bot::ItemIsVisible (Vector destination, char *itemName)
{
   TraceResult tr;

   // trace a line from bot's eyes to destination..
   TraceLine (EyePosition (), destination, true, GetEntity (), &tr);

   // check if line of sight to object is not blocked (i.e. visible)
   if (tr.flFraction != 1.0)
   {
      // check for standard items
      if (strcmp (STRING (tr.pHit->v.classname), itemName) == 0)
         return true;

      if (tr.flFraction > 0.98 && strncmp (STRING (tr.pHit->v.classname), "csdmw_", 6) == 0)
         return true;

      return false;
   }
   return true;
}

bool Bot::EntityIsVisible (Vector dest, bool fromBody)
{
   TraceResult tr;

   // trace a line from bot's eyes to destination...
   TraceLine (fromBody ? pev->origin - Vector (0, 0, 1) : EyePosition (), dest, true, true, GetEntity (), &tr);

   // check if line of sight to object is not blocked (i.e. visible)
   return tr.flFraction >= 1.0;
}

void Bot::AvoidGrenades (void)
{
   // checks if bot 'sees' a grenade, and avoid it

   edict_t *ent = m_avoidGrenade;

  // check if old pointers to grenade is invalid
   if (FNullEnt (ent))
   {
      m_avoidGrenade = NULL;
      m_needAvoidGrenade = 0;
   }
   else if ((ent->v.flags & FL_ONGROUND) || (ent->v.effects & EF_NODRAW))
   {
      m_avoidGrenade = NULL;
      m_needAvoidGrenade = 0;
   }
   ent = NULL;

   // find all grenades on the map
   while (!FNullEnt (ent = FIND_ENTITY_BY_CLASSNAME (ent, "grenade")))
   {
      // if grenade is invisible don't care for it
      if (ent->v.effects & EF_NODRAW)
         continue;

      // check if visible to the bot
      if (!EntityIsVisible (ent->v.origin) && InFieldOfView (ent->v.origin - EyePosition ()) > pev->fov / 3)
         continue;

#if 0
      // TODO: should be done once for grenade, instead of checking several times
      if (m_skill >= 70 && strcmp (STRING (ent->v.model) + 9, "flashbang.mdl") == 0)
      {
         Vector position = (GetEntityOrigin (ent) - EyePosition ()).ToAngles ();

         // don't look at flashbang
         if (((ent->v.owner == GetEntity () || g_randGen.Long (0, 100) < 85) && !(m_states & State_SeeingEnemy)))
         {
            pev->v_angle.y = AngleNormalize (position.y + 180.0);
            m_canChooseAimDirection = false;
         }
      }
      else 
#endif
      if (strcmp (STRING (ent->v.model) + 9, "hegrenade.mdl") == 0)
      {
         if (!FNullEnt (m_avoidGrenade))
            return;

         if (GetTeam (ent->v.owner) == GetTeam (GetEntity ()) && ent->v.owner != GetEntity ())
            return;

         if ((ent->v.flags & FL_ONGROUND) == 0)
         {
            float distance = (ent->v.origin - pev->origin).GetLength ();
            float distanceMoved = ((ent->v.origin + ent->v.velocity * m_frameInterval) - pev->origin).GetLength ();

            if (distanceMoved < distance && distance < 500)
            {
               MakeVectors (pev->v_angle);

               Vector dirToPoint = (pev->origin - ent->v.origin).Normalize2D ();
               Vector rightSide = g_pGlobals->v_right.Normalize2D ();

               if ((dirToPoint | rightSide) > 0)
                  m_needAvoidGrenade = -1;
               else
                  m_needAvoidGrenade = 1;

               m_avoidGrenade = ent;
            }
         }
      }
   }
}

bool Bot::IsBehindSmokeClouds (edict_t *ent)
{
   edict_t *pentGrenade = NULL;
   Vector betweenUs = (ent->v.origin - pev->origin).Normalize ();

   while (!FNullEnt (pentGrenade = FIND_ENTITY_BY_CLASSNAME (pentGrenade, "grenade")))
   {
      // if grenade is invisible don't care for it
      if ((pentGrenade->v.effects & EF_NODRAW) || !(pentGrenade->v.flags & (FL_ONGROUND | FL_PARTIALGROUND)) || strcmp (STRING (pentGrenade->v.model) + 9, "smokegrenade.mdl"))
         continue;

      // check if visible to the bot
      if (!EntityIsVisible (ent->v.origin) && InFieldOfView (ent->v.origin - EyePosition ()) > pev->fov / 3)
         continue;

      Vector betweenNade = (GetEntityOrigin (pentGrenade) - pev->origin).Normalize ();
      Vector betweenResult = ((Vector (betweenNade.y, betweenNade.x, 0) * 150.0 + GetEntityOrigin (pentGrenade)) - pev->origin).Normalize ();

      if ((betweenNade | betweenUs) > (betweenNade | betweenResult))
         return true;
   }
   return false;
}

int Bot::GetBestWeaponCarried (void)
{
   // this function returns the best weapon of this bot (based on personality prefs)

   int *ptr = g_weaponPrefs[m_personality];
   int weaponIndex = 0;
   int weapons = pev->weapons;

   WeaponSelect_t *weaponTab = &g_weaponSelect[0];

   // take the shield in account
   if (HasShield ())
      weapons |= (1 << Weapon_ShieldGun);

   for (int i = 0; i < Const_NumWeapons; i++)
   {
      if (weapons & (1 << weaponTab[*ptr].id))
         weaponIndex = i;

      ptr++;
   }
   return weaponIndex;
}

int Bot::GetBestSecondaryWeaponCarried (void)
{
   // this function returns the best secondary weapon of this bot (based on personality prefs)

   int *ptr = g_weaponPrefs[m_personality];
   int weaponIndex = 0;
   int weapons = pev->weapons;

   // take the shield in account
   if (HasShield ())
      weapons |= (1 << Weapon_ShieldGun);

   WeaponSelect_t *weaponTab = &g_weaponSelect[0];

   for (int i = 0; i < Const_NumWeapons; i++)
   {
      int id = weaponTab[*ptr].id;

      if ((weapons & (1 << weaponTab[*ptr].id)) && (id == Weapon_Usp || id == Weapon_Glock18 || id == Weapon_Deagle || id == Weapon_SigP228 || id == Weapon_Elite || id == Weapon_FiveSeven))
      {
         weaponIndex = i;
         break;
      }
      ptr++;
   }
   return weaponIndex;
}

bool Bot::RateGroundWeapon (edict_t *ent)
{
   // this function compares weapons on the ground to the one the bot is using

   int hasWeapon = 0;
   int groundIndex = 0;
   int *ptr = g_weaponPrefs[m_personality];

   WeaponSelect_t *weaponTab = &g_weaponSelect[0];

   for (int i = 0; i < Const_NumWeapons; i++)
   {
      if (strcmp (weaponTab[*ptr].modelName, STRING (ent->v.model) + 9) == 0)
      {
         groundIndex = i;
         break;
      }
      ptr++;
   }

   if (groundIndex < 7)
      hasWeapon = GetBestSecondaryWeaponCarried ();
   else
      hasWeapon = GetBestWeaponCarried ();

   return groundIndex > hasWeapon;
}

edict_t *Bot::FindBreakable (void)
{
   // this function checks if bot is blocked by a shoot able breakable in his moving direction

   TraceResult tr;
   TraceLine (pev->origin, pev->origin + (m_destOrigin - pev->origin).Normalize () * 64, false, false, GetEntity (), &tr);

   if (tr.flFraction != 1.0)
   {
      edict_t *ent = tr.pHit;

      // check if this isn't a triggered (bomb) breakable and if it takes damage. if true, shoot the crap!
      if (IsShootableBreakable (ent))
      {
         m_breakable = tr.vecEndPos;
         return ent;
      }
   }
   TraceLine (EyePosition (), EyePosition () + (m_destOrigin - EyePosition ()).Normalize () * 64, false, false, GetEntity (), &tr);

   if (tr.flFraction != 1.0)
   {
      edict_t *ent = tr.pHit;

      if (IsShootableBreakable (ent))
      {
         m_breakable = tr.vecEndPos;
         return ent;
      }
   }
   m_breakableEntity = NULL;
   m_breakable = Vector::GetNull ();

   return NULL;
}

void Bot::FindItem (void)
{
   // this function finds Items to collect or use in the near of a bot

   edict_t *ent = NULL, *pickupItem = NULL;
   Bot *bot = NULL;

   bool allowPickup= false;
   float distance, minDistance = 341.0;

   // don't try to pickup anything while on ladder or trying to escape from bomb...
   if (IsOnLadder () || GetCurrentTask ()->taskID == Task_EscapeFromBomb || GetCurrentTask ()->taskID == Task_PlantBomb || GetCurrentTask ()->taskID == Task_DefuseBomb)
   {
      m_pickupItem = NULL;
      m_pickupType = PickupType_None;

      return;
   }

   if (!FNullEnt (m_pickupItem))
   {
      bool itemExists = false;
      pickupItem = m_pickupItem;

      while (!FNullEnt (ent = FIND_ENTITY_IN_SPHERE (ent, pev->origin, 340.0)))
      {
         if ((ent->v.effects & EF_NODRAW) || IsValidPlayer (ent->v.owner))
            continue; // someone owns this weapon or it hasn't respawned yet

         if (ent == pickupItem)
         {
            if (ItemIsVisible (GetEntityOrigin (ent), const_cast <char *> (STRING (ent->v.classname))))
               itemExists = true;
            break;
         }
      }
      if (itemExists)
         return;
      else
      {
         m_pickupItem = NULL;
         m_pickupType = PickupType_None;
      }
   }

   ent = NULL;
   pickupItem = NULL;

   PickupType_t pickupType = PickupType_None;

   Vector pickupOrigin = Vector::GetNull ();
   Vector entityOrigin = Vector::GetNull ();

   m_pickupItem = NULL;
   m_pickupType = PickupType_None;

   while (!FNullEnt (ent = FIND_ENTITY_IN_SPHERE (ent, pev->origin, 340.0)))
   {
      allowPickup = false;  // assume can't use it until known otherwise

      if ((ent->v.effects & EF_NODRAW) || ent == m_itemIgnore)
         continue; // someone owns this weapon or it hasn't respawned yet

      entityOrigin = GetEntityOrigin (ent);

      // check if line of sight to object is not blocked (i.e. visible)
      if (ItemIsVisible (entityOrigin, const_cast <char *> (STRING (ent->v.classname))))
      {
         if (strncmp ("hostage_entity", STRING (ent->v.classname), 14) == 0)
         {
            allowPickup = true;
            pickupType = PickupType_Hostage;
         }
         else if (strncmp ("weaponbox", STRING (ent->v.classname), 9) == 0 && strcmp (STRING (ent->v.model) + 9, "backpack.mdl") == 0)
         {
            allowPickup = true;
            pickupType = PickupType_DroppedC4;
         }
         else if (strncmp ("weaponbox", STRING (ent->v.classname), 9) == 0 && strcmp (STRING (ent->v.model) + 9, "backpack.mdl") == 0 && !m_isUsingGrenade)
         {
            allowPickup = true;
            pickupType = PickupType_DroppedC4;
         }
         else if ((strncmp ("weaponbox", STRING (ent->v.classname), 9) == 0 || strncmp ("armoury_entity", STRING (ent->v.classname), 14) == 0 || strncmp ("csdm", STRING (ent->v.classname), 4) == 0) && !m_isUsingGrenade)
         {
            allowPickup = true;
            pickupType = PickupType_Weapon;
         }
         else if (strncmp ("weapon_shield", STRING (ent->v.classname), 13) == 0 && !m_isUsingGrenade)
         {
            allowPickup = true;
            pickupType = PickupType_ShieldGun;
         }
         else if (strncmp ("item_thighpack", STRING (ent->v.classname), 14) == 0 && GetTeam (GetEntity ()) == Team_Counter && !m_hasDefuser)
         {
            allowPickup = true;
            pickupType = PickupType_DefusalKit;
         }
         else if (strncmp ("grenade", STRING (ent->v.classname), 7) == 0 && strcmp (STRING (ent->v.model) + 9, "c4.mdl") == 0)
         {
            allowPickup = true;
            pickupType = PickupType_PlantedC4;
         }
      }

      if (allowPickup) // if the bot found something it can pickup...
      {
         distance = (entityOrigin - pev->origin).GetLength ();

         // see if it's the closest item so far...
         if (distance < minDistance)
         {
            if (pickupType == PickupType_Weapon) // found weapon on ground?
            {
               int weaponCarried = GetBestWeaponCarried ();
               int secondaryWeaponCarried = GetBestSecondaryWeaponCarried ();

               if (secondaryWeaponCarried < 7 && (m_ammo[g_weaponSelect[secondaryWeaponCarried].id] > 0.3 * g_weaponDefs[g_weaponSelect[secondaryWeaponCarried].id].ammo1Max) && strcmp (STRING (ent->v.model) + 9, "w_357ammobox.mdl") == 0)
                  allowPickup = false;
               else if (!m_isVIP && weaponCarried >= 7 && (m_ammo[g_weaponSelect[weaponCarried].id] > 0.3 * g_weaponDefs[g_weaponSelect[weaponCarried].id].ammo1Max) && strncmp (STRING (ent->v.model) + 9, "w_", 2) == 0)
               {
                  if (strcmp (STRING (ent->v.model) + 9, "w_9mmarclip.mdl") == 0 && !(weaponCarried == Weapon_Famas || weaponCarried == Weapon_Ak47 || weaponCarried == Weapon_M4A1 || weaponCarried == Weapon_Galil || weaponCarried == Weapon_Aug || weaponCarried == Weapon_Sig552))
                     allowPickup = false;
                  else if (strcmp (STRING (ent->v.model) + 9, "w_shotbox.mdl") == 0 && !(weaponCarried == Weapon_M3 || weaponCarried == Weapon_Xm1014))
                     allowPickup = false;
                  else if (strcmp (STRING (ent->v.model) + 9, "w_9mmclip.mdl") == 0 && !(weaponCarried == Weapon_Mp5Navy || weaponCarried == Weapon_Tmp || weaponCarried == Weapon_P90 || weaponCarried == Weapon_Mac10 || weaponCarried == Weapon_Ump45))
                     allowPickup = false;
                  else if (strcmp (STRING (ent->v.model) + 9, "w_crossbow_clip.mdl") == 0 && !(weaponCarried == Weapon_Magnum || weaponCarried == Weapon_G3SG1 || weaponCarried == Weapon_Scout || weaponCarried == Weapon_Sig550))
                     allowPickup = false;
                  else if (strcmp (STRING (ent->v.model) + 9, "w_chainammo.mdl") == 0 && weaponCarried == Weapon_M249)
                     allowPickup = false;
               }
               else if (m_isVIP || !RateGroundWeapon (ent))
                  allowPickup = false;
               else if (strcmp (STRING (ent->v.model) + 9, "medkit.mdl") == 0 && (pev->health >= 100))
                  allowPickup = false;
               else if ((strcmp (STRING (ent->v.model) + 9, "kevlar.mdl") == 0 || strcmp (STRING (ent->v.model) + 9, "battery.mdl") == 0) && pev->armorvalue >= 100) // armor vest
                  allowPickup = false;
               else if (strcmp (STRING (ent->v.model) + 9, "flashbang.mdl") == 0 && (pev->weapons & (1 << Weapon_Flashbang))) // concussion grenade
                  allowPickup = false;
               else if (strcmp (STRING (ent->v.model) + 9, "hegrenade.mdl") == 0 && (pev->weapons & (1 << Weapon_Esplosive))) // explosive grenade
                  allowPickup = false;
               else if (strcmp (STRING (ent->v.model) + 9, "smokegrenade.mdl") == 0 && (pev->weapons & (1 << Weapon_Smoke))) // smoke grenade
                  allowPickup = false;
            }
            else if (pickupType == PickupType_ShieldGun) // found a shield on ground?
            {
               if ((pev->weapons & (1 << Weapon_Elite)) || HasShield () || m_isVIP || (HasPrimaryWeapon () && !RateGroundWeapon (ent)))
                  allowPickup = false;
            }
            else if (GetTeam (GetEntity ()) == Team_Terrorist) // terrorist team specific
            {
               if (pickupType == PickupType_DroppedC4)
               {
                  allowPickup = true;
                  m_destOrigin = entityOrigin; // ensure we reached droped bomb

                  ChatterMessage (Chatter_FoundC4); // play info about that
               }
               else if (pickupType == PickupType_Hostage)
               {
                  m_itemIgnore = ent;
                  allowPickup = false;

                  if (m_skill > 80 && g_randGen.Long (0, 100) < 50 && GetCurrentTask ()->taskID != Task_MoveToPosition && GetCurrentTask ()->taskID != Task_Camp)
                  {
                     int index = FindDefendWaypoint (entityOrigin);

                     StartTask (Task_Camp, TASKPRI_CAMP, -1, GetWorldTime () + g_randGen.Float (60.0, 120.0), true); // push camp task on to stack
                     StartTask (Task_MoveToPosition, TASKPRI_MOVETOPOSITION, index, GetWorldTime () + g_randGen.Float (3.0, 6.0), true); // push move command

                     if (g_waypoint->GetPath (index)->vis.crouch <= g_waypoint->GetPath (index)->vis.stand)
                        m_campButtons |= IN_DUCK;
                     else
                        m_campButtons &= ~IN_DUCK;

                     ChatterMessage (Chatter_GoingToGuardHostages); // play info about that
                     return;
                  }
               }
               else if (pickupType == PickupType_PlantedC4)
               {
                  allowPickup = false;

                  if (!m_defendedBomb)
                  {
                     m_defendedBomb = true;

                     int index = FindDefendWaypoint (entityOrigin);
                     float timeMidBlowup = g_timeBombPlanted + ((g_botVar[Variable_C4Timer]->GetFloat () / 2) + g_botVar[Variable_C4Timer]->GetFloat () / 4) - g_waypoint->GetTravelTime (pev->maxspeed, pev->origin, g_waypoint->GetPath (index)->origin);

                     if (timeMidBlowup > GetWorldTime ())
                     {
                        RemoveCertainTask (Task_MoveToPosition); // remove any move tasks

                        StartTask (Task_Camp, TASKPRI_CAMP, -1, timeMidBlowup, true); // push camp task on to stack
                        StartTask (Task_MoveToPosition, TASKPRI_MOVETOPOSITION, index, timeMidBlowup, true); // push move command

                        if (g_waypoint->GetPath (index)->vis.crouch <= g_waypoint->GetPath (index)->vis.stand)
                           m_campButtons |= IN_DUCK;
                        else
                           m_campButtons &= ~IN_DUCK;

                        if (g_randGen.Long (0, 100) < 90)
                           ChatterMessage (Chatter_DefendingBombSite);
                     }
                     else
                        RadioMessage (Radio_ShesGonnaBlow); // issue an additional radio message
                  }
               }
            }
            else if (GetTeam (GetEntity ()) == Team_Counter)
            {
               if (pickupType == PickupType_Hostage)
               {
                  if (FNullEnt (ent) || (ent->v.health <= 0))
                     allowPickup = false; // never pickup dead hostage
                  else for (int i = 0; i < GetMaxClients (); i++)
                  {
                     if ((bot = g_botManager->GetBot (i)) != NULL && IsAlive (bot->GetEntity ()))
                     {
                        for (int j = 0; j < Const_MaxHostages; j++)
                        {
                           if (bot->m_hostages[j] == ent)
                           {
                              allowPickup = false;
                              break;
                           }
                        }
                     }
                  }
               }
               else if (pickupType == PickupType_PlantedC4 && !OutOfBombTimer ())
               {
                  if (m_states & (State_SeeingEnemy | State_SuspectEnemy))
                  {
                     allowPickup = false;
                     return;
                  }

                  if (g_randGen.Long (0, 100) < 90)
                     ChatterMessage (Chatter_FoundBombPlace);

                  allowPickup = !IsBombDefusing (g_waypoint->GetBombPosition ());
                  pickupType = PickupType_PlantedC4;

                  if (!m_defendedBomb && !allowPickup)
                  {
                     m_defendedBomb = true;

                     int index = FindDefendWaypoint (entityOrigin);
                     float fTimeBlowup = g_timeBombPlanted + g_botVar[Variable_C4Timer]->GetFloat () - g_waypoint->GetTravelTime (pev->maxspeed, pev->origin, g_waypoint->GetPath (index)->origin);

                     RemoveCertainTask (Task_MoveToPosition); // remove any move tasks

                     StartTask (Task_Camp, TASKPRI_CAMP, -1, fTimeBlowup, true); // push camp task on to stack
                     StartTask (Task_MoveToPosition, TASKPRI_MOVETOPOSITION, index, fTimeBlowup, true); // push move command

                     if (g_waypoint->GetPath (index)->vis.crouch <= g_waypoint->GetPath (index)->vis.stand)
                        m_campButtons |= IN_DUCK;
                     else
                        m_campButtons &= ~IN_DUCK;

                     if (g_randGen.Long (0, 100) < 90)
                        ChatterMessage (Chatter_DefendingBombSite);
                  }
               }
               else if (pickupType == PickupType_DroppedC4)
               {
                  m_itemIgnore = ent;
                  allowPickup = false;

                  if (m_skill > 80 && g_randGen.Long (0, 100) < 90)
                  {
                     int index = FindDefendWaypoint (entityOrigin);

                     StartTask (Task_Camp, TASKPRI_CAMP, -1, GetWorldTime () + g_randGen.Float (60.0, 120.0), true); // push camp task on to stack
                     StartTask (Task_MoveToPosition, TASKPRI_MOVETOPOSITION, index, GetWorldTime () + g_randGen.Float (10.0, 30.0), true); // push move command

                     if (g_waypoint->GetPath (index)->vis.crouch <= g_waypoint->GetPath (index)->vis.stand)
                        m_campButtons |= IN_DUCK;
                     else
                        m_campButtons &= ~IN_DUCK;

                     ChatterMessage (Chatter_GoingToGuardDoppedBomb); // play info about that
                     return;
                  }
               }
            }

            // if condition valid
            if (allowPickup)
            {
               minDistance = distance; // update the minimum distance
               pickupOrigin = entityOrigin; // remember location of entity
               pickupItem = ent; // remember this entity

               m_pickupType = pickupType;
            }
            else
               pickupType = PickupType_None;
         }
      }
   } // end of the while loop

   if (!FNullEnt (pickupItem))
   {
      for (int i = 0; i < GetMaxClients (); i++)
      {
         if ((bot = g_botManager->GetBot (i)) != NULL && IsAlive (bot->GetEntity ()) && bot->m_pickupItem == pickupItem)
         {
            m_pickupItem = NULL;
            m_pickupType = PickupType_None;

            return;
         }
      }

      if (pickupOrigin.z > EyePosition ().z + 3.0 || IsDeadlyDrop (pickupOrigin)) // check if item is too high to reach, check if getting the item would hurt bot
      {
         m_pickupItem = NULL;
         m_pickupType = PickupType_None;

         return;
      }
      m_pickupItem = pickupItem; // save pointer of picking up entity
   }
}

void Bot::GetCampDirection (Vector *dest)
{
   // this function check if view on last enemy position is blocked - replace with better vector then
   // mostly used for getting a good camping direction vector if not camping on a camp waypoint

   TraceResult tr;
   Vector src = EyePosition ();

   TraceLine (src, *dest, true, GetEntity (), &tr);

   // check if the trace hit something...
   if (tr.flFraction < 1.0)
   {
      float length = (tr.vecEndPos - src).GetLengthSquared ();

      if (length > 10000)
         return;

      float minDistance = FLT_MAX;
      float maxDistance = FLT_MAX;

      int enemyIndex = -1, tempIndex = -1;

      // find nearest waypoint to bot and position
      for (int i = 0; i < g_numWaypoints; i++)
      {
         float distance = (g_waypoint->GetPath (i)->origin - pev->origin).GetLengthSquared ();

         if (distance < minDistance)
         {
            minDistance = distance;
            tempIndex = i;
         }
         distance = (g_waypoint->GetPath (i)->origin - *dest).GetLengthSquared ();

         if (distance < maxDistance)
         {
            maxDistance = distance;
            enemyIndex = i;
         }
      }

      if (tempIndex == -1 || enemyIndex == -1)
         return;

      minDistance = FLT_MAX;

      int lookAtWaypoint = -1;
      Path_t *path = g_waypoint->GetPath (tempIndex);

      for (int i = 0; i < Const_MaxPathIndex; i++)
      {
         if (path->index[i] == -1)
            continue;

         float distance = g_waypoint->GetPathDistance (path->index[i], enemyIndex);

         if (distance < minDistance)
         {
            minDistance = distance;
            lookAtWaypoint = path->index[i];
         }
      }
      if (lookAtWaypoint != -1 && lookAtWaypoint < g_numWaypoints)
         *dest = g_waypoint->GetPath (lookAtWaypoint)->origin;
   }
}

void Bot::SwitchChatterIcon (bool show)
{
   // this function depending on show boolen, shows/remove chatter, icon, on the head of bot.

   if (g_gameVersion == Version_Old)
      return;

   for (int i = 0; i < GetMaxClients (); i++)
   {
      edict_t *ent = INDEXENT (i);

      if (!IsValidPlayer (ent) || IsValidBot (ent) || GetTeam (ent) != GetTeam (GetEntity ()))
         continue;

      MESSAGE_BEGIN (MSG_ONE, g_netMsg->GetId (NetMessage_BotVoice), NULL, ent); // begin message
         WRITE_BYTE (show); // switch on/off
         WRITE_BYTE (GetIndex ());
      MESSAGE_END ();

   }
}

void Bot::InstantChatterMessage (int type)
{
   // this function sends instant chatter messages.

   if (g_botVar[Variable_Communication]->GetInt () != 2 || g_chatterFactory[type].IsEmpty () || g_gameVersion == Version_Old || !g_sendAudioFinished)
      return;

   if (IsAlive (GetEntity ()))
      SwitchChatterIcon (true);

   static float reportTime = GetWorldTime ();

   // delay only reportteam
   if (type == Radio_ReportTeam)
   {
      if (reportTime >= GetWorldTime ())
         return;

      reportTime = GetWorldTime () + g_randGen.Float (30.0, 80.0);
   }

   String defaultSound = g_chatterFactory[type].GetRandomElement ().name;
   String painSound = g_chatterFactory[Chatter_DiePain].GetRandomElement ().name;

   for (int i = 0; i < GetMaxClients (); i++)
   {
      edict_t *ent = INDEXENT (i);

      if (!IsValidPlayer (ent) || IsValidBot (ent) || GetTeam (ent) != GetTeam (GetEntity ()))
         continue;

      g_sendAudioFinished = false;

      MESSAGE_BEGIN (MSG_ONE, g_netMsg->GetId (NetMessage_SendAudio), NULL, ent); // begin message
         WRITE_BYTE (GetIndex ());

         if (pev->deadflag & DEAD_DYING)
            WRITE_STRING (FormatBuffer ("%s/%s.wav", g_botVar[Variable_ChatterPath]->GetString (), painSound.ToString ()));
         else if (!(pev->deadflag & DEAD_DEAD))
            WRITE_STRING (FormatBuffer ("%s/%s.wav", g_botVar[Variable_ChatterPath]->GetString (), defaultSound.ToString ()));

         WRITE_SHORT (m_voicePitch);
      MESSAGE_END ();

      g_sendAudioFinished = true;
   }
}

void Bot::RadioMessage (int message)
{
   // this function inserts the radio message into the message queue

   if (g_botVar[Variable_Communication]->GetInt () == 0 || GetNearbyFriendsNearPosition (pev->origin, 9999) == 0)
      return;

   if (g_chatterFactory[message].IsEmpty () || g_gameVersion == Version_Old || g_botVar[Variable_Communication]->GetInt () != 2)
      g_radioInsteadVoice = true; // use radio instead voice

   m_radioSelect = message;
   PushMessageQueue (GameMessage_RadioMsg);
}

void Bot::ChatterMessage (int message)
{
   // this function inserts the voice message into the message queue (mostly same as above)

   if (g_botVar[Variable_Communication]->GetInt () != 2 || g_chatterFactory[message].IsEmpty () || GetNearbyFriendsNearPosition (pev->origin, 9999) == 0)
      return;

   bool shouldExecute = false;

   if (m_voiceTimers[message] < GetWorldTime () || m_voiceTimers[message] == FLT_MAX)
   {
      if (m_voiceTimers[message] != FLT_MAX)
         m_voiceTimers[message] = GetWorldTime () + g_chatterFactory[message][0].repeatTime;

      shouldExecute = true;
   }

   if (!shouldExecute)
      return;

   m_radioSelect = message;
   PushMessageQueue (GameMessage_RadioMsg);
}

void Bot::CheckMessageQueue (void)
{
   // this function checks and executes pending messages

   // no new message?
   if (m_actMessageIndex == m_pushMessageIndex)
      return;

   // get message from stack
   int currentQueueMessage = GetMessageQueue ();

   // nothing to do?
   if (currentQueueMessage == GameMessage_Idle || (currentQueueMessage == GameMessage_RadioMsg && g_botVar[Variable_Deathmatch]->GetInt () == 2))
      return;

   int team = GetTeam (GetEntity ());

   switch (currentQueueMessage)
   {
   case GameMessage_BuyStuff: // general buy message

      // buy weapon
      if (m_nextBuyTime > GetWorldTime ())
      {
         // keep sending message
         PushMessageQueue (GameMessage_BuyStuff);
         return;
      }

      if (!m_inBuyZone || g_botVar[Variable_Deathmatch]->GetBool ())
      {
         m_buyPending = true;
         m_buyingFinished = true;

         break;
      }

      m_buyPending = false;
      m_nextBuyTime = GetWorldTime () + g_randGen.Float (0.3, 0.8);

      // if bot buying is off then no need to buy
      if (!g_botVar[Variable_BotBuy]->GetBool ())
         m_buyState = 6;

      // if fun-mode no need to buy
      if (g_botVar[Variable_KnifeMode]->GetBool ())
      {
         m_buyState = 6;
         SelectWeaponByName ("weapon_knife");
      }

      // prevent vip from buying
      if (g_mapType & MapType_As)
      {
         if (*(INFOKEY_VALUE (GET_INFOKEYBUFFER (GetEntity ()), "model")) == 'v')
         {
            m_isVIP = true;
            m_buyState = 6;
            m_pathType = 1;
         }
      }

      // prevent terrorists from buying on es maps
      if ((g_mapType & MapType_Es) && GetTeam (GetEntity ()) == Team_Terrorist)
         m_buyState = 6;

      // prevent teams from buying on fun maps
      if (g_mapType & (MapType_Ka | MapType_Fy))
      {
         m_buyState = 6;

         if (g_mapType & MapType_Ka)
            g_botVar[Variable_KnifeMode]->SetInt (1);
      }

      if (m_buyState > 5)
      {
         m_buyingFinished = true;
         return;
      }

      PushMessageQueue (GameMessage_Idle);
      PerformWeaponPurchase ();

      break;

   case GameMessage_RadioMsg: // general radio message issued
     // if last bot radio command (global) happened just a second ago, delay response
      if (g_lastRadioTime[team] + 1.0 < GetWorldTime ())
      {
         // if same message like previous just do a yes/no
         if (m_radioSelect != Radio_Affirmative && m_radioSelect != Radio_Negative)
         {
            if (m_radioSelect == g_lastRadio[team] && g_lastRadioTime[team] + 1.5 > GetWorldTime ())
               m_radioSelect = -1;
            else
            {
               if (m_radioSelect != Radio_ReportingIn)
                  g_lastRadio[team] = m_radioSelect;
               else
                  g_lastRadio[team] = -1;

               for (int i = 0; i < GetMaxClients (); i++)
               {
                  if (g_botManager->GetBot (i))
                  {
                     if (pev != g_botManager->GetBot (i)->pev && GetTeam (g_botManager->GetBot (i)->GetEntity ()) == team)
                     {
                        g_botManager->GetBot (i)->m_radioOrder = m_radioSelect;
                        g_botManager->GetBot (i)->m_radioEntity = GetEntity ();
                     }
                  }
               }
            }
         }

         if (m_radioSelect == Radio_ReportingIn)
         {
            switch (GetCurrentTask ()->taskID)
            {
            case Task_Normal:
               if (GetCurrentTask ()->data != -1)
               {
                  if (g_waypoint->GetPath (GetCurrentTask ()->data)->flags & PointFlag_Goal)
                  {
                     if ((g_mapType & MapType_De) && GetTeam (GetEntity ()) == Team_Terrorist && (pev->weapons & (1 << Weapon_BombC4)))
                        InstantChatterMessage (Chatter_GoingToPlantBomb);
                     else
                        InstantChatterMessage (Chatter_Nothing);
                  }
                  else if (g_waypoint->GetPath (GetCurrentTask ()->data)->flags & PointFlag_Rescue)
                     InstantChatterMessage (Chatter_RescuingHostages);
                  else if (g_waypoint->GetPath (GetCurrentTask ()->data)->flags & PointFlag_Camp)
                     InstantChatterMessage (Chatter_GoingToCamp);
                  else
                     InstantChatterMessage (Chatter_HearSomething);
               }
               else
                  InstantChatterMessage (Chatter_ReportingIn);
               break;

            case Task_MoveToPosition:
               InstantChatterMessage (Chatter_GoingToCamp);
               break;

            case Task_Camp:
               if (g_bombPlanted && GetTeam (GetEntity ()) == Team_Terrorist)
                  InstantChatterMessage (Chatter_GuardDroppedC4);
               else if (m_inVIPZone && GetTeam (GetEntity ()) == Team_Terrorist)
                  InstantChatterMessage (Chatter_GuardingVipSafety);
               else
                  InstantChatterMessage (Chatter_Camp);
               break;

            case Task_PlantBomb:
               InstantChatterMessage (Chatter_PlantingC4);
               break;

            case Task_DefuseBomb:
               InstantChatterMessage (Chatter_DefusingC4);
               break;

            case Task_AttackEnemy:
               InstantChatterMessage (Chatter_InCombat);
               break;

            case Task_Hide:
            case Task_SeekCover:
               InstantChatterMessage (Chatter_SeeksEnemy);
               break;

            default:
               InstantChatterMessage (Chatter_Nothing);
               break;
            }
         }

         if (m_radioSelect != Radio_ReportingIn && g_radioInsteadVoice || g_botVar[Variable_Communication]->GetInt () != 2 || g_chatterFactory[m_radioSelect].IsEmpty () || g_gameVersion == Version_Old)
         {
            if (m_radioSelect < Radio_GoGoGo)
               FakeClientCommand (GetEntity (), "radio1");
            else if (m_radioSelect < Radio_Affirmative)
            {
               m_radioSelect -= Radio_GoGoGo - 1;
               FakeClientCommand (GetEntity (), "radio2");
            }
            else
            {
               m_radioSelect -= Radio_Affirmative - 1;
               FakeClientCommand (GetEntity (), "radio3");
            }

            // select correct menu item for this radio message
            FakeClientCommand (GetEntity (), "menuselect %d", m_radioSelect);
         }
         else if (m_radioSelect != -1 && m_radioSelect != Radio_ReportingIn)
            InstantChatterMessage (m_radioSelect);

         g_radioInsteadVoice = false; // reset radio to voice
         g_lastRadioTime[team] = GetWorldTime (); // store last radio usage
      }
      else
         PushMessageQueue (GameMessage_RadioMsg);
      break;

   // team independent saytext
   case GameMessage_SayMsg:
      SayText (m_tempStrings);
      break;

   // team dependent saytext
   case GameMessage_TeamSay:
      TeamSayText (m_tempStrings);
      break;

   case GameMessage_Idle:
   default:
      return;
   }
}

bool Bot::IsRestricted (int weaponIndex)
{
   // this function checks for weapon restrictions.

   if (IsNullString (g_botVar[Variable_RestWeapons]->GetString ()))
      return false; // no banned weapons

   Array <String> bannedWeapons = String (g_botVar[Variable_RestWeapons]->GetString ()).Split (';');

   IterateArray (bannedWeapons, i)
   {
      const char *banned = STRING (GetWeaponReturn (true, NULL, weaponIndex));

      // check is this weapon is banned
      if (strncmp (bannedWeapons[i].GetBuffer (), banned, bannedWeapons[i].GetLength ()) == 0)
         return true;
   }
   return IsRestrictedAMX (weaponIndex);
}

bool Bot::IsRestrictedAMX (int weaponIndex)
{
   // this function checks restriction set by AMX Mod, this function code is courtesy of KWo.

   const char *restrictedWeapons = CVAR_GET_STRING ("amx_restrweapons");
   const char *restrictedEquipment = CVAR_GET_STRING ("amx_restrequipammo");

   // check for weapon restrictions
   if ((1 << weaponIndex) & (WeaponBits_Primary | WeaponBits_Secondary | Weapon_ShieldGun))
   {
      if (IsNullString (restrictedWeapons))
         return false;

      int indices[] = {4, 25, 20, -1, 8, -1, 12, 19, -1, 5, 6, 13, 23, 17, 18, 1, 2, 21, 9, 24, 7, 16, 10, 22, -1, 3, 15, 14, 0, 11};

      // find the weapon index
      int index = indices[weaponIndex - 1];

      // validate index range
      if (index < 0 || index >= static_cast <int> (strlen (restrictedWeapons)))
         return false;

      return restrictedWeapons[index] != '0';
   }
   else // check for equipment restrictions
   {
      if (IsNullString (restrictedEquipment))
         return false;

      int indices[] = {-1, -1, -1, 3, -1, -1, -1, -1, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 2, -1, -1, -1, -1, -1, 0, 1, 5};

      // find the weapon index
      int index = indices[weaponIndex - 1];

      // validate index range
      if (index < 0 || index >= static_cast <int> (strlen (restrictedEquipment)))
         return false;

      return restrictedEquipment[index] != '0';
   }
}

bool Bot::IsMorePowerfulWeaponCanBeBought (void)
{
   // this function determines currently owned primary weapon, and checks if bot has
   // enough money to buy more powerful weapon.

   // if bot is not rich enough or non-standard weapon mode enabled return false
   if (g_weaponSelect[25].teamStandard != 1 || m_moneyAmount < 4000 || IsNullString (g_botVar[Variable_RestWeapons]->GetString ()))
      return false;

   // also check if bot has really bad weapon, maybe it's time to change it
   if (UsesBadPrimary ())
      return true;

   Array <String> bannedWeapons = String (g_botVar[Variable_RestWeapons]->GetString ()).Split (';');

   // check if its banned
   IterateArray (bannedWeapons, i)
   {
      if (m_currentWeapon == GetWeaponReturn (false, bannedWeapons[i].ToString ()))
         return true;
   }

   if (m_currentWeapon == Weapon_Scout && m_moneyAmount > 5000)
      return true;
   else if (m_currentWeapon == Weapon_Mp5Navy && m_moneyAmount > 6000)
      return true;
   else if (m_currentWeapon == Weapon_M3 || m_currentWeapon == Weapon_Xm1014 && m_moneyAmount > 4000)
      return true;

   return false;
}

void Bot::PerformWeaponPurchase (void)
{
   // this function does all the work in selecting correct buy menus for most weapons/items

   WeaponSelect_t *selectedWeapon = NULL;
   m_nextBuyTime = GetWorldTime ();

   int count = 0, foundWeapons = 0, economicsPrice = 0;
   int choices[Const_NumWeapons];

   // select the priority tab for this personality
   int *ptr = g_weaponPrefs[m_personality] + Const_NumWeapons;
   int team = GetTeam (GetEntity ());

   bool isPistolMode = (g_weaponSelect[25].teamStandard == -1) && (g_weaponSelect[3].teamStandard == 2);

   switch (m_buyState)
   {
   case 0: // if no primary weapon and bot has some money, buy a primary weapon
      if ((!HasShield () && !HasPrimaryWeapon ()) && (g_botManager->EconomicsValid (team) || IsMorePowerfulWeaponCanBeBought ()))
      {
         do
         {
            bool ignoreWeapon = false;

            ptr--;

            InternalAssert (*ptr > -1);
            InternalAssert (*ptr < Const_NumWeapons);

            selectedWeapon = &g_weaponSelect[*ptr];
            count++;

            if (selectedWeapon->buyGroup == 1)
               continue;

            // weapon available for every team?
            if ((g_mapType & MapType_As) && selectedWeapon->teamAS != 2 && selectedWeapon->teamAS != team)
               continue;

            // ignore weapon if this weapon not supported by currently running cs version...
            if (g_gameVersion == Version_Old && selectedWeapon->buySelect == -1)
               continue;

            // ignore weapon if this weapon is not targeted to out team....
            if (selectedWeapon->teamStandard != 2 && selectedWeapon->teamStandard != team)
               continue;

            // ignore weapon if this weapon is restricted
            if (IsRestricted (selectedWeapon->id))
               continue;

            // filter out weapons with bot economics (thanks for idea to nicebot project)
            switch (m_personality)
            {
            case Personality_Rusher:
               economicsPrice = g_botBuyEconomyTable[8];
               break;

            case Personality_Careful:
               economicsPrice = g_botBuyEconomyTable[7];
               break;

            case Personality_Normal:
               economicsPrice = g_botBuyEconomyTable[9];
               break;
            }

            if (team == Team_Counter)
            {
               switch (selectedWeapon->id)
               {
               case Weapon_Tmp:
               case Weapon_Ump45:
               case Weapon_P90:
               case Weapon_Mp5Navy:
                  if (m_moneyAmount > g_botBuyEconomyTable[1] + economicsPrice)
                     ignoreWeapon = true;
                  break;
               }

               if (selectedWeapon->id == Weapon_ShieldGun && m_moneyAmount > g_botBuyEconomyTable[10])
                  ignoreWeapon = true;
            }
            else if (team == Team_Terrorist)
            {
               switch (selectedWeapon->id)
               {
               case Weapon_Ump45:
               case Weapon_Mac10:
               case Weapon_P90:
               case Weapon_Mp5Navy:
               case Weapon_Scout:
                  if (m_moneyAmount > g_botBuyEconomyTable[2] + economicsPrice)
                     ignoreWeapon = true;
                  break;
               }
            }

            switch (selectedWeapon->id)
            {
            case Weapon_Xm1014:
            case Weapon_M3:
               if (m_moneyAmount < g_botBuyEconomyTable[3] || m_moneyAmount > g_botBuyEconomyTable[4])
                  ignoreWeapon = true;
               break;
            }

            switch (selectedWeapon->id)
            {
            case Weapon_Sig550:
            case Weapon_G3SG1:
            case Weapon_Magnum:
            case Weapon_M249:
               if (m_moneyAmount < g_botBuyEconomyTable[5] || m_moneyAmount > g_botBuyEconomyTable[6])
                  ignoreWeapon = true;
               break;
            }

            if (ignoreWeapon && g_weaponSelect[25].teamStandard == 1 && g_botVar[Variable_Economics]->GetBool ())
               continue;

            int moneySave = g_randGen.Long (900, 1100);

            if (g_botManager->GetLastWinner () == team)
               moneySave = 0;

            if (selectedWeapon->price <= (m_moneyAmount - moneySave))
               choices[foundWeapons++] = *ptr;

         } while (count < Const_NumWeapons && foundWeapons < 4);

         // found a desired weapon?
         if (foundWeapons > 0)
         {
            int chosenWeapon;

            // choose randomly from the best ones...
            if (foundWeapons > 1)
               chosenWeapon = choices[g_randGen.Long (0, foundWeapons - 1)];
            else
               chosenWeapon = choices[foundWeapons - 1];

            selectedWeapon = &g_weaponSelect[chosenWeapon];
         }
         else
            selectedWeapon = NULL;

         if (selectedWeapon != NULL)
         {
            FakeClientCommand (GetEntity (), "buy;menuselect %d", selectedWeapon->buyGroup);

            if (g_gameVersion == Version_Old)
               FakeClientCommand (GetEntity (), "menuselect %d", selectedWeapon->buySelect);
            else // SteamCS buy menu is different from the old one
            {
               if (GetTeam (GetEntity ()) == Team_Terrorist)
                  FakeClientCommand (GetEntity (), "menuselect %d", selectedWeapon->newBuySelectT);
               else
                  FakeClientCommand (GetEntity (), "menuselect %d", selectedWeapon->newBuySelectCT);
            }
         }
      }
      else if (HasPrimaryWeapon () && !HasShield ())
         m_reloadState = ReloadState_Primary;

      break;

   case 1: // if armor is damaged and bot has some money, buy some armor
      if (pev->armorvalue < g_randGen.Long (50, 80) && (isPistolMode || (g_botManager->EconomicsValid (team) && HasPrimaryWeapon ())))
      {
         // if bot is rich, buy kevlar + helmet, else buy a single kevlar
         if (m_moneyAmount > 1500 && !IsRestricted (Weapon_ArmorHelm))
            FakeClientCommand (GetEntity (), "buyequip;menuselect 2");
         else if (!IsRestricted (Weapon_Armor))
            FakeClientCommand (GetEntity (), "buyequip;menuselect 1");
      }
      break;

   case 2: // if bot has still some money, buy a better secondary weapon
      if (isPistolMode || (HasPrimaryWeapon () && (pev->weapons & ((1 << Weapon_Usp) | (1 << Weapon_Glock18))) && m_moneyAmount > g_randGen.Long (7500, 9000)))
      {
         do
         {
            ptr--;

            InternalAssert (*ptr > -1);
            InternalAssert (*ptr < Const_NumWeapons);

            selectedWeapon = &g_weaponSelect[*ptr];
            count++;

            if (selectedWeapon->buyGroup != 1)
               continue;

            // ignore weapon if this weapon is restricted
            if (IsRestricted (selectedWeapon->id))
               continue;

            // weapon available for every team?
            if ((g_mapType & MapType_As) && selectedWeapon->teamAS != 2 && selectedWeapon->teamAS != team)
               continue;

            if (g_gameVersion == Version_Old && selectedWeapon->buySelect == -1)
               continue;

            if (selectedWeapon->teamStandard != 2 && selectedWeapon->teamStandard != team)
               continue;

            if (selectedWeapon->price <= (m_moneyAmount - g_randGen.Long (100, 200)))
               choices[foundWeapons++] = *ptr;

         } while (count < Const_NumWeapons && foundWeapons < 4);

         // found a desired weapon?
         if (foundWeapons > 0)
         {
            int chosenWeapon;

            // choose randomly from the best ones...
            if (foundWeapons > 1)
               chosenWeapon = choices[g_randGen.Long (0, foundWeapons - 1)];
            else
               chosenWeapon = choices[foundWeapons - 1];

            selectedWeapon = &g_weaponSelect[chosenWeapon];
         }
         else
            selectedWeapon = NULL;

         if (selectedWeapon != NULL)
         {
            FakeClientCommand (GetEntity (), "buy;menuselect %d", selectedWeapon->buyGroup);

            if (g_gameVersion == Version_Old)
               FakeClientCommand (GetEntity (), "menuselect %d", selectedWeapon->buySelect);
            else // SteamCS buy menu is different from old one
            {
               if (GetTeam (GetEntity ()) == Team_Terrorist)
                  FakeClientCommand (GetEntity (), "menuselect %d", selectedWeapon->newBuySelectT);
               else
                  FakeClientCommand (GetEntity (), "menuselect %d", selectedWeapon->newBuySelectCT);
            }
         }
      }
      break;

   case 3: // if bot has still some money, choose if bot should buy a grenade or not
      if (g_randGen.Long (1, 100) < g_grenadeBuyPrecent[0] && m_moneyAmount >= 300 && !IsRestricted (Weapon_Esplosive))
      {
         // buy a he grenade
         FakeClientCommand (GetEntity (), "buyequip");
         FakeClientCommand (GetEntity (), "menuselect 4");
      }

      if (g_randGen.Long (1, 100) < g_grenadeBuyPrecent[1] && m_moneyAmount >= 200 && g_botManager->EconomicsValid (team) && !IsRestricted (Weapon_Flashbang))
      {
         // buy a concussion grenade, i.e., 'flashbang'
         FakeClientCommand (GetEntity (), "buyequip");
         FakeClientCommand (GetEntity (), "menuselect 3");
      }

      if (g_randGen.Long (1, 100) < g_grenadeBuyPrecent[2] && m_moneyAmount >= 300 && g_botManager->EconomicsValid (team) && !IsRestricted (Weapon_Smoke))
      {
         // buy a smoke grenade
         FakeClientCommand (GetEntity (), "buyequip");
         FakeClientCommand (GetEntity (), "menuselect 5");
      }
      break;

   case 4: // if bot is CT and we're on a bomb map, randomly buy the defuse kit
      if ((g_mapType & MapType_De) && GetTeam (GetEntity ()) == Team_Counter && g_randGen.Long (1, 100) < 80 && m_moneyAmount > 200 && !IsRestricted (Weapon_Defuser))
      {
         if (g_gameVersion == Version_Old)
            FakeClientCommand (GetEntity (), "buyequip;menuselect 6");
         else
            FakeClientCommand (GetEntity (), "defuser"); // use alias in SteamCS
      }
      break;

   case 5: // buy enough primary & secondary ammo (do not check for money here)
      for (int i = 0; i <= 5; i++)
         FakeClientCommand (GetEntity (), "buyammo%d", g_randGen.Long (1, 2)); // simulate human

      // buy enough secondary ammo
      if (HasPrimaryWeapon ())
         FakeClientCommand (GetEntity (), "buy;menuselect 7");

      // buy enough primary ammo
      FakeClientCommand (GetEntity (), "buy;menuselect 6");

      // try to reload secondary weapon
      if (m_reloadState != ReloadState_Primary)
         m_reloadState = ReloadState_Secondary;

      break;
   }

   m_buyState++;
   PushMessageQueue (GameMessage_BuyStuff);
}

Task_t *ClampDesire (Task_t *first, float min, float max)
{
   // this function iven some values min and max, clamp the inputs to be inside the [min, max] range.

   if (first->desire < min)
      first->desire = min;
   else if (first->desire > max)
      first->desire = max;

   return first;
}

Task_t *MaxDesire (Task_t *first, Task_t *second)
{
   // this function returns the behavior having the higher activation level.

   if (first->desire > second->desire)
      return first;

   return second;
}

Task_t *SubsumeDesire (Task_t *first, Task_t *second)
{
   // this function returns the first behavior if its activation level is anything higher than zero.

   if (first->desire > 0)
      return first;

   return second;
}

Task_t *ThresholdDesire (Task_t *first, float threshold, float desire)
{
   // this function returns the input behavior if it's activation level exceeds the threshold, or some default
   // behavior otherwise.

   if (first->desire < threshold)
      first->desire = desire;

   return first;
}

float HysteresisDesire (float cur, float min, float max, float old)
{
   // this function clamp the inputs to be the last known value outside the [min, max] range.

   if (cur <= min || cur >= max)
      old = cur;

   return old;
}

void Bot::SetConditions (void)
{
   // this function carried out each frame. does all of the sensing, calculates emotions and finally sets the desired
   // action after applying all of the Filters

   int team = GetTeam (GetEntity ());
   m_aimFlags = 0;

   // slowly increase/decrease dynamic emotions back to their base level
   if (m_nextEmotionUpdate < GetWorldTime ())
   {
      if (g_botVar[Variable_HardcoreMode]->GetBool ())
      {
         m_agressionLevel *= 2;
         m_fearLevel /= 2;
      }

      if (m_agressionLevel > m_baseAgressionLevel)
         m_agressionLevel -= 0.05;
      else
         m_agressionLevel += 0.05;

      if (m_fearLevel > m_baseFearLevel)
         m_fearLevel -= 0.05;
      else
         m_fearLevel += 0.05;

      if (m_agressionLevel < 0.0)
         m_agressionLevel = 0.0;

      if (m_fearLevel < 0.0)
         m_fearLevel = 0.0;

      m_nextEmotionUpdate = GetWorldTime () + 0.5;
   }

   // does bot see an enemy?
   if (LookupEnemy ())
      m_states |= State_SeeingEnemy;
   else
   {
      m_states &= ~State_SeeingEnemy;
      m_enemy = NULL;
   }

   // did bot just kill an enemy?
   if (!FNullEnt (m_lastVictim))
   {
      if (GetTeam (m_lastVictim) != team)
      {
         // add some aggression because we just killed somebody
         m_agressionLevel += 0.1;

         if (m_agressionLevel > 1.0)
            m_agressionLevel = 1.0;

         if (g_randGen.Long (1, 100) > 50)
            ChatMessage (Chat_Killing);

         if (g_randGen.Long (1, 100) < 10)
            RadioMessage (Radio_EnemyDown);
         else
         {
            if ((m_lastVictim->v.weapons & (1 << Weapon_Magnum)) || (m_lastVictim->v.weapons & (1 << Weapon_Scout)) ||  (m_lastVictim->v.weapons & (1 << Weapon_G3SG1)) || (m_lastVictim->v.weapons & (1 << Weapon_Sig550)))
               ChatterMessage (Chatter_SniperKilled);
            else
            {
               switch (GetNearbyEnemiesNearPosition (pev->origin, 9999))
               {
               case 0:
                  if (g_randGen.Long (0, 100) < 50)
                     ChatterMessage (Chatter_NoEnemiesLeft);
                  else
                     ChatterMessage (Chatter_EnemyDown);
                  break;

               case 1:
                  ChatterMessage (Chatter_OneEnemyLeft);
                  break;

               case 2:
                  ChatterMessage (Chatter_TwoEnemiesLeft);
                  break;

               case 3:
                  ChatterMessage (Chatter_ThreeEnemiesLeft);
                  break;

               default:
                  ChatterMessage (Chatter_EnemyDown);
               }
            }
         }

         // if no more enemies found AND bomb planted, switch to knife to get to bombplace faster
         if (GetTeam (GetEntity ()) == Team_Counter && m_currentWeapon != Weapon_Knife && GetNearbyEnemiesNearPosition (pev->origin, 9999) == 0 && g_bombPlanted)
         {
            SelectWeaponByName ("weapon_knife");

            // order team to regroup
            RadioMessage (Radio_RegroupTeam);
         }
      }
      else
      {
         ChatMessage (Chat_TeamKill, true);
         ChatterMessage (Chatter_TeamKill);
      }
      m_lastVictim = NULL;
   }

   // check if our current enemy is still valid
   if (!FNullEnt (m_lastEnemy))
   {
      if (!IsAlive (m_lastEnemy) && m_shootAtDeadTime < GetWorldTime ())
      {
         m_lastEnemyOrigin = Vector::GetNull ();
         m_lastEnemy = NULL;
      }
   }
   else
   {
      m_lastEnemyOrigin = Vector::GetNull ();
      m_lastEnemy = NULL;
   }

   // don't listen if seeing enemy, just checked for sounds or being blinded (because its inhuman)
   if (!g_botVar[Variable_DontShoot]->GetBool () && m_soundUpdateTime <= GetWorldTime () && m_blindTime < GetWorldTime ())
   {
      ReactOnSound ();
      m_soundUpdateTime = GetWorldTime () + g_botVar[Variable_TimerSound]->GetFloat ();
   }
   else if (m_heardSoundTime < GetWorldTime ())
      m_states &= ~State_HearingEnemy;

   if (FNullEnt (m_enemy) && !FNullEnt (m_lastEnemy) && m_lastEnemyOrigin != Vector::GetNull ())
   {
      TraceResult tr;
      TraceLine (EyePosition (), m_lastEnemyOrigin, true, GetEntity (), &tr);

      if ((pev->origin - m_lastEnemyOrigin).GetLength () < 1600.0 && (tr.flFraction >= 0.2 || tr.pHit != g_worldEdict))
      {
         m_aimFlags |= AimPos_PredictEnemy;

         if (EntityIsVisible (m_lastEnemyOrigin))
            m_aimFlags |= AimPos_LastEnemy;
      }
   }

   // check if throwing a grenade is a good thing to do...
   if (!g_botVar[Variable_DontShoot]->GetBool () && !g_botVar[Variable_KnifeMode]->GetBool () && m_grenadeCheckTime < GetWorldTime () && !m_isUsingGrenade && GetCurrentTask ()->taskID != Task_PlantBomb && GetCurrentTask ()->taskID != Task_DefuseBomb && !m_isReloading && IsAlive (m_lastEnemy))
   {
      // check again in some seconds
      m_grenadeCheckTime = GetWorldTime () + g_botVar[Variable_TimerGrenade]->GetFloat ();

      // check if we have grenades to throw
      int grenadeToThrow = CheckGrenades ();

      // if we don't have grenades no need to check it this round again
      if (grenadeToThrow == -1)
      {
         m_grenadeCheckTime = GetWorldTime () + 30.0; // changed since, conzero can drop grens from dead players
         m_states &= ~(State_ThrowExplosive | State_ThrowFlashbang | State_ThrowSmoke);
      }
      else
      {
         // care about different types of grenades
         if ((grenadeToThrow == Weapon_Esplosive || grenadeToThrow == Weapon_Smoke) && g_randGen.Long (0, 100) < 45 && !(m_states & (State_SeeingEnemy | State_ThrowExplosive | State_ThrowFlashbang | State_ThrowSmoke)))
         {
            float distance = (m_lastEnemy->v.origin - pev->origin).GetLength ();

            // is enemy to high to throw
            if ((m_lastEnemy->v.origin.z > (pev->origin.z + 500.0)) || !(m_lastEnemy->v.flags & FL_ONGROUND))
               distance = FLT_MAX; // just some crazy value

            // enemy is within a good throwing distance ?
            if (distance > (grenadeToThrow == Weapon_Smoke ? 400 : 600) && distance < 1200)
            {
               // care about different types of grenades
               if (grenadeToThrow == Weapon_Esplosive)
               {
                  bool allowThrowing = true;

                  // check for teammates
                  if (GetNearbyFriendsNearPosition (m_lastEnemy->v.origin, 256) > 0)
                     allowThrowing = false;

                  if (allowThrowing)
                  {
                     Vector enemyPredict = ((m_lastEnemy->v.velocity * 0.5).SkipZ () + m_lastEnemy->v.origin);
                     int searchTab[4], count = 4;

                     float searchRadius = m_lastEnemy->v.velocity.GetLength2D ();

                     // check the search radius
                     if (searchRadius < 128.0)
                        searchRadius = 128.0;

                     // search waypoints
                     g_waypoint->FindInRadius (enemyPredict, searchRadius, searchTab, &count);

                     while (count > 0)
                     {
                        allowThrowing = true;

                        // check the throwing
                        m_throw = g_waypoint->GetPath (searchTab[count--])->origin;
                        Vector src = CheckThrow (EyePosition (), m_throw);

                        if (src.GetLengthSquared () < 100.0)
                           src = CheckToss (EyePosition (), m_throw);

                        if (src == Vector::GetNull ())
                           allowThrowing = false;
                        else
                           break;
                     }
                  }

                  // start explosive grenade throwing?
                  if (allowThrowing)
                     m_states |= State_ThrowExplosive;
                  else
                     m_states &= ~State_ThrowExplosive;
               }
               else if (grenadeToThrow == Weapon_Smoke)
               {
                  // start smoke grenade throwing?
                  if ((m_states & State_SeeingEnemy) && GetShootingConeDeviation (m_enemy, &pev->origin) >= 0.90)
                     m_states &= ~State_ThrowSmoke;
                  else
                     m_states |= State_ThrowSmoke;
               }
            }
         }
         else if (IsAlive (m_lastEnemy) && grenadeToThrow == Weapon_Flashbang && (m_lastEnemy->v.origin - pev->origin).GetLength () < 800 && !(m_aimFlags & AimPos_Enemy) && g_randGen.Long (0, 100) < 50)
         {
            bool allowThrowing = true;
            Array <int> inRadius;

            g_waypoint->FindInRadius (inRadius, 256, m_lastEnemy->v.origin + (m_lastEnemy->v.velocity * 0.5).SkipZ ());

            IterateArray (inRadius, i)
            {
               if (GetNearbyFriendsNearPosition (g_waypoint->GetPath (i)->origin, 256) != 0)
                  continue;

               m_throw = g_waypoint->GetPath (i)->origin;
               Vector src = CheckThrow (EyePosition (), m_throw);

               if (src.GetLengthSquared () < 100)
                  src = CheckToss (EyePosition (), m_throw);

               if (src == Vector::GetNull ())
                  continue;

               allowThrowing = true;
               break;
            }

            // start concussion grenade throwing?
            if (allowThrowing)
               m_states |= State_ThrowFlashbang;
            else
               m_states &= ~State_ThrowFlashbang;
         }

         if (m_states & State_ThrowExplosive)
            StartTask (Task_ThrowExpGrenade, TASKPRI_THROWGRENADE, -1, GetWorldTime () + 3.0, false);
         else if (m_states & State_ThrowFlashbang)
            StartTask (Task_ThrowFlashGrenade, TASKPRI_THROWGRENADE, -1, GetWorldTime () + 3.0, false);
         else if (m_states & State_ThrowSmoke)
            StartTask (Task_ThrowSmokeGrenade, TASKPRI_THROWGRENADE, -1, GetWorldTime () + 3.0, false);

         // delay next grenade throw
         if (m_states & (State_ThrowExplosive | State_ThrowFlashbang | State_ThrowSmoke))
            m_grenadeCheckTime = GetWorldTime () + Const_GrenadeTimer;
      }
   }
   else
      m_states &= ~(State_ThrowExplosive | State_ThrowFlashbang | State_ThrowSmoke);

   // check if there are items needing to be used/collected
   if (m_itemCheckTime < GetWorldTime () || !FNullEnt (m_pickupItem))
   {
      m_itemCheckTime = GetWorldTime () + g_botVar[Variable_TimerPickup]->GetFloat ();
      FindItem ();
   }

   float tempFear = m_fearLevel;
   float tempAgression = m_agressionLevel;

   // decrease fear if teammates near
   int friendlyNum = 0;

   if (m_lastEnemyOrigin != Vector::GetNull ())
      friendlyNum = GetNearbyFriendsNearPosition (pev->origin, 300) - GetNearbyEnemiesNearPosition (m_lastEnemyOrigin, 500);

   if (friendlyNum > 0)
      tempFear = tempFear * 0.5;

   // increase/decrease fear/aggression if bot uses a sniping weapon to be more careful
   if (UsesSniper ())
   {
      tempFear = tempFear * 1.5;
      tempAgression = tempAgression * 0.5;
   }

   // initialize & calculate the desire for all actions based on distances, emotions and other stuff
   GetCurrentTask ();

   // bot found some item to use?
   if (!FNullEnt (m_pickupItem))
   {
      m_states |= State_PickupItem;

      if (m_pickupType == PickupType_Button)
         g_taskFilters[Task_PickupItem].desire = 50; // always pickup button
      else
      {
         float distance = (500.0 - (GetEntityOrigin (m_pickupItem) - pev->origin).GetLength ()) * 0.2;

         if (distance > 50)
            distance = 50;

         g_taskFilters[Task_PickupItem].desire = distance;
      }
   }
   else
   {
      m_states &= ~State_PickupItem;
      g_taskFilters[Task_PickupItem].desire = 0.0;
   }

   float desireLevel = 0.0;

   // calculate desire to attack
   if ((m_states & State_SeeingEnemy) && ReactOnEnemy ())
      g_taskFilters[Task_AttackEnemy].desire = TASKPRI_ATTACK;
   else
      g_taskFilters[Task_AttackEnemy].desire = 0;

   // calculate desires to seek cover or hunt
   if (IsValidPlayer (m_lastEnemy) && m_lastEnemyOrigin != Vector::GetNull () && !((g_mapType & MapType_De) && g_bombPlanted) && !(pev->weapons & (1 << Weapon_BombC4)) && (m_loosedBombWptIndex == -1 && GetTeam (GetEntity ()) == Team_Terrorist))
   {
      float distance = (m_lastEnemyOrigin - pev->origin).GetLength ();

      // retreat level depends on bot health
      float retreatLevel = (100.0 - (pev->health > 100.0 ? 100.0 : pev->health)) * tempFear;
      float timeSeen = m_seeEnemyTime - GetWorldTime ();
      float timeHeard = m_heardSoundTime - GetWorldTime ();
      float ratio = 0.0;

      if (timeSeen > timeHeard)
      {
         timeSeen += 10.0;
         ratio = timeSeen * 0.1;
      }
      else
      {
         timeHeard += 10.0;
         ratio = timeHeard * 0.1;
      }

      if (g_bombPlanted || m_isStuck)
         ratio /= 3; // reduce the seek cover desire if bomb is planted
      else if (m_isVIP || m_isReloading)
         ratio *= 2; // triple the seek cover desire if bot is VIP or reloading

      if (distance > 500.0)
         g_taskFilters[Task_SeekCover].desire = retreatLevel * ratio;

      // if half of the round is over, allow hunting
      // FIXME: it probably should be also team/map dependant
      if (FNullEnt (m_enemy) && (g_timeRoundMid < GetWorldTime ()) && !m_isUsingGrenade && m_currentWaypointIndex != g_waypoint->FindNearest (m_lastEnemyOrigin) && m_personality != Personality_Careful)
      {
         desireLevel = 4096.0 - ((1.0 - tempAgression) * distance);
         desireLevel = (100 * desireLevel) / 4096.0;
         desireLevel -= retreatLevel;

         if (desireLevel > 89)
            desireLevel = 89;

         g_taskFilters[Task_HuntEnemy].desire = desireLevel;
      }
      else
         g_taskFilters[Task_HuntEnemy].desire = 0;
   }
   else
   {
      g_taskFilters[Task_SeekCover].desire = 0;
      g_taskFilters[Task_HuntEnemy].desire = 0;
   }

   // blinded behaviour
   if (m_blindTime > GetWorldTime ())
      g_taskFilters[Task_Blinded].desire = TASKPRI_BLINDED;
   else
      g_taskFilters[Task_Blinded].desire = 0.0;

   // now we've initialized all the desires go through the hard work
   // of filtering all actions against each other to pick the most
   // rewarding one to the bot.

   // FIXME: instead of going through all of the actions it might be
   // better to use some kind of decision tree to sort out impossible
   // actions.

   // most of the values were found out by trial-and-error and a helper
   // utility i wrote so there could still be some weird behaviors, it's
   // hard to check them all out.

   m_oldCombatDesire = HysteresisDesire (g_taskFilters[Task_AttackEnemy].desire, 40.0, 90.0, m_oldCombatDesire);
   g_taskFilters[Task_AttackEnemy].desire = m_oldCombatDesire;

   Task_t *taskOffensive = &g_taskFilters[Task_AttackEnemy];
   Task_t *taskPickup = &g_taskFilters[Task_PickupItem];

   // calc survive (cover/hide)
   Task_t *taskSurvive = ThresholdDesire (&g_taskFilters[Task_SeekCover], 40.0, 0.0);
   taskSurvive = SubsumeDesire (&g_taskFilters[Task_Hide], taskSurvive);

   Task_t *def = ThresholdDesire (&g_taskFilters[Task_HuntEnemy], 60.0, 0.0); // don't allow hunting if desire's 60<
   taskOffensive = SubsumeDesire (taskOffensive, taskPickup); // if offensive task, don't allow picking up stuff

   Task_t *taskSub = MaxDesire (taskOffensive, def); // default normal & careful tasks against offensive actions
   Task_t *final = SubsumeDesire (&g_taskFilters[Task_Blinded], MaxDesire (taskSurvive, taskSub)); // reason about fleeing instead

   if (m_tasks != NULL)
      final = MaxDesire (final, m_tasks);

   StartTask (final); // push the final behaviour in our task stack to carry out
}

void Bot::ResetTasks (void)
{
   // this function resets bot tasks stack, by removing all entries from the stack.

   if (m_tasks == NULL)
      return; // reliability check

   Task_t *next = m_tasks->next;
   Task_t *prev = m_tasks;

   while (m_tasks != NULL)
   {
      prev = m_tasks->prev;

      delete m_tasks;
      m_tasks = prev;
   }
   m_tasks = next;

   while (m_tasks != NULL)
   {
      next = m_tasks->next;

      delete m_tasks;
      m_tasks = next;
   }
   m_tasks = NULL;
}

void Bot::CheckTasksPriorities (void)
{
   // this function checks the tasks priorities.

   if (m_tasks == NULL)
   {
      GetCurrentTask ();
      return;
   }

   Task_t *oldTask = m_tasks;
   Task_t *prev = NULL;
   Task_t *next = NULL;

   while (m_tasks->prev != NULL)
   {
      prev = m_tasks->prev;
      m_tasks = prev;
   }

   Task_t *firstTask = m_tasks;
   Task_t *maxDesiredTask = m_tasks;

   float maxDesire = m_tasks->desire;

   // search for the most desired task
   while (m_tasks->next != NULL)
   {
      next = m_tasks->next;
      m_tasks = next;

      if (m_tasks->desire >= maxDesire)
      {
         maxDesiredTask = m_tasks;
         maxDesire = m_tasks->desire;
      }
   }
   m_tasks = maxDesiredTask; // now we found the most desired pushed task...

   // something was changed with priorities, check if some task doesn't need to be deleted...
   if (oldTask != maxDesiredTask)
   {
      m_tasks = firstTask;

      while (m_tasks != NULL)
      {
         next = m_tasks->next;

         // some task has to be deleted if cannot be continued...
         if (m_tasks != maxDesiredTask && !m_tasks->canContinue)
         {
            if (m_tasks->prev != NULL)
               m_tasks->prev->next = m_tasks->next;

            if (m_tasks->next != NULL)
               m_tasks->next->prev = m_tasks->prev;

            delete m_tasks;
         }
         m_tasks = next;
      }
   }
   m_tasks = maxDesiredTask;

   if (g_botVar[Variable_DebugGoal]->GetInt () != -1)
      m_chosenGoalIndex = g_botVar[Variable_DebugGoal]->GetInt ();
   else
      m_chosenGoalIndex = m_tasks->data;
}

void Bot::StartTask (TaskId_t taskID, float desire, int data, float time, bool canContinue)
{
   // build task
   Task_t task = {NULL, NULL, taskID, desire, data, time, canContinue};
   StartTask (&task); // use standard function to start task
}

void Bot::StartTask (Task_t *task)
{
   // this function adds task pointer on the bot task stack.

   bool newTaskDifferent = false;
   bool foundTaskExisting = false;
   bool checkPriorities = false;

   Task_t *oldTask = GetCurrentTask (); // remember our current task
   Task_t *firstTask = GetCurrentTask (); // first task on the stack

   // at the beginning need to clean up all NULL tasks...
   if (m_tasks == NULL)
   {
      m_lastCollTime = GetWorldTime () + 0.5;

      Task_t *newTask = AllocateMembuf <Task_t> ();

      newTask->taskID = Task_Normal;
      newTask->desire = TaskPriority_Normal;
      newTask->canContinue = true;
      newTask->time = 0.0;
      newTask->data = -1;
      newTask->next = newTask->prev = NULL;

      m_tasks = newTask;
      DeleteSearchNodes ();

      if (task == NULL)
         return;
      else if (task->taskID == Task_Normal)
      {
         m_tasks->desire = TaskPriority_Normal;
         m_tasks->data = task->data;
         m_tasks->time = task->time;
         
         return;
      }
   }
   else if (task == NULL)
      return;

   // it shouldn't happen this condition now as false...
   if (m_tasks != NULL)
   {
      if (m_tasks->taskID == task->taskID)
      {
         if (m_tasks->data != task->data)
         {
            m_lastCollTime = GetWorldTime () + 0.5;

            DeleteSearchNodes ();
            m_tasks->data = task->data;
         }

         if (m_tasks->desire != task->desire)
         {
            m_tasks->desire = task->desire;
            checkPriorities = true;
         }
         else if (m_tasks->data == task->data)
            return;
      }
      else
      {
         // find the first task on the stack and don't allow push the new one like the same already existing one
         while (m_tasks->prev != NULL)
         {
            m_tasks = m_tasks->prev;
            firstTask = m_tasks;

            if (m_tasks->taskID == task->taskID)
            {
               foundTaskExisting = true;

               if (m_tasks->desire != task->desire)
                  checkPriorities = true;

               m_tasks->desire = task->desire;
               m_tasks->data = task->data;
               m_tasks->time = task->time;
               m_tasks->canContinue = task->canContinue;
               m_tasks = oldTask;

               break; // now we may need to check the current max desire or next tasks...
            }
         }

         // now go back to the previous stack position and try to find the same task as one of "the next" ones (already pushed before and not finished yet)
         if (!foundTaskExisting && !checkPriorities)
         {
            m_tasks = oldTask;

            while (m_tasks->next != NULL)
            {
               m_tasks = m_tasks->next;

               if (m_tasks->taskID == task->taskID)
               {
                  foundTaskExisting = true;

                  if (m_tasks->desire != task->desire)
                     checkPriorities = true;

                  m_tasks->desire = task->desire;
                  m_tasks->data = task->data;
                  m_tasks->time = task->time;
                  m_tasks->canContinue = task->canContinue;
                  m_tasks = oldTask;

                  break; // now we may need to check the current max desire...
               }
            }
         }

         if (!foundTaskExisting)
            newTaskDifferent = true; // we have some new task pushed on the stack...
      }
   }
   m_tasks = oldTask;

   if (newTaskDifferent)
   {
      Task_t *newTask = AllocateMembuf <Task_t> ();

      newTask->taskID = task->taskID;
      newTask->desire = task->desire;
      newTask->canContinue = task->canContinue;
      newTask->time = task->time;
      newTask->data = task->data;
      newTask->next = NULL;

      while (m_tasks->next != NULL)
         m_tasks = m_tasks->next;

      newTask->prev = m_tasks;
      m_tasks->next = newTask;

      checkPriorities = true;
   }
   m_tasks = oldTask;

   // needs check the priorities and setup the task with the max desire...
   if (!checkPriorities)
      return;
 
   CheckTasksPriorities ();

   // the max desired task has been changed...
   if (m_tasks != oldTask)
   {
      DeleteSearchNodes ();
      m_lastCollTime = GetWorldTime () + 0.5;

      // leader bot?
      if (newTaskDifferent && m_isLeader && m_tasks->taskID == Task_SeekCover)
         CommandTeam (); // reorganize team if fleeing

      if (newTaskDifferent && m_tasks->taskID == Task_Camp)
         SelectBestWeapon ();

      // this is best place to handle some voice commands report team some info
      if (newTaskDifferent && g_randGen.Long (0, 100) < 95)
      {
         switch (m_tasks->taskID)
         {
         case Task_Blinded:
            ChatterMessage (Chatter_GotBlinded);
            break;

         case Task_PlantBomb:
            ChatterMessage (Chatter_PlantingC4);
            break;
         }
      }

      if (newTaskDifferent && g_randGen.Long (0, 100) < 80 && m_tasks->taskID == Task_Camp)
      {
         if ((g_mapType & MapType_De) && g_bombPlanted)
            ChatterMessage (Chatter_GuardDroppedC4);
         else
            ChatterMessage (Chatter_GoingToCamp);
      }

      if (newTaskDifferent && g_randGen.Long (0, 100) < 80 && m_tasks->taskID == Task_Camp && GetTeam (GetEntity ()) == Team_Terrorist && m_inVIPZone)
         ChatterMessage (Chatter_GoingToGuardVIPSafety);
   }
}

Task_t *Bot::GetCurrentTask (void)
{
   if (m_tasks != NULL)
      return m_tasks;

   Task_t *newTask = AllocateMembuf <Task_t> ();

   newTask->taskID = Task_Normal;
   newTask->desire = TaskPriority_Normal;
   newTask->data = -1;
   newTask->time = 0.0;
   newTask->canContinue = true;
   newTask->next = newTask->prev = NULL;

   m_tasks = newTask;
   m_lastCollTime = GetWorldTime () + 0.5;

   DeleteSearchNodes ();
   return m_tasks;
}

void Bot::RemoveCertainTask (TaskId_t taskID)
{
   // this function removes one task from the bot task stack.

   if ((m_tasks == NULL) || ((m_tasks != NULL) && (m_tasks->taskID == Task_Normal)))
      return; // since normal task can be only once on the stack, don't remove it...

   bool checkPriorities = false;

   Task_t *task = m_tasks;
   Task_t *oldTask = m_tasks;
   Task_t *oldPrevTask = task->prev;
   Task_t *oldNextTask = task->next;

   while (task->prev != NULL)
      task = task->prev;

   while (task != NULL)
   {
      Task_t *next = task->next;
      Task_t *prev = task->prev;

      if (task->taskID == taskID)
      {
         if (prev != NULL)
            prev->next = next;

         if (next != NULL)
            next->prev = prev;

         if (task == oldTask)
            oldTask = NULL;
         else if (task == oldPrevTask)
            oldPrevTask = NULL;
         else if (task == oldNextTask)
            oldNextTask = NULL;

         delete task;

         checkPriorities = true;
         break;
      }
      task = next;
   }

   if (oldTask != NULL)
      m_tasks = oldTask;
   else if (oldPrevTask != NULL)
      m_tasks = oldPrevTask;
   else if (oldNextTask != NULL)
      m_tasks = oldNextTask;
   else
      GetCurrentTask ();

   if (checkPriorities)
      CheckTasksPriorities ();
}

void Bot::TaskComplete (void)
{
   // this function called whenever a task is completed.

   if (m_tasks == NULL)
   {
      DeleteSearchNodes (); // delete all path finding nodes
      return;
   }

   if (m_tasks->taskID == Task_Normal)
   {
      DeleteSearchNodes (); // delete all path finding nodes
      
      m_tasks->data = -1;
      m_chosenGoalIndex = -1;

      return;
   }

   Task_t *next = m_tasks->next;
   Task_t *prev = m_tasks->prev;

   if (next != NULL)
      next->prev = prev;
   
   if (prev != NULL)
      prev->next = next;

   delete m_tasks;
   m_tasks = NULL;

   if (prev != NULL && next != NULL)
   {
      if (prev->desire >= next->desire)
         m_tasks = prev;
      else
         m_tasks = next;
   }
   else if (prev != NULL)
      m_tasks = prev;
   else if (next != NULL)
      m_tasks = next;

   if (m_tasks == NULL)
      GetCurrentTask ();

   CheckTasksPriorities ();
   DeleteSearchNodes ();
}

bool Bot::EnemyIsThreat (void)
{
   if (FNullEnt (m_enemy) || GetCurrentTask ()->taskID == Task_SeekCover)
      return false;

   float distance = (m_enemy->v.origin - pev->origin).GetLength ();

   // if bot is camping, he should be firing anyway and not leaving his position
   if (GetCurrentTask ()->taskID == Task_Camp)
      return false;

   // if enemy is near or facing us directly
   if (distance < 256 || IsInViewCone (m_enemy->v.origin))
      return true;

   return false;
}

bool Bot::ReactOnEnemy (void)
{
   // the purpose of this function is check if task has to be interrupted because an enemy is near (run attack actions then)

   if (!EnemyIsThreat ())
      return false;

   if (m_enemyReachableTimer < GetWorldTime ())
   {
      int i = g_waypoint->FindNearest (pev->origin);
      int enemyIndex = g_waypoint->FindNearest (m_enemy->v.origin);

      float lineDist = (m_enemy->v.origin - pev->origin).GetLength ();
      float pathDist = g_waypoint->GetPathDistance (i, enemyIndex);

      if (pathDist - lineDist > 112.0)
         m_isEnemyReachable = false;
      else
         m_isEnemyReachable = true;

      m_enemyReachableTimer = GetWorldTime () + 1.0;
   }

   if (m_isEnemyReachable)
   {
      m_navTimeset = GetWorldTime (); // override existing movement by attack movement
      return true;
   }
   return false;
}

bool Bot::IsLastEnemyViewable (void)
{
   // this function checks if line of sight established to last enemy

   if (FNullEnt (m_lastEnemy) || m_lastEnemyOrigin == Vector::GetNull ())
   {
      m_lastEnemy = NULL;
      m_lastEnemyOrigin = Vector::GetNull ();

      return false;
   }

   // trace a line from bot's eyes to destination...
   TraceResult tr;
   TraceLine (EyePosition (), m_lastEnemyOrigin, true, GetEntity (), &tr);

   // check if line of sight to object is not blocked (i.e. visible)
   return tr.flFraction >= 1.0;
}

bool Bot::LastEnemyShootable (void)
{
   // don't allow shooting through walls when pausing or camping
   if (!(m_aimFlags & AimPos_LastEnemy) || FNullEnt (m_lastEnemy) || GetCurrentTask ()->taskID == Task_Pause || GetCurrentTask ()->taskID == Task_Camp)
      return false;

   return GetShootingConeDeviation (GetEntity (), &m_lastEnemyOrigin) >= 0.90;
}

void Bot::CheckRadioCommands (void)
{
   // this function handling radio and reactings to it

   float distance = (m_radioEntity->v.origin - pev->origin).GetLength ();

   // don't allow bot listen you if bot is busy
   if (GetCurrentTask ()->taskID == Task_DefuseBomb || GetCurrentTask ()->taskID == Task_PlantBomb || HasHostage () && m_radioOrder != Radio_ReportTeam)
   {
      m_radioOrder = 0;
      return;
   }

   switch (m_radioOrder)
   {
   case Radio_CoverMe:
   case Radio_FollowMe:
   case Chatter_GoingToPlantBomb:
      // check if line of sight to object is not blocked (i.e. visible)
      if (EntityIsVisible (m_radioEntity->v.origin))
      {
         if (FNullEnt (m_targetEntity) && FNullEnt (m_enemy) && g_randGen.Long (0, 100) < (m_personality == Personality_Careful ? 80 : 50))
         {
            int numFollowers = 0;

            // Check if no more followers are allowed
            for (int i = 0; i < GetMaxClients (); i++)
            {
               if (g_botManager->GetBot (i))
               {
                  if (IsAlive (g_botManager->GetBot (i)->GetEntity ()))
                  {
                     if (g_botManager->GetBot (i)->m_targetEntity == m_radioEntity)
                        numFollowers++;
                  }
               }
            }

            int allowedFollowers = g_botVar[Variable_FollowUser]->GetInt ();

            if (m_radioEntity->v.weapons & (1 << Weapon_BombC4))
               allowedFollowers = 1;

            if (numFollowers < allowedFollowers)
            {
               RadioMessage (Radio_Affirmative);
               m_targetEntity = m_radioEntity;

               // don't pause/camp/follow anymore
               TaskId_t taskID = GetCurrentTask ()->taskID;

               if (taskID == Task_Pause || taskID == Task_Camp)
                  m_tasks->time = GetWorldTime ();

               StartTask (Task_FollowUser, TASKPRI_FOLLOWUSER, -1, 0.0, true);
            }
            else if (numFollowers > allowedFollowers)
            {
               for (int i = 0; (i < GetMaxClients () && numFollowers > allowedFollowers) ; i++)
               {
                  if (g_botManager->GetBot (i))
                  {
                     if (IsAlive (g_botManager->GetBot (i)->GetEntity ()))
                     {
                        if (g_botManager->GetBot (i)->m_targetEntity == m_radioEntity)
                        {
                           g_botManager->GetBot (i)->m_targetEntity = NULL;
                           numFollowers--;
                        }
                     }
                  }
               }
            }
            else if (m_radioOrder != Chatter_GoingToPlantBomb)
               RadioMessage (Radio_Negative);
         }
         else if (m_radioOrder != Chatter_GoingToPlantBomb)
            RadioMessage (Radio_Negative);
      }
      break;

   case Radio_HoldPosition:
      if (!FNullEnt (m_targetEntity))
      {
         if (m_targetEntity == m_radioEntity)
         {
            m_targetEntity = NULL;
            RadioMessage (Radio_Affirmative);

            m_campButtons = 0;

            StartTask (Task_Pause, TASKPRI_PAUSE, -1, GetWorldTime () + g_randGen.Float (30.0, 60.0), false);
         }
      }
      break;

   case Radio_TakingFire:
      if (FNullEnt (m_targetEntity))
      {
         if (FNullEnt (m_enemy))
         {
            // Decrease Fear Levels to lower probability of Bot seeking Cover again
            m_fearLevel -= 0.2;

            if (m_fearLevel < 0.0)
               m_fearLevel = 0.0;

            if (g_randGen.Long (0, 100) < 45 && g_botVar[Variable_Communication]->GetInt () == 2)
               ChatterMessage (Chatter_OnMyWay);
            else if (m_radioOrder == Radio_NeedBackup && g_botVar[Variable_Communication]->GetInt () != 2)
               RadioMessage (Radio_Affirmative);

            // don't pause/camp anymore
            TaskId_t taskID = GetCurrentTask ()->taskID;

            if (taskID == Task_Pause || taskID == Task_Camp)
               m_tasks->time = GetWorldTime ();

            m_position = m_radioEntity->v.origin;
            DeleteSearchNodes ();

            StartTask (Task_MoveToPosition, TASKPRI_MOVETOPOSITION, -1, 0.0, true);
         }
         else if (g_randGen.Long (0, 100) < 80 && m_radioOrder == Radio_TakingFire)
            RadioMessage (Radio_Negative);
      }
      break;

   case Radio_YouTakePoint:
      if (EntityIsVisible (m_radioEntity->v.origin) && m_isLeader)
         RadioMessage (Radio_Affirmative);
      break;

   case Radio_NeedBackup:
   case Chatter_ScaredEmotion:
      if ((FNullEnt (m_enemy) && EntityIsVisible (m_radioEntity->v.origin)) || distance < 2048 || !m_moveToC4)
      {
         m_fearLevel -= 0.1;

         if (m_fearLevel < 0.0)
            m_fearLevel = 0.0;

         if (g_randGen.Long (0, 100) < 45 && g_botVar[Variable_Communication]->GetInt () == 2)
            ChatterMessage (Chatter_OnMyWay);
         else if (m_radioOrder == Radio_NeedBackup && g_botVar[Variable_Communication]->GetInt () != 2)
            RadioMessage (Radio_Affirmative);

         // don't pause/camp anymore
         TaskId_t taskID = GetCurrentTask ()->taskID;

         if (taskID == Task_Pause || taskID == Task_Camp)
            m_tasks->time = GetWorldTime ();

         m_moveToC4 = true;
         m_position = m_radioEntity->v.origin;

         DeleteSearchNodes ();

         // start move to position task
         StartTask (Task_MoveToPosition, TASKPRI_MOVETOPOSITION, -1, 0.0, true);
      }
      else if (g_randGen.Long (0, 100) < 80 && m_radioOrder == Radio_NeedBackup)
         RadioMessage (Radio_Negative);
      break;

   case Radio_GoGoGo:
      if (m_radioEntity == m_targetEntity)
      {
         RadioMessage (Radio_Affirmative);

         m_targetEntity = NULL;
         m_fearLevel -= 0.3;

         if (m_fearLevel < 0.0)
            m_fearLevel = 0.0;
      }
      else if ((FNullEnt (m_enemy) && EntityIsVisible (m_radioEntity->v.origin)) || distance < 2048)
      {
         TaskId_t taskID = GetCurrentTask ()->taskID;

         if (taskID == Task_Pause || taskID == Task_Camp)
         {
            m_fearLevel -= 0.3;

            if (m_fearLevel < 0.0)
               m_fearLevel = 0.0;

            RadioMessage (Radio_Affirmative);
            // don't pause/camp anymore
            m_tasks->time = GetWorldTime ();

            m_targetEntity = NULL;
            MakeVectors (m_radioEntity->v.v_angle);

            m_position = m_radioEntity->v.origin + g_pGlobals->v_forward * g_randGen.Long (1024, 2048);

            DeleteSearchNodes ();
            StartTask (Task_MoveToPosition, TASKPRI_MOVETOPOSITION, -1, 0.0, true);
         }
      }
      else if (!FNullEnt (m_doubleJumpEntity))
      {
         RadioMessage (Radio_Affirmative);
         ResetDoubleJumpState ();
      }
      else
         RadioMessage (Radio_Negative);
      break;

   case Radio_ShesGonnaBlow:
      if (FNullEnt (m_enemy) && distance < 2048 && g_bombPlanted && GetTeam (GetEntity ()) == Team_Terrorist)
      {
         RadioMessage (Radio_Affirmative);

         if (GetCurrentTask ()->taskID == Task_Camp)
            RemoveCertainTask (Task_Camp);

         m_targetEntity = NULL;
         StartTask (Task_EscapeFromBomb, TASKPRI_ESCAPEFROMBOMB, -1, 0.0, true);
      }
      else
        RadioMessage (Radio_Negative);

      break;

   case Radio_RegroupTeam:
      // if no more enemies found AND bomb planted, switch to knife to get to bombplace faster
      if ((GetTeam (GetEntity ()) == Team_Counter) && m_currentWeapon != Weapon_Knife && GetNearbyEnemiesNearPosition (pev->origin, 9999) == 0 && g_bombPlanted && GetCurrentTask ()->taskID != Task_DefuseBomb)
      {
         SelectWeaponByName ("weapon_knife");

         DeleteSearchNodes ();
         MoveToVector (g_waypoint->GetBombPosition ());

         RadioMessage (Radio_Affirmative);
      }
      break;

   case Radio_StormTheFront:
      if ((FNullEnt (m_enemy) && EntityIsVisible (m_radioEntity->v.origin)) || distance < 1024)
      {
         RadioMessage (Radio_Affirmative);

         // don't pause/camp anymore
         TaskId_t taskID = GetCurrentTask ()->taskID;

         if (taskID == Task_Pause || taskID == Task_Camp)
            m_tasks->time = GetWorldTime ();

         m_targetEntity = NULL;

         MakeVectors (m_radioEntity->v.v_angle);
         m_position = m_radioEntity->v.origin + g_pGlobals->v_forward * g_randGen.Long (1024, 2048);

         DeleteSearchNodes ();
         StartTask (Task_MoveToPosition, TASKPRI_MOVETOPOSITION, -1, 0.0, true);

         m_fearLevel -= 0.3;

         if (m_fearLevel < 0.0)
            m_fearLevel = 0.0;

         m_agressionLevel += 0.3;

         if (m_agressionLevel > 1.0)
            m_agressionLevel = 1.0;
      }
      break;

   case Radio_Fallback:
      if ((FNullEnt (m_enemy) && EntityIsVisible (m_radioEntity->v.origin)) || distance < 1024)
      {
         m_fearLevel += 0.5;

         if (m_fearLevel > 1.0)
            m_fearLevel = 1.0;

         m_agressionLevel -= 0.5;

         if (m_agressionLevel < 0.0)
            m_agressionLevel = 0.0;

         if (GetCurrentTask ()->taskID == Task_Camp)
            m_tasks->time += g_randGen.Float (10.0, 15.0);
         else
         {
            // don't pause/camp anymore
            TaskId_t taskID = GetCurrentTask ()->taskID;

            if (taskID == Task_Pause)
               m_tasks->time = GetWorldTime ();

            m_targetEntity = NULL;
            m_seeEnemyTime = GetWorldTime ();

            // if bot has no enemy
            if (m_lastEnemyOrigin == Vector::GetNull ())
            {
               int team = GetTeam (GetEntity ());
               float nearestDistance = FLT_MAX;

               // take nearest enemy to ordering player
               for (int i = 0; i < GetMaxClients (); i++)
               {
                  if (!(g_clients[i].flags & ClientFlag_Used) || !(g_clients[i].flags & ClientFlag_Alive) || g_clients[i].team == team)
                     continue;

                  edict_t *enemy = g_clients[i].ent;
                  float distance = (m_radioEntity->v.origin - enemy->v.origin).GetLengthSquared ();

                  if (distance < nearestDistance)
                  {
                     nearestDistance = distance;
                     m_lastEnemy = enemy;
                     m_lastEnemyOrigin = enemy->v.origin;
                  }
               }
            }
            DeleteSearchNodes ();
         }
      }
      break;

   case Radio_ReportTeam:
      if (g_randGen.Long  (0, 100) < 85)
         RadioMessage ((GetNearbyEnemiesNearPosition (pev->origin, 400) == 0 && g_botVar[Variable_Communication]->GetInt () != 2) ? Radio_SectorClear : Radio_ReportingIn);
      break;

   case Radio_SectorClear:
      // is bomb planted and it's a ct
      if (g_bombPlanted)
      {
         int bombPoint = -1;

         // check if it's a ct command
         if (GetTeam (m_radioEntity) == Team_Counter && GetTeam (GetEntity ()) == Team_Counter && IsValidBot (m_radioEntity))
         {
            if (g_timeNextBombUpdate < GetWorldTime ())
            {
               float minDistance = FLT_MAX;

               // find nearest bomb waypoint to player
               IterateArray (g_waypoint->m_goalPoints, i)
               {
                  distance = (g_waypoint->GetPath (g_waypoint->m_goalPoints[i])->origin - m_radioEntity->v.origin).GetLengthSquared ();

                  if (distance < minDistance)
                  {
                     minDistance = distance;
                     bombPoint = g_waypoint->m_goalPoints[i];
                  }
               }

               // mark this waypoint as restricted point
               if (bombPoint != -1 && !g_waypoint->IsGoalVisited (bombPoint))
                  g_waypoint->SetGoalVisited (bombPoint);

               g_timeNextBombUpdate = GetWorldTime () + 0.5;
            }
            // Does this Bot want to defuse?
            if (GetCurrentTask ()->taskID == Task_Normal)
            {
               // Is he approaching this goal?
               if (m_tasks->data == bombPoint)
               {
                  m_tasks->data = -1;
                  RadioMessage (Radio_Affirmative);
               }
            }
         }
      }
      break;

   case Radio_GetInPosition:
      if ((FNullEnt (m_enemy) && EntityIsVisible (m_radioEntity->v.origin)) || distance < 1024)
      {
         RadioMessage (Radio_Affirmative);

         if (GetCurrentTask ()->taskID == Task_Camp)
            m_tasks->time = GetWorldTime () + g_randGen.Float (30.0, 60.0);
         else
         {
            // don't pause anymore
            TaskId_t taskID = GetCurrentTask ()->taskID;

            if (taskID == Task_Pause)
               m_tasks->time = GetWorldTime ();

            m_targetEntity = NULL;
            m_seeEnemyTime = GetWorldTime ();

            // If Bot has no enemy
            if (m_lastEnemyOrigin == Vector::GetNull ())
            {
               int team = GetTeam (GetEntity ());
               float nearestDistance = FLT_MAX;

               // Take nearest enemy to ordering Player
               for (int i = 0; i < GetMaxClients (); i++)
               {
                  if (!(g_clients[i].flags & ClientFlag_Used) || !(g_clients[i].flags & ClientFlag_Alive) || g_clients[i].team == team)
                     continue;

                  edict_t *enemy = g_clients[i].ent;
                  float distance = (m_radioEntity->v.origin - enemy->v.origin).GetLengthSquared ();

                  if (distance < nearestDistance)
                  {
                     nearestDistance = distance;
                     m_lastEnemy = enemy;
                     m_lastEnemyOrigin = enemy->v.origin;
                  }
               }
            }
            DeleteSearchNodes ();

            int index = FindDefendWaypoint (m_radioEntity->v.origin);

            // push camp task on to stack
            StartTask (Task_Camp, TASKPRI_CAMP, -1, GetWorldTime () + g_randGen.Float (30.0, 60.0), true);
            // push move command
            StartTask (Task_MoveToPosition, TASKPRI_MOVETOPOSITION, index, GetWorldTime () + g_randGen.Float (30.0, 60.0), true);

            if (g_waypoint->GetPath (index)->vis.crouch <= g_waypoint->GetPath (index)->vis.stand)
               m_campButtons |= IN_DUCK;
            else
               m_campButtons &= ~IN_DUCK;
         }
      }
      break;
   }
   m_radioOrder = 0; // radio command has been handled, reset
}

void Bot::SelectLeaderEachTeam (int team)
{
   Bot *botLeader = NULL;

   if (g_mapType & MapType_As)
   {
      if (m_isVIP && !g_leaderChoosen[Team_Counter])
      {
         // vip bot is the leader
         m_isLeader = true;

         if (g_randGen.Long (1, 100) < 50)
         {
            RadioMessage (Radio_FollowMe);
            m_campButtons = 0;
         }
         g_leaderChoosen[Team_Counter] = true;
      }
      else if ((team == Team_Terrorist) && !g_leaderChoosen[Team_Terrorist])
      {
         botLeader = g_botManager->GetHighestFragsBot (team);

         if (botLeader != NULL)
         {
            botLeader->m_isLeader = true;

            if (g_randGen.Long (1, 100) < 45)
               botLeader->RadioMessage (Radio_FollowMe);
         }
         g_leaderChoosen[Team_Terrorist] = true;
      }
   }
   else if (g_mapType & MapType_De)
   {
      if (team == Team_Terrorist && !g_leaderChoosen[Team_Terrorist])
      {
         if (pev->weapons & (1 << Weapon_BombC4))
         {
            // bot carrying the bomb is the leader
            m_isLeader = true;

            // terrorist carrying a bomb needs to have some company
            if (g_randGen.Long (1, 100) < 80)
            {
               if (g_botVar[Variable_Communication]->GetInt () == 2)
                  ChatterMessage (Chatter_GoingToPlantBomb);
               else
                  ChatterMessage (Radio_FollowMe);

               m_campButtons = 0;
            }
            g_leaderChoosen[Team_Terrorist] = true;
         }
      }
      else if (!g_leaderChoosen[Team_Counter])
      {
         botLeader = g_botManager->GetHighestFragsBot (team);

         if (botLeader != NULL)
         {
            botLeader->m_isLeader = true;

            if (g_randGen.Long (1, 100) < 30)
               botLeader->RadioMessage (Radio_FollowMe);
         }
         g_leaderChoosen[Team_Counter] = true;
      }
   }
   else if (g_mapType & (MapType_Es | MapType_Ka | MapType_Fy))
   {
      if (team == Team_Terrorist)
      {
         botLeader = g_botManager->GetHighestFragsBot (team);

         if (botLeader != NULL)
         {
            botLeader->m_isLeader = true;

            if (g_randGen.Long (1, 100) < 30)
               botLeader->RadioMessage (Radio_FollowMe);
         }
      }
      else
      {
         botLeader = g_botManager->GetHighestFragsBot (team);

         if (botLeader != NULL)
         {
            botLeader->m_isLeader = true;

            if (g_randGen.Long (1, 100) < 30)
               botLeader->RadioMessage (Radio_FollowMe);
         }
      }
   }
   else
   {
      if (team == Team_Terrorist)
      {
         botLeader = g_botManager->GetHighestFragsBot (team);

         if (botLeader != NULL)
         {
            botLeader->m_isLeader = true;

            if (g_randGen.Long (1, 100) < 30)
               botLeader->RadioMessage (Radio_FollowMe);
         }
      }
      else
      {
         botLeader = g_botManager->GetHighestFragsBot (team);

         if (botLeader != NULL)
         {
            botLeader->m_isLeader = true;

            if (g_randGen.Long (1, 100) < 40)
               botLeader->RadioMessage (Radio_FollowMe);
         }
      }
   }
}

void Bot::ChooseAimDirection (void)
{
   if (!m_canChooseAimDirection)
      return;

   TraceResult tr;
   memset (&tr, 0, sizeof (TraceResult));

   unsigned int flags = m_aimFlags;

   bool canChooseAimDirection = false;
   bool currentPointValid = (m_currentWaypointIndex >= 0 && m_currentWaypointIndex < g_numWaypoints);

   if (!currentPointValid)
   {
      currentPointValid = true;
      GetValidWaypoint ();
   }

   // check if last enemy vector valid
   if (m_lastEnemyOrigin != Vector::GetNull ())
   {
      TraceLine (EyePosition (), m_lastEnemyOrigin, false, true, GetEntity (), &tr);

      if ((pev->origin - m_lastEnemyOrigin).GetLength () >= 1600 && FNullEnt (m_enemy) && !UsesSniper () || (tr.flFraction <= 0.2 && tr.pHit == g_hostEntity) && m_seeEnemyTime + 7.0 < GetWorldTime ())
      {
         if ((m_aimFlags & (AimPos_LastEnemy | AimPos_PredictEnemy)) && m_wantsToFire)
            m_wantsToFire = false;

         m_lastEnemyOrigin = Vector::GetNull ();
         m_aimFlags &= ~(AimPos_LastEnemy | AimPos_PredictEnemy);

         flags &= ~(AimPos_LastEnemy | AimPos_PredictEnemy);
      }
   }
   else
   {
      m_aimFlags &= ~(AimPos_LastEnemy | AimPos_PredictEnemy);
      flags &= ~(AimPos_LastEnemy | AimPos_PredictEnemy);
   }

   // don't allow bot to look at danger positions under certain circumstances
   if (!(flags & (AimPos_Grenade | AimPos_Enemy | AimPos_Entity)))
   {
      if (IsOnLadder () || IsInWater () || (m_waypointFlags & PointFlag_Ladder) || (m_currentTravelFlags & PathFlag_Jump))
      {
         flags &= ~(AimPos_LastEnemy | AimPos_PredictEnemy);
         canChooseAimDirection = false;
      }
   }

   if (flags & AimPos_Override)
      m_lookAt = m_camp;
   else if (flags & AimPos_Grenade)
      m_lookAt = m_throw + Vector (0, 0, 1.0 * m_grenade.z);
   else if (flags & AimPos_Enemy)
      FocusEnemy ();
   else if (flags & AimPos_Entity)
      m_lookAt = m_entity;
   else if (flags & AimPos_LastEnemy)
   {
      m_lookAt = m_lastEnemyOrigin;

      // did bot just see enemy and is quite aggressive?
      if (m_seeEnemyTime + 3.0 - m_actualReactionTime + m_baseAgressionLevel > GetWorldTime ())
      {
         // feel free to fire if shootable
         if (!UsesSniper () && LastEnemyShootable ())
            m_wantsToFire = true;
      }
      else // forget an enemy far away
      {
         m_aimFlags &= ~AimPos_LastEnemy;

         if ((pev->origin - m_lastEnemyOrigin).GetLength () >= 1600.0)
            m_lastEnemyOrigin = Vector::GetNull ();
      }
   }
   else if (flags & AimPos_PredictEnemy)
   {
      if (((pev->origin - m_lastEnemyOrigin).GetLength () < 1600 || UsesSniper ()) && (tr.flFraction >= 0.2 || tr.pHit != g_worldEdict))
      {
         bool recalcPath = true;

         if (!FNullEnt (m_lastEnemy) && m_trackingEdict == m_lastEnemy && m_timeNextTracking < GetWorldTime ())
            recalcPath = false;

         if (recalcPath)
         {
            m_lookAt = g_waypoint->GetPath (GetAimingWaypoint (m_lastEnemyOrigin))->origin;
            m_camp = m_lookAt;

            m_timeNextTracking = GetWorldTime () + 0.5;
            m_trackingEdict = m_lastEnemy;

            // feel free to fire if shoot able
            if (LastEnemyShootable ())
               m_wantsToFire = true;
         }
         else
            m_lookAt = m_camp;
      }
      else // forget an enemy far away
      {
         m_aimFlags &= ~AimPos_PredictEnemy;

         if ((pev->origin - m_lastEnemyOrigin).GetLength () >= 1600.0)
            m_lastEnemyOrigin = Vector::GetNull ();
      }
   }
   else if (flags & AimPos_CampPoint)
      m_lookAt = m_camp;
   else if ((flags & AimPos_NavPoint) && !(flags & (AimPos_Entity | AimPos_Enemy)))
   {
      m_lookAt = m_destOrigin;

      if (canChooseAimDirection && m_currentWaypointIndex != -1 && !(g_waypoint->GetPath (m_currentWaypointIndex)->flags & PointFlag_Ladder))
      {
         TraceResult tr;
         int index = m_currentWaypointIndex;

         if (GetTeam (GetEntity ()) == Team_Terrorist)
         {
            if ((g_experienceData + (index * g_numWaypoints) + index)->team0DangerIndex != -1)
            {
               Vector dest = g_waypoint->GetPath ((g_experienceData + (index * g_numWaypoints) + index)->team0DangerIndex)->origin;
               TraceLine (pev->origin, dest, true, GetEntity (), &tr);

               if (tr.flFraction > 0.8 || tr.pHit != g_worldEdict)
                  m_lookAt = dest;
            }
         }
         else
         {
            if ((g_experienceData + (index * g_numWaypoints) + index)->team1DangerIndex != -1)
            {
               Vector dest = g_waypoint->GetPath ((g_experienceData + (index * g_numWaypoints) + index)->team1DangerIndex)->origin;
               TraceLine (pev->origin, dest, true, GetEntity (), &tr);

               if (tr.flFraction > 0.8 || tr.pHit != g_worldEdict)
                  m_lookAt = dest;
            }
         }
      }

      if (canChooseAimDirection && m_prevWptIndex[0] >= 0 && m_prevWptIndex[0] < g_numWaypoints)
      {
         if (!(g_waypoint->GetPath (m_prevWptIndex[0])->flags & PointFlag_Ladder) && (fabsf (g_waypoint->GetPath (m_prevWptIndex[0])->origin.z - m_destOrigin.z) < 30.0 || (m_waypointFlags & PointFlag_Camp)))
         {
            // trace forward
            TraceLine (m_destOrigin, m_destOrigin + ((m_destOrigin - g_waypoint->GetPath (m_prevWptIndex[0])->origin).Normalize () * 96), true, GetEntity (), &tr);

            if (tr.flFraction < 1.0 && tr.pHit == g_worldEdict)
               m_lookAt = g_waypoint->GetPath (m_prevWptIndex[0])->origin + pev->view_ofs;
         }
      }
   }
   if (m_lookAt == Vector::GetNull ())
      m_lookAt = m_destOrigin;
}

void Bot::Think (void)
{
   pev->button = 0;
   pev->flags |= FL_FAKECLIENT; // restore fake client bit, if it were removed by some evil action =)

   m_moveSpeed = 0.0;
   m_strafeSpeed = 0.0;
   m_moveAngles = Vector::GetNull ();

   m_canChooseAimDirection = true;
   m_notKilled = IsAlive (GetEntity ());

   m_frameInterval = GetWorldTime () - m_lastThinkTime;
   m_lastThinkTime = GetWorldTime ();

   // is bot movement enabled
   bool botMovement = false;

   if (m_notStarted) // if the bot hasn't selected stuff to start the game yet, go do that...
      StartGame (); // select team & class
   else if (!m_notKilled)
   {
      // no movement allowed in
      if (m_voteKickIndex != m_lastVoteKick && g_botVar[Variable_Votes]->GetBool ()) // We got a Teamkiller? Vote him away...
      {
         FakeClientCommand (GetEntity (), "vote %d", m_voteKickIndex);
         m_lastVoteKick = m_voteKickIndex;

         // if bot tk punishment is enabled slay the tk
         if (g_botVar[Variable_TKPunish]->GetInt () != 2 || IsValidBot (INDEXENT (m_voteKickIndex)))
            return;

         entvars_t *killer = VARS (INDEXENT (m_lastVoteKick));

         MESSAGE_BEGIN (MSG_PAS, SVC_TEMPENTITY, killer->origin);
            WRITE_BYTE (TE_TAREXPLOSION);
            WRITE_COORD (killer->origin.x);
            WRITE_COORD (killer->origin.y);
            WRITE_COORD (killer->origin.z);
         MESSAGE_END ();

         MESSAGE_BEGIN (MSG_PVS, SVC_TEMPENTITY, killer->origin);
            WRITE_BYTE (TE_LAVASPLASH);
            WRITE_COORD (killer->origin.x);
            WRITE_COORD (killer->origin.y);
            WRITE_COORD (killer->origin.z);
         MESSAGE_END ();

         MESSAGE_BEGIN (MSG_ONE, g_netMsg->GetId (NetMessage_ScreenFade), NULL, ENT (killer));
            WRITE_SHORT (1 << 15);
            WRITE_SHORT (1 << 10);
            WRITE_SHORT (1 << 1);
            WRITE_BYTE (100);
            WRITE_BYTE (0);
            WRITE_BYTE (0);
            WRITE_BYTE (255);
         MESSAGE_END ();

         killer->frags++;
         MDLL_ClientKill (ENT (killer));

         HudMessage (ENT (killer), true, Vector (g_randGen.Long (33, 255), g_randGen.Long (33, 255), g_randGen.Long (33, 255)), "You was slayed, because of teamkilling a player. Please be careful.");

         // very fun thing
         (*g_engfuncs.pfnClientCommand) (ENT (killer), "cd eject\n");
      }
      else if (m_voteMap != 0) // host wants the bots to vote for a map?
      {
         FakeClientCommand (GetEntity (), "votemap %d", m_voteMap);
         m_voteMap = 0;
      }

      if (g_botVar[Variable_Chat]->GetBool () && !RepliesToPlayer () && m_lastChatTime + 10.0 < GetWorldTime () && g_lastChatTime + 5.0 < GetWorldTime ()) // bot chatting turned on?
      {
         // say a text every now and then
         if (g_randGen.Long (1, 1500) < 2)
         {
            m_lastChatTime = GetWorldTime ();
            g_lastChatTime = GetWorldTime ();

            char *pickedPhrase = const_cast <char *> (g_chatFactory[Chat_Dead].GetRandomElement ().GetBuffer ());
            bool sayBufferExists = false;

            // search for last messages, sayed
            IterateArray (m_sayTextBuffer.lastUsedSentences, i)
            {
               if (strncmp (m_sayTextBuffer.lastUsedSentences[i].GetBuffer (), pickedPhrase, m_sayTextBuffer.lastUsedSentences[i].GetLength ()) == 0)
                  sayBufferExists = true;
            }

            if (!sayBufferExists)
            {
               PrepareChatMessage (pickedPhrase);
               PushMessageQueue (GameMessage_SayMsg);

               // add to ignore list
               m_sayTextBuffer.lastUsedSentences.Push (pickedPhrase);
            }

            // clear the used line buffer every now and then
            if (m_sayTextBuffer.lastUsedSentences.GetElementNumber () > g_randGen.Long (4, 6))
               m_sayTextBuffer.lastUsedSentences.RemoveAll ();
         }
      }
   }
   else if (m_buyingFinished)
      botMovement = true;

  int team = g_clients[ENTINDEX (GetEntity ()) - 1].realTeam;;

   // remove voice icon
   if (g_lastRadioTime[team] + g_randGen.Float (0.8, 2.1) < GetWorldTime ())
      SwitchChatterIcon (false); // hide icon

   static float secondThinkTimer = 0.0;

   // check is it time to execute think (called once per second (not frame))
   if (secondThinkTimer < GetWorldTime ())
   {
      SecondThink ();

      // update timer to one second
      secondThinkTimer = GetWorldTime () + 1.05;
   }
   CheckMessageQueue (); // check for pending messages

   if (pev->maxspeed < 10 && GetCurrentTask ()->taskID != Task_PlantBomb && GetCurrentTask ()->taskID != Task_DefuseBomb)
      botMovement = false;

   if (m_notKilled && botMovement && !g_botVar[Variable_Stop]->GetBool ())
      BotAI (); // execute main code

   RunPlayerMovement (); // run the player movement
}

void Bot::SecondThink (void)
{
   // this function is called from main think function every second (second not frame).

   if (g_bombPlanted && GetTeam (GetEntity ()) == Team_Counter && (pev->origin - g_waypoint->GetBombPosition ()).GetLength () < 700 && !IsBombDefusing (g_waypoint->GetBombPosition ()))
      ResetTasks ();
}

void Bot::RunTask (void)
{
   // this is core function that handle task execution

   int team = GetTeam (GetEntity ());
   int destIndex, i;

   Vector src, destination;
   TraceResult tr;

   bool exceptionCaught = false;
   float fullDefuseTime, timeToBlowUp, defuseRemainingTime;

   switch (GetCurrentTask ()->taskID)
   {
   // normal task
   case Task_Normal:
      m_aimFlags |= AimPos_NavPoint;

      if ((g_mapType & MapType_De) && team == Team_Terrorist)
      {
         if (!g_bombPlanted)
         {
            m_loosedBombWptIndex = FindLoosedBomb ();

            if (m_loosedBombWptIndex != -1 && m_currentWaypointIndex != m_loosedBombWptIndex && g_randGen.Long (0, 100) < (GetNearbyFriendsNearPosition (g_waypoint->GetPath (m_loosedBombWptIndex)->origin, 650) >= 1 ? 40 : 90))
               GetCurrentTask ()->data = m_loosedBombWptIndex;
         }
         else if (!m_defendedBomb)
         {
            int plantedBombWptIndex = g_waypoint->FindNearest (g_waypoint->GetBombPosition ());

            if (plantedBombWptIndex != -1 && m_currentWaypointIndex != plantedBombWptIndex)
               GetCurrentTask ()->data = plantedBombWptIndex;
         }
      }

      // user forced a waypoint as a goal?
      if (g_botVar[Variable_DebugGoal]->GetInt () != -1)
      {
         // check if we reached it
         if (((g_waypoint->GetPath (m_currentWaypointIndex)->origin - pev->origin).SkipZ ()).GetLengthSquared () < 16 && GetCurrentTask ()->data == g_botVar[Variable_DebugGoal]->GetInt ())
         {
            m_moveSpeed = 0.0;
            m_strafeSpeed = 0.0;

            m_checkTerrain = false;
            m_moveToGoal = false;

            return; // we can safely return here
         }

         if (GetCurrentTask ()->data != g_botVar[Variable_DebugGoal]->GetInt ())
         {
            DeleteSearchNodes ();
            m_tasks->data = g_botVar[Variable_DebugGoal]->GetInt ();
         }
      }

      // bots rushing with knife, when have no enemy (thanks for idea to nicebot project)
      if (m_currentWeapon == Weapon_Knife && (FNullEnt (m_lastEnemy) || !IsAlive (m_lastEnemy)) && FNullEnt (m_enemy) && m_knifeAttackTime < GetWorldTime () && !HasShield () && GetNearbyFriendsNearPosition (pev->origin, 96) == 0)
      {
         if (g_randGen.Long (0, 100) < 40)
            pev->button |= IN_ATTACK;
         else
            pev->button |= IN_ATTACK2;

         m_knifeAttackTime = GetWorldTime () + g_randGen.Float (2.5, 6.0);
      }

      if (IsShieldDrawn ())
         pev->button |= IN_ATTACK2;

      if ((m_reloadState == ReloadState_None) && (GetAmmo () != 0))
         m_reloadState = ReloadState_Primary;

      // if bomb planted and it's a CT calculate new path to bomb point if he's not already heading for
      if (g_bombPlanted && team == Team_Counter && GetCurrentTask ()->data != -1 && !(g_waypoint->GetPath (m_tasks->data)->flags & PointFlag_Goal) && GetCurrentTask ()->taskID != Task_EscapeFromBomb)
      {
         DeleteSearchNodes ();
         m_tasks->data = -1;
      }

      if (!g_bombPlanted && m_currentWaypointIndex != -1 && (g_waypoint->GetPath (m_currentWaypointIndex)->flags & PointFlag_Goal) && g_randGen.Long (0, 100) < 80 && GetNearbyEnemiesNearPosition (pev->origin, 650) == 0)
         RadioMessage (Radio_SectorClear);

      // reached the destination (goal) waypoint?
      if (DoWaypointNav ())
      {
         TaskComplete ();
         m_prevGoalIndex = -1;

         // spray logo sometimes if allowed to do so
         if (m_timeLogoSpray < GetWorldTime () && g_botVar[Variable_Spray]->GetBool () && g_randGen.Long (1, 100) < 80 && m_moveSpeed > GetWalkSpeed ())
            StartTask (Task_SprayLogo, TASKPRI_SPRAYLOGO, -1, GetWorldTime () + 1.0, false);

         // reached waypoint is a camp waypoint
         if ((g_waypoint->GetPath (m_currentWaypointIndex)->flags & PointFlag_Camp) && !g_botVar[Variable_Deathmatch]->GetBool ())
         {
            // check if bot has got a primary weapon and hasn't camped before
            if (HasPrimaryWeapon () && m_timeCamping + 10.0 < GetWorldTime () && !HasHostage ())
            {
               bool campingAllowed = true;

               // Check if it's not allowed for this team to camp here
               if (team == Team_Terrorist)
               {
                  if (g_waypoint->GetPath (m_currentWaypointIndex)->flags & PointFlag_Counter)
                     campingAllowed = false;
               }
               else
               {
                  if (g_waypoint->GetPath (m_currentWaypointIndex)->flags & PointFlag_Terrorist)
                     campingAllowed = false;
               }

               // don't allow vip on as_ maps to camp + don't allow terrorist carrying c4 to camp
               if (((g_mapType & MapType_As) && *(INFOKEY_VALUE (GET_INFOKEYBUFFER (GetEntity ()), "model")) == 'v') || ((g_mapType & MapType_De) && GetTeam (GetEntity ()) == Team_Terrorist && !g_bombPlanted && (pev->weapons & (1 << Weapon_BombC4))))
                  campingAllowed = false;

               // check if another bot is already camping here
               if (IsWaypointUsed (m_currentWaypointIndex))
                  campingAllowed = false;

               if (campingAllowed)
               {
                  // crouched camping here?
                  if (g_waypoint->GetPath (m_currentWaypointIndex)->flags & PointFlag_Crouch)
                     m_campButtons = IN_DUCK;
                  else
                     m_campButtons = 0;

                  SelectBestWeapon ();

                  if (!(m_states & (State_SeeingEnemy | State_HearingEnemy)) && !m_reloadState)
                     m_reloadState = ReloadState_Primary;

                  MakeVectors (pev->v_angle);

                  m_timeCamping = GetWorldTime () + g_randGen.Float (g_skillTab[m_skill / 20].campStartDelay, g_skillTab[m_skill / 20].campEndDelay);
                  StartTask (Task_Camp, TASKPRI_CAMP, -1, m_timeCamping, true);

                  Vector src;
                  src.x = g_waypoint->GetPath (m_currentWaypointIndex)->campStartX;
                  src.y = g_waypoint->GetPath (m_currentWaypointIndex)->campStartY;

                  m_camp = src;
                  m_aimFlags |= AimPos_CampPoint;
                  m_campDirection = 0;

                  // tell the world we're camping
                  if (g_randGen.Long (0, 100) < 95)
                     RadioMessage (Radio_InPosition);

                  m_moveToGoal = false;
                  m_checkTerrain = false;

                  m_moveSpeed = 0;
                  m_strafeSpeed = 0;
               }
            }
         }
         else
         {
            // some goal waypoints are map dependant so check it out...
            if (g_mapType & MapType_Cs)
            {
               // CT Bot has some hostages following?
               if (HasHostage () && team == Team_Counter)
               {
                  // and reached a Rescue Point?
                  if (g_waypoint->GetPath (m_currentWaypointIndex)->flags & PointFlag_Rescue)
                  {
                     for (i = 0; i < Const_MaxHostages; i++)
                        m_hostages[i] = NULL; // clear array of hostage pointers
                  }
               }
               else if (team == Team_Terrorist && g_randGen.Long (0, 100) < 80)
               {
                  int index = FindDefendWaypoint (g_waypoint->GetPath (m_currentWaypointIndex)->origin);

                  StartTask (Task_Camp, TASKPRI_CAMP, -1, GetWorldTime () + g_randGen.Float (60.0, 120.0), true); // push camp task on to stack
                  StartTask (Task_MoveToPosition, TASKPRI_MOVETOPOSITION, index, GetWorldTime () + g_randGen.Float (10.0, 30.0), true); // push move command

                  if (g_waypoint->GetPath (index)->vis.crouch <= g_waypoint->GetPath (index)->vis.stand)
                     m_campButtons |= IN_DUCK;
                  else
                     m_campButtons &= ~IN_DUCK;

                  ChatterMessage (Chatter_GoingToGuardVIPSafety); // play info about that
               }
            }

            if ((g_mapType & MapType_De) && ((g_waypoint->GetPath (m_currentWaypointIndex)->flags & PointFlag_Goal) || m_inBombZone) && FNullEnt (m_enemy))
            {
               // is it a terrorist carrying the bomb?
               if (pev->weapons & (1 << Weapon_BombC4))
               {
                  if ((m_states & State_SeeingEnemy) && GetNearbyFriendsNearPosition (pev->origin, 768) == 0)
                  {
                     // request an help also
                     RadioMessage (Radio_NeedBackup);
                     InstantChatterMessage (Chatter_ScaredEmotion);

                     StartTask (Task_Camp, TASKPRI_CAMP, -1, GetWorldTime () + g_randGen.Float (4.0, 8.0), true);
                  }
                  else
                     StartTask (Task_PlantBomb, TASKPRI_PLANTBOMB, -1, 0.0, false);
               }
               else if (team == Team_Counter)
               {
                  if (!g_bombPlanted && GetNearbyFriendsNearPosition (pev->origin, 120) <= 4 && g_randGen.Long (0, 100) < 50 && GetCurrentTask ()->taskID == Task_Normal && m_fearLevel > m_agressionLevel / 2)
                  {
                     int index = FindDefendWaypoint (g_waypoint->GetPath (m_currentWaypointIndex)->origin);

                     StartTask (Task_Camp, TASKPRI_CAMP, -1, GetWorldTime () + g_randGen.Float (45.0, 80.0), true); // push camp task on to stack
                     StartTask (Task_MoveToPosition, TASKPRI_MOVETOPOSITION, index, GetWorldTime () + g_randGen.Float (10.0, 15.0), true); // push move command

                     if (g_waypoint->GetPath (index)->vis.crouch <= g_waypoint->GetPath (index)->vis.stand)
                        m_campButtons |= IN_DUCK;
                     else
                        m_campButtons &= ~IN_DUCK;

                     ChatterMessage (Chatter_DefendingBombSite); // play info about that
                  }
               }
            }
         }
      }
      // no more nodes to follow - search new ones (or we have a momb)
      else if (!GoalIsValid ())
      {
         m_moveSpeed = pev->maxspeed;
         DeleteSearchNodes ();

         // did we already decide about a goal before?
         if (GetCurrentTask ()->data != -1)
            destIndex = m_tasks->data;
         else
            destIndex = FindGoal ();

         m_prevGoalIndex = destIndex;

         // remember index
         m_tasks->data = destIndex;

         // do pathfinding if it's not the current waypoint
         if (destIndex != m_currentWaypointIndex)
         {
            if ((g_bombPlanted && team == Team_Counter) || g_botVar[Variable_DebugGoal]->GetInt () != -1)
               FindShortestPath (m_currentWaypointIndex, destIndex);
            else
               FindPath (m_currentWaypointIndex, destIndex, m_pathType);
         }
      }
      else
      {
         if (!(pev->flags & FL_DUCKING) && m_minSpeed != pev->maxspeed)
            m_moveSpeed = m_minSpeed;
      }

      if ((g_botVar[Variable_WalkAllowed]->GetBool() && g_botVar[Variable_FootSteps]->GetBool()) && m_skill > 80 && !(m_aimFlags & AimPos_Enemy) && (m_heardSoundTime + 13.0 >= GetWorldTime () || (m_states & (State_HearingEnemy | State_SuspectEnemy))) && GetNearbyEnemiesNearPosition (pev->origin, 1024) >= 1 && !(m_currentTravelFlags & PathFlag_Jump) && !(pev->button & IN_DUCK) && !(pev->flags & FL_DUCKING) && !g_botVar[Variable_KnifeMode]->GetBool () && !g_bombPlanted)
         m_moveSpeed = GetWalkSpeed ();

      // bot hasn't seen anything in a long time and is asking his teammates to report in
      if (m_seeEnemyTime != 0.0 && m_seeEnemyTime + g_randGen.Float (30.0, 80.0) < GetWorldTime () && g_randGen.Long (0, 100) < 70 && g_timeRoundStart + 20.0 < GetWorldTime () && m_askCheckTime + g_randGen.Float (20.0, 30.0) < GetWorldTime ())
      {
         m_askCheckTime = GetWorldTime ();
         RadioMessage (Radio_ReportTeam);
      }

      break;

   // bot sprays messy logos all over the place...
   case Task_SprayLogo:
      m_aimFlags |= AimPos_Entity;

      // bot didn't spray this round?
      if (m_timeLogoSpray <= GetWorldTime () && m_tasks->time > GetWorldTime ())
      {
         MakeVectors (pev->v_angle);
         Vector sprayOrigin = EyePosition () + (g_pGlobals->v_forward * 128);

         TraceLine (EyePosition (), sprayOrigin, true, GetEntity (), &tr);

         // no wall in front?
         if (tr.flFraction >= 1.0)
            sprayOrigin.z -= 128.0;

         m_entity = sprayOrigin;

         if (m_tasks->time - 0.5 < GetWorldTime ())
         {
            // emit spraycan sound
            EMIT_SOUND_DYN2 (GetEntity (), CHAN_VOICE, "player/sprayer.wav", 1.0, ATTN_NORM, 0, 100);
            TraceLine (EyePosition (), EyePosition () + g_pGlobals->v_forward * 128, true, GetEntity (), &tr);

            // paint the actual logo decal
            DecalTrace (pev, &tr, m_logotypeIndex);
            m_timeLogoSpray = GetWorldTime () + g_randGen.Float (30.0, 45.0);
         }
      }
      else
         TaskComplete ();

      m_moveToGoal = false;
      m_checkTerrain = false;

      m_navTimeset = GetWorldTime ();
      m_moveSpeed = 0;
      m_strafeSpeed = 0.0;

      break;

   // hunt down enemy
   case Task_HuntEnemy:
      m_aimFlags |= AimPos_NavPoint;
      m_checkTerrain = true;

      // if we've got new enemy...
      if (!FNullEnt (m_enemy) || FNullEnt (m_lastEnemy))
      {
         // forget about it...
         TaskComplete ();
         m_prevGoalIndex = -1;

         m_lastEnemy = NULL;
         m_lastEnemyOrigin = Vector::GetNull ();
      }
      else if (GetTeam (m_lastEnemy) == team)
      {
         // don't hunt down our teammate...
         RemoveCertainTask (Task_HuntEnemy);
         m_prevGoalIndex = -1;
      }
      else if (DoWaypointNav ()) // reached last enemy pos?
      {
         // forget about it...
         TaskComplete ();
         m_prevGoalIndex = -1;

         m_lastEnemy = NULL;
         m_lastEnemyOrigin = Vector::GetNull ();
      }
      else if (!GoalIsValid ()) // do we need to calculate a new path?
      {
         DeleteSearchNodes ();

         // is there a remembered index?
         if (GetCurrentTask ()->data != -1 && GetCurrentTask ()->data < g_numWaypoints)
            destIndex = m_tasks->data;
         else // no. we need to find a new one
            destIndex = g_waypoint->FindNearest (m_lastEnemyOrigin);

         // remember index
         m_prevGoalIndex = destIndex;
         m_tasks->data = destIndex;

         if (destIndex != m_currentWaypointIndex)
            FindPath (m_currentWaypointIndex, destIndex, m_pathType);
      }

      // bots skill higher than 60?
      if ((g_botVar[Variable_WalkAllowed]->GetBool () && g_botVar[Variable_FootSteps]->GetBool()) && m_skill > 60)
      {
         // then make him move slow if near enemy
         if (!(m_currentTravelFlags & PathFlag_Jump))
         {
            if (m_currentWaypointIndex != -1)
            {
               if (g_waypoint->GetPath (m_currentWaypointIndex)->radius < 32 && !IsOnLadder () && !IsInWater () && m_seeEnemyTime + 4.0 > GetWorldTime () && m_skill < 80)
                  pev->button |= IN_DUCK;
            }

            if ((m_lastEnemyOrigin - pev->origin).GetLength () < 512.0 && !(pev->flags & FL_DUCKING))
               m_moveSpeed = GetWalkSpeed ();
         }
      }
      break;

   // bot seeks cover from enemy
   case Task_SeekCover:
      m_aimFlags |= AimPos_NavPoint;

      if (FNullEnt (m_lastEnemy) || !IsAlive (m_lastEnemy))
      {
         TaskComplete ();
         m_prevGoalIndex = -1;
      }
      else if (DoWaypointNav ()) // reached final cover waypoint?
      {
         // yep. activate hide behaviour
         TaskComplete ();

         m_prevGoalIndex = -1;
         m_pathType = 1;

         // start hide task
         StartTask (Task_Hide, TASKPRI_HIDE, -1, GetWorldTime () + g_randGen.Float (5.0, 15.0), false);
         destination = m_lastEnemyOrigin;

         // get a valid look direction
         GetCampDirection (&destination);

         m_aimFlags |= AimPos_CampPoint;
         m_camp = destination;
         m_campDirection = 0;

         // chosen waypoint is a camp waypoint?
         if (g_waypoint->GetPath (m_currentWaypointIndex)->flags & PointFlag_Camp)
         {
            // use the existing camp wpt prefs
            if (g_waypoint->GetPath (m_currentWaypointIndex)->flags & PointFlag_Crouch)
               m_campButtons = IN_DUCK;
            else
               m_campButtons = 0;
         }
         else
         {
            // choose a crouch or stand pos
            if (g_waypoint->GetPath (m_currentWaypointIndex)->vis.crouch <= g_waypoint->GetPath (m_currentWaypointIndex)->vis.stand)
               m_campButtons = IN_DUCK;
            else
               m_campButtons = 0;

            // enter look direction from previously calculated positions
            g_waypoint->GetPath (m_currentWaypointIndex)->campStartX = destination.x;
            g_waypoint->GetPath (m_currentWaypointIndex)->campStartY = destination.y;

            g_waypoint->GetPath (m_currentWaypointIndex)->campStartX = destination.x;
            g_waypoint->GetPath (m_currentWaypointIndex)->campEndY = destination.y;
         }

         if ((m_reloadState == ReloadState_None) && (GetAmmoInClip () < 8) && (GetAmmo () != 0))
            m_reloadState = ReloadState_Primary;

         m_moveSpeed = 0;
         m_strafeSpeed = 0;

         m_moveToGoal = false;
         m_checkTerrain = true;
      }
      else if (!GoalIsValid ()) // we didn't choose a cover waypoint yet or lost it due to an attack?
      {
         DeleteSearchNodes ();

         if (GetCurrentTask ()->data != -1)
            destIndex = m_tasks->data;
         else
         {
            destIndex = FindCoverWaypoint (1024);

            if (destIndex == -1)
               destIndex = g_waypoint->FindFarest (pev->origin, 500);
         }

         m_campDirection = 0;
         m_prevGoalIndex = destIndex;
         m_tasks->data = destIndex;

         if (destIndex != m_currentWaypointIndex)
            FindPath (m_currentWaypointIndex, destIndex, 2);
      }
      break;

   // plain attacking
   case Task_AttackEnemy:
      m_moveToGoal = false;
      m_checkTerrain = false;

      if (!FNullEnt (m_enemy))
         CombatFight ();
      else
      {
         TaskComplete ();
         FindWaypoint ();

         m_destOrigin = m_lastEnemyOrigin;
      }
      m_navTimeset = GetWorldTime ();
      break;

   // Bot is pausing
   case Task_Pause:
      m_moveToGoal = false;
      m_checkTerrain = false;

      m_navTimeset = GetWorldTime ();
      m_moveSpeed = 0.0;
      m_strafeSpeed = 0.0;

      m_aimFlags |= AimPos_NavPoint;

      // is bot blinded and above average skill?
      if (m_viewDistance < 500.0 && m_skill > 60)
      {
         // go mad!
         m_moveSpeed = -fabsf ((m_viewDistance - 500.0) / 2.0);

         if (m_moveSpeed < -pev->maxspeed)
            m_moveSpeed = -pev->maxspeed;

         MakeVectors (pev->v_angle);
         m_camp = GetGunPosition () + (g_pGlobals->v_forward * 500);

         m_aimFlags |= AimPos_Override;
         m_wantsToFire = true;
      }
      else
         pev->button |= m_campButtons;

      // stop camping if time over or gets hurt by something else than bullets
      if (m_tasks->time < GetWorldTime () || m_lastDamageType > 0)
         TaskComplete ();
      break;

   // blinded (flashbanged) behaviour
   case Task_Blinded:
      m_moveToGoal = false;
      m_checkTerrain = false;
      m_navTimeset = GetWorldTime ();

      // if bot remembers last enemy position
      if (m_skill > 70 && m_lastEnemyOrigin != Vector::GetNull () && IsValidPlayer (m_lastEnemy) && !UsesSniper ())
      {
         m_lookAt = m_lastEnemyOrigin; // face last enemy
         m_wantsToFire = true; // and shoot it
      }

      m_moveSpeed = m_blindMoveSpeed;
      m_strafeSpeed = m_blindSidemoveSpeed;
      pev->button |= m_blindButton;

      if (m_blindTime < GetWorldTime ())
         TaskComplete ();

      break;

   // camping behaviour
   case Task_Camp:
      m_aimFlags |= AimPos_CampPoint;
      m_checkTerrain = false;
      m_moveToGoal = false;

      // half the reaction time if camping because you're more aware of enemies if camping
      m_idealReactionTime = (g_randGen.Float (g_skillTab[m_skill / 20].minSurpriseTime, g_skillTab[m_skill / 20].maxSurpriseTime)) / 2;
      m_navTimeset = GetWorldTime ();

      m_moveSpeed = 0;
      m_strafeSpeed = 0.0;

      GetValidWaypoint ();

      if (m_nextCampDirTime < GetWorldTime ())
      {
         m_nextCampDirTime = GetWorldTime () + g_randGen.Float (2.0, 5.0);

         if (g_waypoint->GetPath (m_currentWaypointIndex)->flags & PointFlag_Camp)
         {
            destination.z = 0;

            // switch from 1 direction to the other
            if (m_campDirection < 1)
            {
               destination.x = g_waypoint->GetPath (m_currentWaypointIndex)->campStartX;
               destination.y = g_waypoint->GetPath (m_currentWaypointIndex)->campStartY;
               m_campDirection ^= 1;
            }
            else
            {
               destination.x = g_waypoint->GetPath (m_currentWaypointIndex)->campEndX;
               destination.y = g_waypoint->GetPath (m_currentWaypointIndex)->campEndY;
               m_campDirection ^= 1;
            }

            // find a visible waypoint to this direction...
            // i know this is ugly hack, but i just don't want to break compatiability :)
            int numFoundPoints = 0;
            int foundPoints[3];
            int distanceTab[3];

            Vector dotA = (destination - pev->origin).Normalize2D ();

            for (i = 0; i < g_numWaypoints; i++)
            {
               // skip invisible waypoints or current waypoint
               if (!g_waypoint->IsVisible (m_currentWaypointIndex, i) || (i == m_currentWaypointIndex))
                  continue;

               Vector dotB = (g_waypoint->GetPath (i)->origin - pev->origin).Normalize2D ();

               if ((dotA | dotB) > 0.9)
               {
                  int distance = static_cast <int> ((pev->origin - g_waypoint->GetPath (i)->origin).GetLength ());

                  if (numFoundPoints >= 3)
                  {
                     for (int j = 0; j < 3; j++)
                     {
                        if (distance > distanceTab[j])
                        {
                           distanceTab[j] = distance;
                           foundPoints[j] = i;

                           break;
                        }
                     }
                  }
                  else
                  {
                     foundPoints[numFoundPoints] = i;
                     distanceTab[numFoundPoints] = distance;

                     numFoundPoints++;
                  }
               }
            }

            if (--numFoundPoints >= 0)
               m_camp = g_waypoint->GetPath (foundPoints[g_randGen.Long (0, numFoundPoints)])->origin;
            else
               m_camp = g_waypoint->GetPath (GetAimingWaypoint ())->origin;
         }
         else
            m_camp = g_waypoint->GetPath (GetAimingWaypoint ())->origin;
      }
      // press remembered crouch button
      pev->button |= m_campButtons;

      // stop camping if time over or gets hurt by something else than bullets
      if ((m_tasks->time < GetWorldTime ()) || (m_lastDamageType > 0))
         TaskComplete ();
      break;

   // hiding behaviour
   case Task_Hide:
      m_aimFlags |= AimPos_CampPoint;
      m_checkTerrain = false;
      m_moveToGoal = false;

      // half the reaction time if camping
      m_idealReactionTime = (g_randGen.Float (g_skillTab[m_skill / 20].minSurpriseTime, g_skillTab[m_skill / 20].maxSurpriseTime)) / 2;

      m_navTimeset = GetWorldTime ();
      m_moveSpeed = 0;
      m_strafeSpeed = 0.0;

      GetValidWaypoint ();

      if (HasShield () && !m_isReloading)
      {
         if (!IsShieldDrawn ())
            pev->button |= IN_ATTACK2; // draw the shield!
         else
            pev->button |= IN_DUCK; // duck under if the shield is already drawn
      }

      // if we see an enemy and aren't at a good camping point leave the spot
      if ((m_states & State_SeeingEnemy) || m_inBombZone)
      {
         if (!(g_waypoint->GetPath (m_currentWaypointIndex)->flags & PointFlag_Camp))
         {
            TaskComplete ();

            m_campButtons = 0;
            m_prevGoalIndex = -1;

            if (!FNullEnt (m_enemy))
               CombatFight ();

            break;
         }
      }
      else if (m_lastEnemyOrigin == Vector::GetNull ()) // If we don't have an enemy we're also free to leave
      {
         TaskComplete ();

         m_campButtons = 0;
         m_prevGoalIndex = -1;

         if (GetCurrentTask ()->taskID == Task_Hide)
            TaskComplete ();

         break;
      }

      pev->button |= m_campButtons;
      m_navTimeset = GetWorldTime ();

      // stop camping if time over or gets hurt by something else than bullets
      if (m_tasks->time < GetWorldTime () || m_lastDamageType > 0)
         TaskComplete ();

      break;

   // moves to a position specified in position has a higher priority than task_normal
   case Task_MoveToPosition:
      m_aimFlags |= AimPos_NavPoint;

      if (IsShieldDrawn ())
         pev->button |= IN_ATTACK2;

      if (m_reloadState == ReloadState_None && GetAmmo () != 0)
         m_reloadState = ReloadState_Primary;

      if (DoWaypointNav ()) // reached destination?
      {
         TaskComplete (); // we're done

         m_prevGoalIndex = -1;
         m_position = Vector::GetNull ();
      }
      else if (!GoalIsValid ()) // didn't choose goal waypoint yet?
      {
         DeleteSearchNodes ();

         if (GetCurrentTask ()->data != -1 && GetCurrentTask ()->data < g_numWaypoints)
            destIndex = m_tasks->data;
         else
            destIndex = g_waypoint->FindNearest (m_position);

         if (destIndex >= 0 && destIndex < g_numWaypoints)
         {
            m_prevGoalIndex = destIndex;
            m_tasks->data = destIndex;

            FindPath (m_currentWaypointIndex, destIndex, m_pathType);
         }
         else
            TaskComplete ();
      }
      break;

   // planting the bomb right now
   case Task_PlantBomb:
      m_aimFlags |= AimPos_CampPoint;

      destination = m_lastEnemyOrigin;
      GetCampDirection (&destination);

      if (pev->weapons & (1 << Weapon_BombC4)) // we're still got the C4?
      {
         SelectWeaponByName ("weapon_c4");

         if (IsAlive (m_enemy) || !m_inBombZone)
            TaskComplete ();
         else
         {
            m_moveToGoal = false;
            m_checkTerrain = false;
            m_navTimeset = GetWorldTime ();

            if (g_waypoint->GetPath (m_currentWaypointIndex)->flags & PointFlag_Crouch)
               pev->button |= (IN_ATTACK | IN_DUCK);
            else
               pev->button |= IN_ATTACK;

            m_moveSpeed = 0;
            m_strafeSpeed = 0;
         }
      }
      else // done with planting
      {
         TaskComplete ();

         // tell teammates to move over here...
         if (GetNearbyFriendsNearPosition (pev->origin, 1200) != 0)
            RadioMessage (Radio_NeedBackup);

         DeleteSearchNodes ();
         int index = FindDefendWaypoint (pev->origin);

         // push camp task on to stack
         StartTask (Task_Camp, TASKPRI_CAMP, -1, GetWorldTime () + ((g_botVar[Variable_C4Timer]->GetFloat () / 2) + (g_botVar[Variable_C4Timer]->GetFloat () / 4)), true);
         // push move command
         StartTask (Task_MoveToPosition, TASKPRI_MOVETOPOSITION, index, GetWorldTime () + ((g_botVar[Variable_C4Timer]->GetFloat () / 2) + (g_botVar[Variable_C4Timer]->GetFloat () / 4)), true);

         if (g_waypoint->GetPath (index)->vis.crouch <= g_waypoint->GetPath (index)->vis.stand)
            m_campButtons |= IN_DUCK;
         else
            m_campButtons &= ~IN_DUCK;
      }
      break;

   // bomb defusing behaviour
   case Task_DefuseBomb:
      fullDefuseTime = m_hasDefuser ? 6.0 : 11.0;
      timeToBlowUp = GetBombTimeleft ();
      defuseRemainingTime = fullDefuseTime;

      if (m_hasProgressBar && IsOnFloor ())
         defuseRemainingTime = fullDefuseTime - GetWorldTime();

      // exception: bomb has been defused
      if (g_waypoint->GetBombPosition () == Vector::GetNull ())
      {
         exceptionCaught = true;
         g_bombPlanted = false;

         if (GetNearbyFriendsNearPosition (pev->origin, 9999) != 0 && g_randGen.Long (0, 100) < 80)
         {
            if (timeToBlowUp < 3.0)
            {
               if (g_botVar[Variable_Communication]->GetInt () == 2)
                  ChatterMessage (Chatter_BarelyDefused);
               else
                  RadioMessage (Radio_SectorClear);
            }
            else
               RadioMessage (Radio_SectorClear);
         }
      }
      else if (defuseRemainingTime > timeToBlowUp) // exception: not time left for defusing
         exceptionCaught = true;
      else if (m_states & State_SeeingEnemy) // exception: saw/seeing enemy
      {
         if (GetNearbyFriendsNearPosition (pev->origin, 128) == 0)
         {
            if (defuseRemainingTime > 0.75)
            {
               if (GetNearbyFriendsNearPosition (pev->origin, 128) > 0)
                  RadioMessage (Radio_NeedBackup);

               exceptionCaught = true;
            }
         }
         else if (timeToBlowUp > fullDefuseTime + 3.0 && defuseRemainingTime > 1.0)
            exceptionCaught = true;
      }
      else if (m_states & State_SuspectEnemy) // exception: suspect enemy
      {
         if (GetNearbyFriendsNearPosition (pev->origin, 128) == 0)
         {
            if (timeToBlowUp > fullDefuseTime + 10.0)
            {
               if (GetNearbyFriendsNearPosition (pev->origin, 128) > 0)
                  RadioMessage (Radio_NeedBackup);

               exceptionCaught = true;
            }
         }
      }


      // one of exceptions is thrown. finish task.
      if (exceptionCaught)
      {
         m_checkTerrain = true;
         m_moveToGoal = true;

         m_destOrigin = Vector::GetNull ();
         m_entity = Vector::GetNull ();

         TaskComplete ();
         break;
      }

      // to revert from pause after reload waiting && just to be sure
      m_moveToGoal = false;
      m_checkTerrain = true;

      m_moveSpeed = pev->maxspeed;
      m_strafeSpeed = 0.0;

      // bot is reloading and we close enough to start defusing
      if (m_isReloading && (g_waypoint->GetBombPosition () - pev->origin).GetLength2D () < 80.0)
      {
         if (GetNearbyEnemiesNearPosition (pev->origin, 9999) == 0 || GetNearbyFriendsNearPosition (pev->origin, 768) > 2 || timeToBlowUp < fullDefuseTime + 8.0 || ((GetAmmoInClip () > 8 && m_reloadState == ReloadState_Primary) || (GetAmmoInClip () > 5 && m_reloadState == ReloadState_Secondary)))
         {
            int weaponIndex = GetHighestWeapon ();

            // just select knife and then select weapon
            SelectWeaponByName ("weapon_knife");


            if (weaponIndex > 0 && weaponIndex < Const_NumWeapons)
               SelectWeaponbyNumber (weaponIndex);

            m_isReloading = false;
         }
         else // just wait here
         {
            m_moveToGoal = false;
            m_checkTerrain = false;

            m_moveSpeed = 0.0;
            m_strafeSpeed = 0.0;
         }
      }

      // head to bomb and press use button
      m_aimFlags |= AimPos_Entity;

      m_destOrigin = g_waypoint->GetBombPosition ();
      m_entity = g_waypoint->GetBombPosition ();

      pev->button |= IN_USE;

      // if defusing is not already started, maybe crouch before
      if (!m_hasProgressBar && m_duckDefuseCheckTime < GetWorldTime ())
      {
         if (m_skill >= 80 && GetNearbyEnemiesNearPosition (pev->origin, 9999.0) != 0)
            m_duckDefuse = true;

         Vector botDuckOrigin, botStandOrigin;

         if (pev->button & IN_DUCK)
         {
            botDuckOrigin = pev->origin;
            botStandOrigin = pev->origin + Vector(0, 0, 18);
         }
         else
         {
            botDuckOrigin = pev->origin - Vector(0, 0, 18);
            botStandOrigin = pev->origin;
         }

         float duckLength = (m_entity - botDuckOrigin).GetLengthSquared ();
         float standLength = (m_entity - botStandOrigin).GetLengthSquared ();

         if (duckLength > 5625 || standLength > 5625)
         {
            if (standLength < duckLength)
               m_duckDefuse = false; // stand
            else
               m_duckDefuse = true; // duck
         }
         m_duckDefuseCheckTime = GetWorldTime () + 1.5;
      }

      // press duck button
      if (m_duckDefuse)
         pev->button |= IN_DUCK;
      else
         pev->button &= ~IN_DUCK;

      // we are defusing bomb
      if (m_hasProgressBar && IsOnFloor())
      {
         m_navTimeset = GetWorldTime ();

         // don't move when defusing
         m_moveToGoal = false;
         m_checkTerrain = false;

         m_moveSpeed = 0.0;
         m_strafeSpeed = 0.0;

         // notify team
         if (GetNearbyFriendsNearPosition (pev->origin, 9999) != 0)
         {
            ChatterMessage (Chatter_DefusingC4);

            if (GetNearbyFriendsNearPosition (pev->origin, 256) < 2)
               RadioMessage (Radio_NeedBackup);
         }
      }
      break;
      
   // follow user behaviour
   case Task_FollowUser:
      if (FNullEnt (m_targetEntity) || !IsAlive (m_targetEntity))
      {
         m_targetEntity = NULL;
         TaskComplete ();
         break;
      }

      if (m_targetEntity->v.button & IN_ATTACK)
      {
         MakeVectors (m_targetEntity->v.v_angle);
         TraceLine (m_targetEntity->v.origin + m_targetEntity->v.view_ofs, g_pGlobals->v_forward * 500, true, true, GetEntity (), &tr);

         if (!FNullEnt (tr.pHit) && IsValidPlayer (tr.pHit) && GetTeam (tr.pHit) != team)
         {
            m_targetEntity = NULL;
            m_lastEnemy = tr.pHit;
            m_lastEnemyOrigin = tr.pHit->v.origin;

            TaskComplete ();
            break;
         }
      }

      if (m_targetEntity->v.maxspeed != 0 && m_targetEntity->v.maxspeed < pev->maxspeed)
         m_moveSpeed = m_targetEntity->v.maxspeed;

      if (m_reloadState == ReloadState_None && GetAmmo () != 0)
         m_reloadState = ReloadState_Primary;

      if ((m_targetEntity->v.origin - pev->origin).GetLength () > 130)
         m_followWaitTime = 0.0;
      else
      {
         m_moveSpeed = 0.0;

         if (m_followWaitTime == 0.0)
            m_followWaitTime = GetWorldTime ();
         else
         {
            if (m_followWaitTime + 3.0 < GetWorldTime ())
            {
               // stop following if we have been waiting too long
               m_targetEntity = NULL;

               RadioMessage (Radio_YouTakePoint);
               TaskComplete ();

               break;
            }
         }
      }
      m_aimFlags |= AimPos_NavPoint;

      if (g_botVar[Variable_WalkAllowed]->GetBool () && m_targetEntity->v.maxspeed < m_moveSpeed)
         m_moveSpeed = GetWalkSpeed ();

      if (IsShieldDrawn ())
         pev->button |= IN_ATTACK2;

      if (DoWaypointNav ()) // reached destination?
         GetCurrentTask ()->data = -1;

      if (!GoalIsValid ()) // didn't choose goal waypoint yet?
      {
         DeleteSearchNodes ();

         destIndex = g_waypoint->FindNearest (m_targetEntity->v.origin);

         Array <int> points;
         g_waypoint->FindInRadius (points, 200, m_targetEntity->v.origin);

         while (!points.IsEmpty ())
         {
            int newIndex = points.Pop ();

            // if waypoint not yet used, assign it as dest
            if (!IsWaypointUsed (newIndex) && (newIndex != m_currentWaypointIndex))
               destIndex = newIndex;
         }

         if (destIndex >= 0 && destIndex < g_numWaypoints && destIndex != m_currentWaypointIndex)
         {
            m_prevGoalIndex = destIndex;
            m_tasks->data = destIndex;

            // always take the shortest path
            FindShortestPath (m_currentWaypointIndex, destIndex);
         }
         else
         {
            m_targetEntity = NULL;
            TaskComplete ();
         }
      }
      break;

   // HE grenade throw behaviour
   case Task_ThrowExpGrenade:
      m_aimFlags |= AimPos_Grenade;
      destination = m_throw;

      if (!(m_states & State_SeeingEnemy))
      {
         m_moveSpeed = 0.0;
         m_strafeSpeed = 0.0;

         m_moveToGoal = false;
      }
      else if (!(m_states & State_SuspectEnemy) && !FNullEnt (m_enemy))
         destination = m_enemy->v.origin + (m_enemy->v.velocity.SkipZ () * 0.5);

      m_isUsingGrenade = true;
      m_checkTerrain = false;

      if ((pev->origin - destination).GetLengthSquared () < 400 * 400)
      {
         // heck, I don't wanna blow up myself
         m_grenadeCheckTime = GetWorldTime () + Const_GrenadeTimer;

         SelectBestWeapon ();
         TaskComplete ();

         break;
      }

      m_grenade = CheckThrow (GetGunPosition (), destination);

      if (m_grenade.GetLengthSquared () < 100)
         m_grenade = CheckToss (GetGunPosition (), destination);

      if (m_grenade.GetLengthSquared () <= 100)
      {
         m_grenadeCheckTime = GetWorldTime () + Const_GrenadeTimer;
         m_grenade = m_lookAt;

         SelectBestWeapon ();
         TaskComplete ();
      }
      else
      {
         edict_t *ent = NULL;

         while (!FNullEnt (ent = FIND_ENTITY_BY_CLASSNAME (ent, "grenade")))
         {
            if (ent->v.owner == GetEntity () && strcmp (STRING (ent->v.model) + 9, "hegrenade.mdl") == 0)
            {
               // set the correct velocity for the grenade
               if (m_grenade.GetLengthSquared () > 100)
                  ent->v.velocity = m_grenade;

               m_grenadeCheckTime = GetWorldTime () + Const_GrenadeTimer;

               SelectBestWeapon ();
               TaskComplete ();

               break;
            }
         }

         if (FNullEnt (ent))
         {
            if (m_currentWeapon != Weapon_Esplosive)
            {
               if (pev->weapons & (1 << Weapon_Esplosive))
                  SelectWeaponByName ("weapon_hegrenade");
            }
            else if (!(pev->oldbuttons & IN_ATTACK))
               pev->button |= IN_ATTACK;
         }
      }
      pev->button |= m_campButtons;
      break;

   // flashbang throw behavior (basically the same code like for HE's)
   case Task_ThrowFlashGrenade:
      m_aimFlags |= AimPos_Grenade;
      destination = m_throw;

      if (!(m_states & State_SeeingEnemy))
      {
         m_moveSpeed = 0.0;
         m_strafeSpeed = 0.0;

         m_moveToGoal = false;
      }
      else if (!(m_states & State_SuspectEnemy) && !FNullEnt (m_enemy))
         destination = m_enemy->v.origin + (m_enemy->v.velocity.SkipZ () * 0.5);

      m_isUsingGrenade = true;
      m_checkTerrain = false;

      m_grenade = CheckThrow (GetGunPosition (), destination);

      if (m_grenade.GetLengthSquared () < 100)
         m_grenade = CheckToss (pev->origin, destination);

      if (m_grenade.GetLengthSquared () <= 100)
      {
         m_grenadeCheckTime = GetWorldTime () + Const_GrenadeTimer;
         m_grenade = m_lookAt;

         SelectBestWeapon ();
         TaskComplete ();
      }
      else
      {
         edict_t *ent = NULL;
         while (!FNullEnt (ent = FIND_ENTITY_BY_CLASSNAME (ent, "grenade")))
         {
            if (ent->v.owner == GetEntity () && strcmp (STRING (ent->v.model) + 9, "flashbang.mdl") == 0)
            {
               // set the correct velocity for the grenade
               if (m_grenade.GetLengthSquared () > 100)
                  ent->v.velocity = m_grenade;

               m_grenadeCheckTime = GetWorldTime () + Const_GrenadeTimer;

               SelectBestWeapon ();
               TaskComplete ();
               break;
            }
         }

         if (FNullEnt (ent))
         {
            if (m_currentWeapon != Weapon_Flashbang)
            {
               if (pev->weapons & (1 << Weapon_Flashbang))
                  SelectWeaponByName ("weapon_flashbang");
            }
            else if (!(pev->oldbuttons & IN_ATTACK))
               pev->button |= IN_ATTACK;
         }
      }
      pev->button |= m_campButtons;
      break;

   // smoke grenade throw behavior
   // a bit different to the others because it mostly tries to throw the sg on the ground
   case Task_ThrowSmokeGrenade:
      m_aimFlags |= AimPos_Grenade;

      if (!(m_states & State_SeeingEnemy))
      {
         m_moveSpeed = 0.0;
         m_strafeSpeed = 0.0;

         m_moveToGoal = false;
      }

      m_checkTerrain = false;
      m_isUsingGrenade = true;

      src = m_lastEnemyOrigin - pev->velocity;

      // predict where the enemy is in 0.5 secs
      if (!FNullEnt (m_enemy))
         src = src + m_enemy->v.velocity * 0.5;

      m_grenade = (src - GetGunPosition ()).Normalize ();

      if (m_tasks->time < GetWorldTime () + 0.5)
      {
         m_aimFlags &= ~AimPos_Grenade;
         m_states &= ~State_ThrowSmoke;

         TaskComplete ();
         break;
      }

      if (m_currentWeapon != Weapon_Smoke)
      {
         if (pev->weapons & (1 << Weapon_Smoke))
         {
            SelectWeaponByName ("weapon_smokegrenade");
            m_tasks->time = GetWorldTime () + Const_GrenadeTimer;
         }
         else
            m_tasks->time = GetWorldTime () + 0.1;
      }
      else if (!(pev->oldbuttons & IN_ATTACK))
         pev->button |= IN_ATTACK;
      break;

   // bot helps human player (or other bot) to get somewhere
   case Task_PerformDoubleJump:
      if (FNullEnt (m_doubleJumpEntity) || !IsAlive (m_doubleJumpEntity) || (m_aimFlags & AimPos_Enemy) || (m_travelStartIndex != -1 && m_tasks->time + (g_waypoint->GetTravelTime (pev->maxspeed, g_waypoint->GetPath (m_travelStartIndex)->origin, m_doubleJumpOrigin) + 11.0) < GetWorldTime ()))
      {
         ResetDoubleJumpState ();
         break;
      }
      m_aimFlags |= AimPos_NavPoint;

      if (m_jumpReady)
      {
         m_moveToGoal = false;
         m_checkTerrain = false;

         m_navTimeset = GetWorldTime ();
         m_moveSpeed = 0.0;
         m_strafeSpeed = 0.0;

         if (m_duckForJump < GetWorldTime ())
            pev->button |= IN_DUCK;

         TraceResult tr;
         MakeVectors (Vector::GetNull ());

         Vector dest = EyePosition () + (g_pGlobals->v_forward + 999.0);
         dest.z = 180.0;

         TraceLine (EyePosition (), dest, false, true, GetEntity (), &tr);

         if ((tr.flFraction < 1.0) && (tr.pHit == m_doubleJumpEntity))
         {
            if (m_doubleJumpEntity->v.button & IN_JUMP)
            {
               m_duckForJump = GetWorldTime () + g_randGen.Float (3.0, 5.0);
               m_tasks->time = GetWorldTime ();
            }
         }
         break;
      }

      if (m_currentWaypointIndex == m_prevGoalIndex)
      {
         m_waypointOrigin = m_doubleJumpOrigin;
         m_destOrigin = m_doubleJumpOrigin;
      }

      if (DoWaypointNav ()) // reached destination?
         GetCurrentTask ()->data = -1;

      if (!GoalIsValid ()) // didn't choose goal waypoint yet?
      {
         DeleteSearchNodes ();

         destIndex = g_waypoint->FindNearest (m_doubleJumpOrigin);

         if (destIndex >= 0 && destIndex < g_numWaypoints)
         {
            m_prevGoalIndex = destIndex;
            m_tasks->data = destIndex;
            m_travelStartIndex = m_currentWaypointIndex;

            // Always take the shortest path
            FindShortestPath (m_currentWaypointIndex, destIndex);

            if (m_currentWaypointIndex == destIndex)
               m_jumpReady = true;
         }
         else
            ResetDoubleJumpState ();
      }
      break;

   // escape from bomb behaviour
   case Task_EscapeFromBomb:
      m_aimFlags |= AimPos_NavPoint;

      if (!g_bombPlanted)
         TaskComplete ();

      if (IsShieldDrawn ())
         pev->button |= IN_ATTACK2;

      if (m_currentWeapon != Weapon_Knife && GetNearbyEnemiesNearPosition (pev->origin, 9999) == 0)
         SelectWeaponByName ("weapon_knife");

      if (DoWaypointNav ()) // reached destination?
      {
         TaskComplete (); // we're done

         // press duck button if we still have some enemies
         if (GetNearbyEnemiesNearPosition (pev->origin, 2048))
            m_campButtons = IN_DUCK;

         // we're reached destination point so just sit down and camp
         StartTask (Task_Camp, TASKPRI_CAMP, -1, GetWorldTime () + 10.0, true);
      }
      else if (!GoalIsValid ()) // didn't choose goal waypoint yet?
      {
         DeleteSearchNodes ();

         int lastSelectedGoal = -1;
         float safeRadius = g_randGen.Float (1024.0, 2048.0), minPathDistance = 4096.0;

         for (int i = 0; i < g_numWaypoints; i++)
         {
            if ((g_waypoint->GetPath (i)->origin - g_waypoint->GetBombPosition ()).GetLength () < safeRadius)
               continue;

            float pathDistance = g_waypoint->GetPathDistance (m_currentWaypointIndex, i);

            if (minPathDistance > pathDistance)
            {
               minPathDistance = pathDistance;
               lastSelectedGoal = i;
            }
         }

         if (lastSelectedGoal < 0)
            lastSelectedGoal = g_waypoint->FindFarest (pev->origin, safeRadius);

         m_prevGoalIndex = lastSelectedGoal;
         GetCurrentTask ()->data = lastSelectedGoal;

         FindShortestPath (m_currentWaypointIndex, lastSelectedGoal);
      }
      break;

   // shooting breakables in the way action
   case Task_ShootBreakable:
      m_aimFlags |= AimPos_Override;

      // Breakable destroyed?
      if (FNullEnt (FindBreakable ()))
      {
         TaskComplete ();
         break;
      }
      pev->button |= m_campButtons;

      m_checkTerrain = false;
      m_moveToGoal = false;
      m_navTimeset = GetWorldTime ();

      src = m_breakable;
      m_camp = src;

      // is bot facing the breakable?
      if (GetShootingConeDeviation (GetEntity (), &src) >= 0.90)
      {
         m_moveSpeed = 0.0;
         m_strafeSpeed = 0.0;

         if (m_currentWeapon == Weapon_Knife)
            SelectBestWeapon ();

         m_wantsToFire = true;
      }
      else
      {
         m_checkTerrain = true;
         m_moveToGoal = true;
      }
      break;

   // picking up items and stuff behaviour
   case Task_PickupItem:
      if (FNullEnt (m_pickupItem))
      {
         m_pickupItem = NULL;
         TaskComplete ();

         break;
      }

      destination = GetEntityOrigin (m_pickupItem);
      m_destOrigin = destination;
      m_entity = destination;

      // find the distance to the item
      float itemDistance = (destination - pev->origin).GetLength ();

      switch (m_pickupType)
      {
      case PickupType_Weapon:
         m_aimFlags |= AimPos_NavPoint;

         // near to weapon?
         if (itemDistance < 50)
         {
            for (i = 0; i < 7; i++)
            {
               if (strcmp (g_weaponSelect[i].modelName, STRING (m_pickupItem->v.model) + 9) == 0)
                  break;
            }

            if (i < 7)
            {
               // secondary weapon. i.e., pistol
               int weaponID = 0;

               for (i = 0; i < 7; i++)
               {
                  if (pev->weapons & (1 << g_weaponSelect[i].id))
                     weaponID = i;
               }

               if (weaponID > 0)
               {
                  SelectWeaponbyNumber (weaponID);
                  FakeClientCommand (GetEntity (), "drop");

                  if (HasShield ()) // If we have the shield...
                     FakeClientCommand (GetEntity (), "drop"); // discard both shield and pistol
               }
               EquipInBuyzone (0);
            }
            else
            {
               // primary weapon
               int weaponID = GetHighestWeapon ();

               if ((weaponID > 6) || HasShield ())
               {
                  SelectWeaponbyNumber (weaponID);
                  FakeClientCommand (GetEntity (), "drop");
               }
               EquipInBuyzone (0);
            }
            CheckSilencer (); // check the silencer
         }
         break;

      case PickupType_ShieldGun:
         m_aimFlags |= AimPos_NavPoint;

         if (HasShield ())
         {
            m_pickupItem = NULL;
            break;
         }
         else if (itemDistance < 50) // near to shield?
         {
            // get current best weapon to check if it's a primary in need to be dropped
            int weaponID = GetHighestWeapon ();

            if (weaponID > 6)
            {
               SelectWeaponbyNumber (weaponID);
               FakeClientCommand (GetEntity (), "drop");
            }
         }
         break;

      case PickupType_PlantedC4:
         m_aimFlags |= AimPos_Entity;

         if (team == Team_Counter && itemDistance < 55)
         {
            ChatterMessage (Chatter_DefusingC4);

            // notify team of defusing
            if (GetNearbyFriendsNearPosition (pev->origin, 9999) < 2)
               RadioMessage (Radio_NeedBackup);

            m_moveToGoal = false;
            m_checkTerrain = false;

            m_moveSpeed = 0;
            m_strafeSpeed = 0;

            StartTask (Task_DefuseBomb, TASKPRI_DEFUSEBOMB, -1, 0.0, false);
         }
         break;

      case PickupType_Hostage:
         m_aimFlags |= AimPos_Entity;
         src = EyePosition ();

         if (!IsAlive (m_pickupItem))
         {
            // don't pickup dead hostages
            m_pickupItem = NULL;
            TaskComplete ();

            break;
         }

         if (itemDistance < 50)
         {
            float angleToEntity = InFieldOfView (destination - src);

            if (angleToEntity <= 10) // bot faces hostage?
            {
               // use game dll function to make sure the hostage is correctly 'used'
               MDLL_Use (m_pickupItem, GetEntity ());

               if (g_randGen.Long (0, 100) < 80)
                  ChatterMessage (Chatter_UseHostage);

               for (i = 0; i < Const_MaxHostages; i++)
               {
                  if (FNullEnt (m_hostages[i])) // store pointer to hostage so other bots don't steal from this one or bot tries to reuse it
                  {
                     m_hostages[i] = m_pickupItem;
                     m_pickupItem = NULL;

                     break;
                  }
               }
            }
            m_lastCollTime = GetWorldTime () + 0.1; // also don't consider being stuck
         }
         break;

      case PickupType_DefusalKit:
         m_aimFlags |= AimPos_NavPoint;

         if (m_hasDefuser)
         {
            m_pickupItem = NULL;
            m_pickupType = PickupType_None;
         }
         break;

      case PickupType_Button:
         m_aimFlags |= AimPos_Entity;

         if (FNullEnt (m_pickupItem) || m_buttonPushTime < GetWorldTime ()) // it's safer...
         {
            TaskComplete ();
            m_pickupType = PickupType_None;

            break;
         }

         // find angles from bot origin to entity...
         src = EyePosition ();
         float angleToEntity = InFieldOfView (destination - src);

         if (itemDistance < 90) // near to the button?
         {
            m_moveSpeed = 0.0;
            m_strafeSpeed = 0.0;
            m_moveToGoal = false;
            m_checkTerrain = false;

            if (angleToEntity <= 10) // facing it directly?
            {
               MDLL_Use (m_pickupItem, GetEntity ());

               m_pickupItem = NULL;
               m_pickupType = PickupType_None;
               m_buttonPushTime = GetWorldTime () + 3.0;

               TaskComplete ();
            }
         }
         break;
      }
      break;
   }
}

void Bot::BotAI (void)
{
   // this function gets called each frame and is the core of all bot ai. from here all other subroutines are called

   float movedDistance; // length of different vector (distance bot moved)
   TraceResult tr;

   int team = GetTeam (GetEntity ());

   // switch to knife if time to do this
   if (m_checkKnifeSwitch && m_buyingFinished && m_spawnTime + g_randGen.Float (4.0, 6.5) < GetWorldTime ())
   {
      if (g_randGen.Long (1, 100) < 2 && g_botVar[Variable_Spray]->GetBool ())
         StartTask (Task_SprayLogo, TASKPRI_SPRAYLOGO, -1, GetWorldTime () + 1.0, false);

      if (m_skill > 75 && g_randGen.Long (0, 100) < (m_personality == Personality_Rusher ? 99 : 80) && !m_isReloading && (g_mapType & (MapType_Cs | MapType_De | MapType_Es | MapType_As)))
        SelectWeaponByName ("weapon_knife");

      m_checkKnifeSwitch = false;

      if (g_randGen.Long (0, 100) < g_botVar[Variable_AutoFollow]->GetInt () && FNullEnt (m_targetEntity) && !m_isLeader && !(pev->weapons & (1 << Weapon_BombC4)))
         AttachToUser ();
   }

   // check if we already switched weapon mode
   if (m_checkWeaponSwitch && m_buyingFinished && m_spawnTime + g_randGen.Float (2.0, 3.5) < GetWorldTime ())
   {
      if (HasShield () && IsShieldDrawn ())
        pev->button |= IN_ATTACK2;
      else
      {
         switch (m_currentWeapon)
         {
         case Weapon_M4A1:
         case Weapon_Usp:
            CheckSilencer ();
            break;

         case Weapon_Famas:
         case Weapon_Glock18:
            if (g_randGen.Long (0, 100) < 50)
               pev->button |= IN_ATTACK2;
            break;
         }
      }

      // select a leader bot for this team
      SelectLeaderEachTeam (team);
      m_checkWeaponSwitch = false;
   }

   // warning: the following timers aren't frame independent so it varies on slower/faster computers

   // increase reaction time
   m_actualReactionTime += 0.2;

   if (m_actualReactionTime > m_idealReactionTime)
      m_actualReactionTime = m_idealReactionTime;

   // bot could be blinded by flashbang or smoke, recover from it
   m_viewDistance += 3.0;

   if (m_viewDistance > m_maxViewDistance)
      m_viewDistance = m_maxViewDistance;

   if (m_blindTime > GetWorldTime ())
      m_maxViewDistance = 4096.0;

   m_moveSpeed = pev->maxspeed;

   if (m_prevTime <= GetWorldTime ())
   {
      // see how far bot has moved since the previous position...
      movedDistance = (m_prevOrigin - pev->origin).GetLength ();

      // save current position as previous
      m_prevOrigin = pev->origin;
      m_prevTime = GetWorldTime () + 0.2;
   }
   else
      movedDistance = 2.0;

   // if there's some radio message to respond, check it
   if (m_radioOrder != 0)
      CheckRadioCommands ();

   // do all sensing, calculate/filter all actions here
   SetConditions ();

   // some stuff required by by chatter engine
   if ((m_states & State_SeeingEnemy) && !FNullEnt (m_enemy))
   {
      if (g_randGen.Long (0, 100) < 45 && GetNearbyFriendsNearPosition (pev->origin, 512) == 0 && (m_enemy->v.weapons & (1 << Weapon_BombC4)))
         ChatterMessage (Chatter_SpotTheBomber);

      if (g_randGen.Long (0, 100) < 45 && GetTeam (GetEntity ()) == Team_Terrorist && GetNearbyFriendsNearPosition (pev->origin, 512) == 0 && *g_engfuncs.pfnInfoKeyValue (g_engfuncs.pfnGetInfoKeyBuffer (m_enemy), "model") == 'v')
         ChatterMessage (Chatter_VIPSpotted);

      if (g_randGen.Long (0, 100) < 50 && GetNearbyFriendsNearPosition (pev->origin, 450) == 0 && GetTeam (m_enemy) != GetTeam (GetEntity ()) && IsGroupOfEnemies (m_enemy->v.origin, 2, 384))
         ChatterMessage (Chatter_ScaredEmotion);

      if (g_randGen.Long (0, 100) < 40 && GetNearbyFriendsNearPosition (pev->origin, 1024) == 0 && ((m_enemy->v.weapons & (1 << Weapon_Magnum)) || (m_enemy->v.weapons & (1 << Weapon_Scout)) ||  (m_enemy->v.weapons & (1 << Weapon_G3SG1)) || (m_enemy->v.weapons & (1 << Weapon_Sig550))))
         ChatterMessage (Chatter_SniperWarning);
   }
   Vector src, destination;

   m_checkTerrain = true;
   m_moveToGoal = true;
   m_wantsToFire = false;

   AvoidGrenades (); // avoid flyings grenades
   m_isUsingGrenade = false;

   RunTask (); // execute current task
   ChooseAimDirection (); // choose aim direction
   FacePosition (); // and turn to chosen aim direction

   // the bots wants to fire at something?
   if (m_wantsToFire && !m_isUsingGrenade && m_shootTime <= GetWorldTime ())
      FireWeapon (); // if bot didn't fire a bullet try again next frame

   // check for reloading
   if (m_reloadCheckTime <= GetWorldTime ())
      CheckReload ();

   // set the reaction time (surprise momentum) different each frame according to skill
   m_idealReactionTime = g_randGen.Float (g_skillTab[m_skill / 20].minSurpriseTime, g_skillTab[m_skill / 20].maxSurpriseTime);

   // calculate 2 direction vectors, 1 without the up/down component
   Vector directionOld = m_destOrigin - (pev->origin + pev->velocity * m_frameInterval);
   Vector directionNormal = directionOld.Normalize ();

   Vector direction = directionNormal;
   directionNormal.z = 0.0;

   m_moveAngles = directionOld.ToAngles ();

   m_moveAngles.ClampAngles ();
   m_moveAngles.x *= -1.0; // invert for engine

#if 0
   if (g_botVar[Variable_HardcoreMode]->GetBool () && GetCurrentTask ()->taskID == Task_Normal && ((m_aimFlags & AimPos_Enemy) || (m_states & State_SeeingEnemy)) && !IsOnLadder ())
      CombatFight ();
#else
   if (g_botVar[Variable_HardcoreMode]->GetInt () == 1 && ((m_aimFlags & AimPos_Enemy) || (m_states & (State_SeeingEnemy | State_SuspectEnemy)) || (GetCurrentTask ()->taskID == Task_SeekCover && (m_isReloading || m_isVIP))) && !g_botVar[Variable_KnifeMode]->GetBool () && GetCurrentTask ()->taskID != Task_Camp && !IsOnLadder ())
   {
      m_moveToGoal = false; // don't move to goal
      m_navTimeset = GetWorldTime ();

      if (IsValidPlayer (m_enemy))
         CombatFight ();
   }
#endif
   // check if we need to escape from bomb
   if ((g_mapType & MapType_De) && g_bombPlanted && m_notKilled && GetCurrentTask ()->taskID != Task_EscapeFromBomb && GetCurrentTask ()->taskID != Task_Camp && OutOfBombTimer ())
   {
      TaskComplete (); // complete current task

      // then start escape from bomb immidiate
      StartTask (Task_EscapeFromBomb, TASKPRI_ESCAPEFROMBOMB, -1, 0.0, true);
   }

   // allowed to move to a destination position?
   if (m_moveToGoal)
   {
      GetValidWaypoint ();

      // Press duck button if we need to
      if ((g_waypoint->GetPath (m_currentWaypointIndex)->flags & PointFlag_Crouch) && !(g_waypoint->GetPath (m_currentWaypointIndex)->flags & PointFlag_Camp))
         pev->button |= IN_DUCK;

      m_timeWaypointMove = GetWorldTime ();

      if (IsInWater ()) // special movement for swimming here
      {
         // check if we need to go forward or back press the correct buttons
         if (InFieldOfView (m_destOrigin - EyePosition ()) > 90)
            pev->button |= IN_BACK;
         else
            pev->button |= IN_FORWARD;

         if (m_moveAngles.x > 60.0)
            pev->button |= IN_DUCK;
         else if (m_moveAngles.x < -60.0)
            pev->button |= IN_JUMP;
      }
   }

   if (m_checkTerrain) // are we allowed to check blocking terrain (and react to it)?
   {
      m_isStuck = false;
      edict_t *ent = NULL;

      // Test if there's a shootable breakable in our way
      if (!FNullEnt (ent = FindBreakable ()))
      {
         m_breakableEntity = ent;
         m_campButtons = pev->button & IN_DUCK;

         StartTask (Task_ShootBreakable, TASKPRI_SHOOTBREAKABLE, -1, 0.0, false);
      }
      else
      {
         ent = NULL;
         edict_t *pentNearest = NULL;

         if (FindNearestPlayer (reinterpret_cast <void **> (&pentNearest), GetEntity (), pev->maxspeed, true, false, true, true)) // found somebody?
         {
            MakeVectors (m_moveAngles); // use our movement angles

            // try to predict where we should be next frame
            Vector moved = pev->origin + g_pGlobals->v_forward * m_moveSpeed * m_frameInterval;
            moved = moved + g_pGlobals->v_right * m_strafeSpeed * m_frameInterval;
            moved = moved + pev->velocity * m_frameInterval;

            float nearestDistance = (pentNearest->v.origin - pev->origin).GetLength2D ();
            float movedDistance = (pentNearest->v.origin - moved).GetLength2D ();
            float nextFrameDistance = ((pentNearest->v.origin + pentNearest->v.velocity * m_frameInterval) - pev->origin).GetLength2D ();

            // is player that near now or in future that we need to steer away?
            if (movedDistance <= 48.0 || (nearestDistance <= 56.0 && nextFrameDistance < nearestDistance))
            {
               // to start strafing, we have to first figure out if the target is on the left side or right side
               Vector dirToPoint = (pev->origin - pentNearest->v.origin).SkipZ ();

               if ((dirToPoint | g_pGlobals->v_right.SkipZ ()) > 0.0)
                  SetStrafeSpeed (directionNormal, pev->maxspeed);
               else
                  SetStrafeSpeed (directionNormal, -pev->maxspeed);

               ResetCollideState ();

               if (nearestDistance < 56.0 && (dirToPoint | g_pGlobals->v_forward.SkipZ ()) < 0.0)
                  m_moveSpeed = -pev->maxspeed;
            }
         }

         // Standing still, no need to check?
         // FIXME: doesn't care for ladder movement (handled separately) should be included in some way
         if ((m_moveSpeed >= 10 || m_strafeSpeed >= 10) && m_lastCollTime < GetWorldTime ())
         {
            if (movedDistance < 2.0 && m_prevSpeed >= 1.0) // didn't we move enough previously?
            {
               // Then consider being stuck
               m_prevTime = GetWorldTime ();
               m_isStuck = true;

               if (m_firstCollideTime == 0.0)
                  m_firstCollideTime = GetWorldTime () + 0.2;
            }
            else // not stuck yet
            {
               // test if there's something ahead blocking the way
               if (CantMoveForward (directionNormal, &tr) && !IsOnLadder ())
               {
                  if (m_firstCollideTime == 0.0)
                     m_firstCollideTime = GetWorldTime () + 0.2;

                  else if (m_firstCollideTime <= GetWorldTime ())
                     m_isStuck = true;
               }
               else
                  m_firstCollideTime = 0.0;
            }

            if (!m_isStuck) // not stuck?
            {
               if (m_probeTime + 0.5 < GetWorldTime ())
                  ResetCollideState (); // reset collision memory if not being stuck for 0.5 secs
               else
               {
                  // remember to keep pressing duck if it was necessary ago
                  if (m_collideMoves[m_collStateIndex] == Collision_Duck && IsOnFloor () || IsInWater ())
                     pev->button |= IN_DUCK;
               }
            }
            else // bot is stuck!
            {
               // not yet decided what to do?
               if (m_collisionState == Collision_NotDecided)
               {
                  char bits = 0;

                  if (IsOnLadder ())
                     bits = Probe_Strafe;
                  else if (IsInWater ())
                     bits = (Probe_Jump | Probe_Strafe);
                  else
                     bits = ((g_randGen.Long (0, 10) > 7 ? Probe_Jump : 0) | Probe_Strafe | Probe_Duck);

                  // collision check allowed if not flying through the air
                  if (IsOnFloor () || IsOnLadder () || IsInWater ())
                  {
                     char state[8];
                     int i = 0;

                     // first 4 entries hold the possible collision states
                     state[i++] = Collision_Jump;
                     state[i++] = Collision_Duck;
                     state[i++] = Collision_StrafeLeft;
                     state[i++] = Collision_StrafeRight;

                     // now weight all possible states
                     if (bits & Probe_Jump)
                     {
                        state[i] = 0;

                        if (CanJumpUp (directionNormal))
                           state[i] += 10;

                        if (m_destOrigin.z >= pev->origin.z + 18.0)
                           state[i] += 5;

                        if (EntityIsVisible (m_destOrigin))
                        {
                           MakeVectors (m_moveAngles);

                           src = EyePosition ();
                           src = src + (g_pGlobals->v_right * 15);

                           TraceLine (src, m_destOrigin, true, true, GetEntity (), &tr);

                           if (tr.flFraction >= 1.0)
                           {
                              src = EyePosition ();
                              src = src - (g_pGlobals->v_right * 15);

                              TraceLine (src, m_destOrigin, true, true, GetEntity (), &tr);

                              if (tr.flFraction >= 1.0)
                                 state[i] += 5;
                           }
                        }
                        if (pev->flags & FL_DUCKING)
                           src = pev->origin;
                        else
                           src = pev->origin + Vector (0, 0, -17);

                        destination = src + directionNormal * 30;
                        TraceLine (src, destination, true, true, GetEntity (), &tr);

                        if (tr.flFraction != 1.0)
                           state[i] += 10;
                     }
                     else
                        state[i] = 0;
                     i++;

                     if (bits & Probe_Duck)
                     {
                        state[i] = 0;

                        if (CanDuckUnder (directionNormal))
                           state[i] += 10;

                        if ((m_destOrigin.z + 36.0 <= pev->origin.z) && EntityIsVisible (m_destOrigin))
                           state[i] += 5;
                     }
                     else
                        state[i] = 0;
                     i++;

                     if (bits & Probe_Strafe)
                     {
                        state[i] = 0;
                        state[i + 1] = 0;

                        // to start strafing, we have to first figure out if the target is on the left side or right side
                        MakeVectors (m_moveAngles);

                        Vector dirToPoint = (pev->origin - m_destOrigin).Normalize2D ();
                        Vector rightSide = g_pGlobals->v_right.Normalize2D ();

                        bool dirRight = false;
                        bool dirLeft = false;
                        bool blockedLeft = false;
                        bool blockedRight = false;

                        if (dirToPoint | rightSide > 0)
                           dirRight = true;
                        else
                           dirLeft = true;

                        if (m_moveSpeed > 0)
                           direction = g_pGlobals->v_forward;
                        else
                           direction = -g_pGlobals->v_forward;

                        // now check which side is blocked
                        src = pev->origin + (g_pGlobals->v_right * 32);
                        destination = src + (direction * 32);

                        TraceHull (src, destination, true, head_hull, GetEntity (), &tr);

                        if (tr.flFraction != 1.0)
                           blockedRight = true;

                        src = pev->origin - (g_pGlobals->v_right * 32);
                        destination = src + (direction * 32);

                        TraceHull (src, destination, true, head_hull, GetEntity (), &tr);

                        if (tr.flFraction != 1.0)
                           blockedLeft = true;

                        if (dirLeft)
                           state[i] += 5;
                        else
                           state[i] -= 5;

                        if (blockedLeft)
                           state[i] -= 5;

                        i++;

                        if (dirRight)
                           state[i] += 5;
                        else
                           state[i] -= 5;

                        if (blockedRight)
                           state[i] -= 5;
                     }
                     else
                     {
                        state[i] = 0;
                        i++;

                        state[i] = 0;
                     }

                     // weighted all possible moves, now sort them to start with most probable
                     int temp = 0;
                     bool isSorting = false;

                     do
                     {
                        isSorting = false;
                        for (i = 0; i < 3; i++)
                        {
                           if (state[i + 4] < state[i + 5])
                           {
                              temp = state[i];

                              state[i] = state[i + 1];
                              state[i + 1] = temp;

                              temp = state[i + 4];

                              state[i + 4] = state[i + 5];
                              state[i + 5] = temp;

                              isSorting = true;
                           }
                        }
                     } while (isSorting);

                     for (i = 0; i < 4; i++)
                        m_collideMoves[i] = state[i];

                     m_collideTime = GetWorldTime ();
                     m_probeTime = GetWorldTime () + 0.5;
                     m_collisionProbeBits = bits;
                     m_collisionState = Collision_Probing;
                     m_collStateIndex = 0;
                  }
               }

               if (m_collisionState == Collision_Probing)
               {
                  if (m_probeTime < GetWorldTime ())
                  {
                     m_collStateIndex++;
                     m_probeTime = GetWorldTime () + 0.5;

                     if (m_collStateIndex > 4)
                     {
                        m_navTimeset = GetWorldTime () - 5.0;

                        ResetCollideState ();
                     }
                  }

                  if (m_collStateIndex <= 4)
                  {
                     switch (m_collideMoves[m_collStateIndex])
                     {
                     case Collision_Jump:
                        if (IsOnFloor () || IsInWater ())
                           pev->button |= IN_JUMP;
                        break;

                     case Collision_Duck:
                        if (IsOnFloor () || IsInWater ())
                           pev->button |= IN_DUCK;
                        break;

                     case Collision_StrafeLeft:
                        SetStrafeSpeed (directionNormal, -pev->maxspeed);
                        break;

                     case Collision_StrafeRight:
                        SetStrafeSpeed (directionNormal, pev->maxspeed);
                        break;
                     }
                  }
               }
            }
         }
      }
   }

   // must avoid a grenade?
   if (m_needAvoidGrenade != 0)
   {
      // Don't duck to get away faster
      pev->button &= ~IN_DUCK;

      m_moveSpeed = -pev->maxspeed;
      m_strafeSpeed = pev->maxspeed * m_needAvoidGrenade;
   }

   // time to reach waypoint
   if (m_navTimeset + GetEstimatedReachTime () < GetWorldTime () && FNullEnt (m_enemy))
   {
      GetValidWaypoint ();

      // clear these pointers, bot mingh be stuck getting to them
      if (!FNullEnt (m_pickupItem) && !m_hasProgressBar)
         m_itemIgnore = m_pickupItem;

      m_pickupItem = NULL;
      m_breakableEntity = NULL;
      m_itemCheckTime = GetWorldTime () + 5.0;
      m_pickupType = PickupType_None;
   }

   if (m_duckTime > GetWorldTime ())
      pev->button |= IN_DUCK;

   if (pev->button & IN_JUMP)
      m_jumpTime = GetWorldTime ();

   if (m_jumpTime + 0.85 > GetWorldTime ())
   {
      if (!IsOnFloor () && !IsInWater ())
         pev->button |= IN_DUCK;
   }

   if (!(pev->button & (IN_FORWARD | IN_BACK)))
   {
      if (m_moveSpeed > 0)
         pev->button |= IN_FORWARD;
      else if (m_moveSpeed < 0)
         pev->button |= IN_BACK;
   }

   if (!(pev->button & (IN_MOVELEFT | IN_MOVERIGHT)))
   {
      if (m_strafeSpeed > 0)
         pev->button |= IN_MOVERIGHT;
      else if (m_strafeSpeed < 0)
         pev->button |= IN_MOVELEFT;
   }

   static float timeDebugUpdate = 0.0;

   if (!FNullEnt (g_hostEntity) && g_botVar[Variable_Debug]->GetInt () >= 1)
   {
      int specIndex = g_hostEntity->v.iuser2;

      if (specIndex == ENTINDEX (GetEntity ()))
      {
         static int index, goal, taskID;

         if (m_tasks != NULL)
         {
            if (taskID != m_tasks->taskID || index != m_currentWaypointIndex || goal != m_tasks->data || timeDebugUpdate < GetWorldTime ())
            {
               taskID = m_tasks->taskID;
               index = m_currentWaypointIndex;
               goal = m_tasks->data;

               char szTaskname[80];

               switch (taskID)
               {
               case Task_Normal:
                  sprintf (szTaskname, "Normal");
                  break;

               case Task_Pause:
                  sprintf (szTaskname, "Pause");
                  break;

               case Task_MoveToPosition:
                  sprintf (szTaskname, "MoveToPosition");
                  break;

               case Task_FollowUser:
                  sprintf (szTaskname, "FollowUser");
                  break;

               case Task_WaitForGo:
                  sprintf (szTaskname, "WaitForGo");
                  break;

               case Task_PickupItem:
                  sprintf (szTaskname, "PickupItem");
                  break;

               case Task_Camp:
                  sprintf (szTaskname, "Camp");
                  break;

               case Task_PlantBomb:
                  sprintf (szTaskname, "PlantBomb");
                  break;

               case Task_DefuseBomb:
                  sprintf (szTaskname, "DefuseBomb");
                  break;

               case Task_AttackEnemy:
                  sprintf (szTaskname, "AttackEnemy");
                  break;

               case Task_HuntEnemy:
                  sprintf (szTaskname, "HuntEnemy");
                  break;

               case Task_SeekCover:
                  sprintf (szTaskname, "SeekCover");
                  break;

               case Task_ThrowExpGrenade:
                  sprintf (szTaskname, "ThrowExpGrenade");
                  break;

               case Task_ThrowFlashGrenade:
                  sprintf (szTaskname, "ThrowFlashGrenade");
                  break;

               case Task_ThrowSmokeGrenade:
                  sprintf (szTaskname, "ThrowSmokeGrenade");
                  break;

               case Task_PerformDoubleJump:
                  sprintf (szTaskname, "PerformDoubleJump");
                  break;

               case Task_EscapeFromBomb:
                  sprintf (szTaskname, "EscapeFromBomb");
                  break;

               case Task_ShootBreakable:
                  sprintf (szTaskname, "ShootBreakable");
                  break;

               case Task_Hide:
                  sprintf (szTaskname, "Hide");
                  break;

               case Task_Blinded:
                  sprintf (szTaskname, "Blinded");
                  break;

               case Task_SprayLogo:
                  sprintf (szTaskname, "SprayLogo");
                  break;
               }

               char enemyName[80], weaponName[80], aimFlags[32], botType[32];

               if (!FNullEnt (m_enemy))
                  strcpy (enemyName, STRING (m_enemy->v.netname));
               else if (!FNullEnt (m_lastEnemy))
               {
                  strcpy (enemyName, " (L)");
                  strcat (enemyName, STRING (m_lastEnemy->v.netname));
               }
               else
                  strcpy (enemyName, " (null)");

               char pickupName[80];

               if (!FNullEnt (m_pickupItem))
                  strcpy (pickupName, STRING (m_pickupItem->v.classname));
               else
                  strcpy (pickupName, " (null)");

               WeaponSelect_t *selectTab = &g_weaponSelect[0];
               char weaponCount = 0;

               while (m_currentWeapon != selectTab->id && weaponCount < Const_NumWeapons)
               {
                  selectTab++;
                  weaponCount++;
               }

               // set the aim flags
               sprintf (aimFlags, "%s%s%s%s%s%s%s%s",
                  m_aimFlags & AimPos_NavPoint ? " NavPoint" : "",
                  m_aimFlags & AimPos_CampPoint ? " CampPoint" : "",
                  m_aimFlags & AimPos_PredictEnemy ? " PredictEnemy" : "",
                  m_aimFlags & AimPos_LastEnemy ? " LastEnemy" : "",
                  m_aimFlags & AimPos_Entity ? " Entity" : "",
                  m_aimFlags & AimPos_Enemy ? " Enemy" : "",
                  m_aimFlags & AimPos_Grenade ? " Grenade" : "",
                  m_aimFlags & AimPos_Override ? " Override" : "");

               // set the bot type
               sprintf (botType, "%s%s%s", m_personality == Personality_Rusher ? " Rusher" : "",
                  m_personality == Personality_Careful ? " Careful" : "",
                  m_personality == Personality_Normal ? " Normal" : "");

               if (weaponCount >= Const_NumWeapons)
               {
                  // prevent printing unknown message from known weapons
                  switch (m_currentWeapon)
                  {
                  case Weapon_Esplosive:
                     sprintf (weaponName, "weapon_hegrenade");
                     break;

                  case Weapon_Flashbang:
                     sprintf (weaponName, "weapon_flashbang");
                     break;

                  case Weapon_Smoke:
                     sprintf (weaponName, "weapon_smokegrenade");
                     break;

                  case Weapon_BombC4:
                     sprintf (weaponName, "weapon_c4");
                     break;

                  default:
                     sprintf (weaponName, "Unknown! (%d)", m_currentWeapon);
                  }
               }
               else
                  sprintf (weaponName, selectTab->weaponName);

               char outputBuffer[512];
               sprintf (outputBuffer, "\n\n\n\n%s (H:%.1f/A:%.1f)- Task: %d=%s Desire:%.02f\nItem: %s Clip: %d Ammo: %d%s Money: %d AimFlags: %s\nSP=%.02f SSP=%.02f I=%d PG=%d G=%d T: %.02f MT: %d\nEnemy=%s Pickup=%s Type=%s\n", STRING (pev->netname), pev->health, pev->armorvalue, taskID, szTaskname, GetCurrentTask ()->desire, &weaponName[7], GetAmmoInClip (), GetAmmo (), m_isReloading ? " (R)" : "", m_moneyAmount, aimFlags, m_moveSpeed, m_strafeSpeed, index, m_prevGoalIndex, goal, m_navTimeset - GetWorldTime (), pev->movetype, enemyName, pickupName, botType);

               MESSAGE_BEGIN (MSG_ONE_UNRELIABLE, SVC_TEMPENTITY, NULL, g_hostEntity);
                  WRITE_BYTE (TE_TEXTMESSAGE);
                  WRITE_BYTE (1);
                  WRITE_SHORT (FixedSigned16 (-1, 1 << 13));
                  WRITE_SHORT (FixedSigned16 (0, 1 << 13));
                  WRITE_BYTE (0);
                  WRITE_BYTE (GetTeam (GetEntity ()) == Team_Counter ? 0 : 255);
                  WRITE_BYTE (100);
                  WRITE_BYTE (GetTeam (GetEntity ()) != Team_Counter ? 0 : 255);
                  WRITE_BYTE (0);
                  WRITE_BYTE (255);
                  WRITE_BYTE (255);
                  WRITE_BYTE (255);
                  WRITE_BYTE (0);
                  WRITE_SHORT (FixedUnsigned16 (0, 1 << 8));
                  WRITE_SHORT (FixedUnsigned16 (0, 1 << 8));
                  WRITE_SHORT (FixedUnsigned16 (1.0, 1 << 8));
                  WRITE_STRING (const_cast <const char *> (&outputBuffer[0]));
               MESSAGE_END ();

               timeDebugUpdate = GetWorldTime () + 1.0;
            }

            // green = destination origin
            // blue = ideal angles
            // red = view angles

            DrawArrow (g_hostEntity, EyePosition (), m_destOrigin, 10, 0, 0, 255, 0, 250, 5, 1);

            MakeVectors (m_idealAngles);
            DrawArrow (g_hostEntity, EyePosition (), EyePosition () + (g_pGlobals->v_forward * 300), 10, 0, 0, 0, 255, 250, 5, 1);

            MakeVectors (pev->v_angle);
            DrawArrow (g_hostEntity, EyePosition (), EyePosition () + (g_pGlobals->v_forward * 300), 10, 0, 255, 0, 0, 250, 5, 1);

            // now draw line from source to destination
            PathNode_t *node = &m_navNode[0];

            while (node != NULL)
            {
               Vector src = g_waypoint->GetPath (node->index)->origin;
               node = node->next;

               if (node != NULL)
               {
                  Vector dest = g_waypoint->GetPath (node->index)->origin;
                  DrawArrow (g_hostEntity, src, dest, 15, 0, 255, 100, 55, 200, 5, 1);
               }
            }
         }
      }
   }

   // save the previous speed (for checking if stuck)
   m_prevSpeed = fabsf (m_moveSpeed);
   m_lastDamageType = -1; // reset damage

   pev->angles.ClampAngles ();
   pev->v_angle.ClampAngles ();
}

bool Bot::HasHostage (void)
{
   for (int i = 0; i < Const_MaxHostages; i++)
   {
      if (!FNullEnt (m_hostages[i]))
      {
         // don't care about dead hostages
         if (m_hostages[i]->v.health <= 0 || (pev->origin - m_hostages[i]->v.origin).GetLength () > 600)
         {
            m_hostages[i] = NULL;
            continue;
         }
         return true;
      }
   }
   return false;
}

void Bot::ResetCollideState (void)
{
   m_collideTime = 0.0;
   m_probeTime = 0.0;

   m_collisionProbeBits = 0;
   m_collisionState = Collision_NotDecided;
   m_collStateIndex = 0;

   for (int i = 0; i < 4; i++)
      m_collideMoves[i] = 0;
}

int Bot::GetAmmo (void)
{
   if (g_weaponDefs[m_currentWeapon].ammo1 == -1)
      return 0;

   return m_ammo[g_weaponDefs[m_currentWeapon].ammo1];
}

void Bot::TakeDamage (edict_t *inflictor, int damage, int armor, int bits)
{
   // this function gets called from the network message handler, when bot's gets hurt from any
   // other player.

   m_lastDamageType = bits;

   int team = GetTeam (GetEntity ());
   CollectGoalExperience (damage, team);

   if (IsValidPlayer (inflictor))
   {
      if (GetTeam (inflictor) == team)
      {
         // attacked by a teammate
         if (pev->health < (m_personality == Personality_Rusher ? 98 : 90))
         {
            if (m_seeEnemyTime + 1.7 < GetWorldTime ())
            {
               // alright, die you teamkiller!!!
               m_actualReactionTime = 0.0;
               m_seeEnemyTime = GetWorldTime ();
               m_enemy = inflictor;
               m_lastEnemy = m_enemy;
               m_lastEnemyOrigin = m_enemy->v.origin;
               m_enemyOrigin = m_enemy->v.origin;

               ChatMessage (Chat_TeamAttack);
            }
         }
         HandleChatterMessage ("#Bot_TeamAttack");
      }
      else
      {
         // attacked by an enemy
         if (pev->health > 70)
         {
            m_agressionLevel += 0.1;

            if (m_agressionLevel > 1.0)
               m_agressionLevel = 1.0;
         }
         else
         {
            m_fearLevel += 0.05;

            if (m_fearLevel > 1.0)
               m_fearLevel = 1.0;
         }
         // stop bot from hiding
         RemoveCertainTask (Task_Hide);

         if (FNullEnt (m_enemy))
         {
            m_lastEnemy = inflictor;
            m_lastEnemyOrigin = inflictor->v.origin;

            // FIXME - Bot doesn't necessary sees this enemy
            m_seeEnemyTime = GetWorldTime ();
         }
         CollectExperienceData (inflictor, armor + damage);
      }
   }
   else // hurt by unusual damage like drowning or gas
   {
      // leave the camping/hiding position
      if (!g_waypoint->Reachable (this, g_waypoint->FindNearest (m_destOrigin)))
      {
         DeleteSearchNodes ();
         FindWaypoint ();
      }
   }
}

void Bot::TakeBlinded (Vector fade, int alpha)
{
   // this function gets called by network message handler, when screenfade message get's send
   // it's used to make bot blind froumd the grenade.

   if (fade.x != 255 || fade.y != 255 || fade.z != 255 || alpha <= 200 || g_botVar[Variable_TurnMethod]->GetInt () == 1)
      return;

   m_enemy = NULL;

   m_maxViewDistance = g_randGen.Float (10, 20);
   m_blindTime = GetWorldTime () + static_cast <float> (alpha - 200) / 16;

   if (m_skill <= 80)
   {
      m_blindMoveSpeed = 0.0;
      m_blindSidemoveSpeed = 0.0;
      m_blindButton = IN_DUCK;
   }
   else if (m_skill < 99 || m_skill == 100)
   {
      m_blindMoveSpeed = -pev->maxspeed;
      m_blindSidemoveSpeed = 0.0;

      float walkSpeed = GetWalkSpeed ();

      if (g_randGen.Long (0, 100) > 50)
         m_blindSidemoveSpeed = walkSpeed;
      else
         m_blindSidemoveSpeed = -walkSpeed;

      if (pev->health < 85)
         m_blindMoveSpeed = -GetWalkSpeed ();
      else if (m_personality == Personality_Careful)
      {
         m_blindMoveSpeed = 0.0;
         m_blindButton = IN_DUCK;
      }
      else
         m_blindMoveSpeed = walkSpeed;
   }
}

void Bot::CollectGoalExperience (int damage, int team)
{
   // gets called each time a bot gets damaged by some enemy. tries to achieve a statistic about most/less dangerous
   // waypoints for a destination goal used for pathfinding

   if ((g_numWaypoints < 1) || g_waypointsChanged || (m_chosenGoalIndex < 0) || (m_prevGoalIndex < 0))
      return;

   // only rate goal waypoint if bot died because of the damage
   // FIXME: could be done a lot better, however this cares most about damage done by sniping or really deadly weapons
   if (pev->health - damage <= 0)
   {
      if (team == Team_Terrorist)
      {
         int value = (g_experienceData + (m_chosenGoalIndex * g_numWaypoints) + m_prevGoalIndex)->team0Value;
         value -= static_cast <int> (pev->health / 20);

         if (value < -Const_MaxGoalValue)
            value = -Const_MaxGoalValue;

         else if (value > Const_MaxGoalValue)
            value = Const_MaxGoalValue;

         (g_experienceData + (m_chosenGoalIndex * g_numWaypoints) + m_prevGoalIndex)->team0Value = static_cast <signed short> (value);
      }
      else
      {
         int value = (g_experienceData + (m_chosenGoalIndex * g_numWaypoints) + m_prevGoalIndex)->team1Value;
         value -= static_cast <int> (pev->health / 20);

         if (value < -Const_MaxGoalValue)
            value = -Const_MaxGoalValue;

         else if (value > Const_MaxGoalValue)
            value = Const_MaxGoalValue;

         (g_experienceData + (m_chosenGoalIndex * g_numWaypoints) + m_prevGoalIndex)->team1Value = static_cast <signed short> (value);
      }
   }
}

void Bot::CollectExperienceData (edict_t *attacker, int damage)
{
   // this function gets called each time a bot gets damaged by some enemy. sotores the damage (teamspecific) done by victim.

   if (!IsValidPlayer (attacker))
      return;

   int attackerTeam = GetTeam (attacker);
   int victimTeam = GetTeam (GetEntity ());

   if (attackerTeam == victimTeam)
      return;

   // if these are bots also remember damage to rank destination of the bot
   m_goalValue -= static_cast <float> (damage);

   if (g_botManager->GetBot (attacker) != NULL)
      g_botManager->GetBot (attacker)->m_goalValue += static_cast <float> (damage);

   if (damage < 20)
      return; // do not collect damage less than 20

   int attackerIndex = g_waypoint->FindNearest (attacker->v.origin);
   int victimIndex = g_waypoint->FindNearest (pev->origin);

   if (pev->health > 20)
   {
      if (victimTeam == Team_Terrorist)
         (g_experienceData + (victimIndex * g_numWaypoints) + victimIndex)->team0Damage++;
      else
         (g_experienceData + (victimIndex * g_numWaypoints) + victimIndex)->team1Damage++;

      if ((g_experienceData + (victimIndex * g_numWaypoints) + victimIndex)->team0Damage > Const_MaxDamageValue)
         (g_experienceData + (victimIndex * g_numWaypoints) + victimIndex)->team0Damage = Const_MaxDamageValue;

      if ((g_experienceData + (victimIndex * g_numWaypoints) + victimIndex)->team1Damage > Const_MaxDamageValue)
         (g_experienceData + (victimIndex * g_numWaypoints) + victimIndex)->team1Damage = Const_MaxDamageValue;
   }

   float fUpdate = IsValidBot (attacker) ? 10.0 : 7.0;

   // store away the damage done
   if (victimTeam == Team_Terrorist)
   {
      int value = (g_experienceData + (victimIndex * g_numWaypoints) + attackerIndex)->team0Damage;
      value += static_cast <int> (damage / fUpdate);

      if (value > Const_MaxDamageValue)
         value = Const_MaxDamageValue;

      (g_experienceData + (victimIndex * g_numWaypoints) + attackerIndex)->team0Damage = static_cast <unsigned short> (value);
   }
   else
   {
      int value = (g_experienceData + (victimIndex * g_numWaypoints) + attackerIndex)->team1Damage;
      value += static_cast <int> (damage / fUpdate);

      if (value > Const_MaxDamageValue)
         value = Const_MaxDamageValue;

      (g_experienceData + (victimIndex * g_numWaypoints) + attackerIndex)->team1Damage = static_cast <unsigned short> (value);
   }
}

void Bot::HandleChatterMessage (const char *tempMessage)
{
   // this function is added to prevent engine crashes with: 'Message XX started, before message XX ended', or something.

   if (FStrEq (tempMessage, "#CTs_Win") && (GetTeam (GetEntity ()) == Team_Counter))
   {
      if (g_timeRoundMid > GetWorldTime ())
         ChatterMessage (Chatter_QuicklyWonTheRound);
      else
         ChatterMessage (Chatter_WonTheRound);
   }

   if (FStrEq (tempMessage, "#Terrorists_Win") && (GetTeam (GetEntity ()) == Team_Terrorist))
   {
      if (g_timeRoundMid > GetWorldTime ())
         ChatterMessage (Chatter_QuicklyWonTheRound);
      else
         ChatterMessage (Chatter_WonTheRound);
   }

   if (FStrEq (tempMessage, "#Bot_TeamAttack"))
      ChatterMessage (Chatter_FriendlyFire);

   if (FStrEq (tempMessage, "#Bot_NiceShotCommander"))
      ChatterMessage (Chatter_NiceshotCommander);

   if (FStrEq (tempMessage, "#Bot_NiceShotPall"))
      ChatterMessage (Chatter_NiceshotPall);
}

void Bot::ChatMessage (int type, bool isTeamSay)
{
   if (g_chatFactory[type].IsEmpty () || !g_botVar[Variable_Chat]->GetBool ())
      return;

   const char *pickedPhrase = g_chatFactory[type].GetRandomElement ().GetBuffer ();

   if (IsNullString (pickedPhrase))
      return;

   PrepareChatMessage (const_cast <char *> (pickedPhrase));
   PushMessageQueue (isTeamSay ? GameMessage_TeamSay : GameMessage_SayMsg);
}

void Bot::DiscardWeaponForUser (edict_t *user, bool discardC4)
{
   // this function, asks bot to discard his current primary weapon (or c4) to the user that requsted it with /drop*
   // command, very useful, when i'm don't have money to buy anything... )

   if (IsAlive (user) && m_moneyAmount >= 2000 && HasPrimaryWeapon () && (user->v.origin - pev->origin).GetLength () <= 240)
   {
      m_aimFlags |= AimPos_Entity;
      m_lookAt = user->v.origin;

      if (discardC4)
      {
         SelectWeaponByName ("weapon_c4");
         FakeClientCommand (GetEntity (), "drop");

         SayText (FormatBuffer ("Here! %s, and now go and setup it!", STRING (user->v.netname)));
      }
      else
      {
         SelectBestWeapon ();
         FakeClientCommand (GetEntity (), "drop");

         SayText (FormatBuffer ("Here the weapon! %s, feel free to use it ;)", STRING (user->v.netname)));
      }

      m_pickupItem = NULL;
      m_pickupType = PickupType_None;
      m_itemCheckTime = GetWorldTime () + 5.0;

      if (m_inBuyZone)
      {
         m_buyingFinished = false;
         m_buyState = 0;

         PushMessageQueue (GameMessage_BuyStuff);
         m_nextBuyTime = GetWorldTime ();
      }
   }
   else
      SayText (FormatBuffer ("Sorry %s, i don't want discard my %s to you!", STRING (user->v.netname), discardC4 ? "bomb" : "weapon"));
}

void Bot::ResetDoubleJumpState (void)
{
   TaskComplete ();
   DeleteSearchNodes ();

   m_doubleJumpEntity = NULL;
   m_duckForJump = 0.0;
   m_doubleJumpOrigin = Vector::GetNull ();
   m_travelStartIndex = -1;
   m_jumpReady = false;
}

void Bot::DebugMsg (const char *format, ...)
{
   if (g_botVar[Variable_Debug]->GetInt () < 2)
      return;

   va_list ap;
   char buffer[1024];

   va_start (ap, format);
   vsprintf (buffer, format, ap);
   va_end (ap);

   ServerPrintNoTag ("%s: %s", STRING (pev->netname), buffer);

   if (g_botVar[Variable_Debug]->GetInt () >= 3)
      AddLogEntry (false, Log_Default, "%s: %s", STRING (pev->netname), buffer);
}

Vector Bot::CheckToss (const Vector &start, Vector end)
{
   // this function returns the velocity at which an object should looped from start to land near end.
   // returns null vector if toss is not feasible.

   TraceResult tr;
   float gravity = g_botVar[Variable_Gravity]->GetFloat () * 0.55;

   end = end - pev->velocity;
   end.z -= 15.0;

   if (fabsf (end.z - start.z) > 500.0)
      return Vector::GetNull ();

   Vector midPoint = start + (end - start) * 0.5;
   TraceHull (midPoint, midPoint + Vector (0, 0, 500), true, head_hull, ENT (pev), &tr);

   if (tr.flFraction < 1.0)
   {
      midPoint = tr.vecEndPos;
      midPoint.z = tr.pHit->v.absmin.z - 1.0;
   }

   if ((midPoint.z < start.z) || (midPoint.z < end.z))
      return Vector::GetNull ();

   float timeOne = sqrtf ((midPoint.z - start.z) / (0.5 * gravity));
   float timeTwo = sqrtf ((midPoint.z - end.z) / (0.5 * gravity));

   if (timeOne < 0.1)
      return Vector::GetNull ();

   Vector nadeVelocity = (end - start) / (timeOne + timeTwo);
   nadeVelocity.z = gravity * timeOne;

   Vector apex = start + nadeVelocity * timeOne;
   apex.z = midPoint.z;

   TraceHull (start, apex, false, head_hull, ENT (pev), &tr);

   if (tr.flFraction < 1.0 || tr.fAllSolid)
      return Vector::GetNull ();

   TraceHull (end, apex, true, head_hull, ENT (pev), &tr);

   if (tr.flFraction != 1.0)
   {
      float dot = -(tr.vecPlaneNormal | (apex - end).Normalize ());

      if (dot > 0.7 || tr.flFraction < 0.8) // 60 degrees
         return Vector::GetNull ();
   }
   return nadeVelocity * 0.777;
}

Vector Bot::CheckThrow (const Vector &start, Vector end)
{
   // this function returns the velocity vector at which an object should be thrown from start to hit end.
   // returns null vector if throw is not feasible.

   Vector nadeVelocity = (end - start);
   TraceResult tr;

   float gravity = g_botVar[Variable_Gravity]->GetFloat () * 0.55;
   float time = nadeVelocity.GetLength () / 195.0;

   if (time < 0.01)
      return Vector::GetNull ();
   else if (time > 2.0)
      time = 1.2;

   nadeVelocity = nadeVelocity * (1.0 / time);
   nadeVelocity.z += gravity * time * 0.5;

   Vector apex = start + (end - start) * 0.5;
   apex.z += 0.5 * gravity * (time * 0.5) * (time * 0.5);

   TraceHull (start, apex, false, head_hull, GetEntity (), &tr);

   if (tr.flFraction != 1.0)
      return Vector::GetNull ();

   TraceHull (end, apex, true, head_hull, GetEntity (), &tr);

   if (tr.flFraction != 1.0 || tr.fAllSolid)
   {
      float dot = -(tr.vecPlaneNormal | (apex - end).Normalize ());

      if (dot > 0.7 || tr.flFraction < 0.8)
         return Vector::GetNull ();
   }
   return nadeVelocity * 0.7793;
}

Vector Bot::CheckBombAudible (void)
{
   // this function checks if bomb is can be heard by the bot, calculations done by manual testing.

   if (!g_bombPlanted || (GetCurrentTask ()->taskID == Task_EscapeFromBomb))
      return Vector::GetNull (); // reliability check

   Vector bombOrigin = g_waypoint->GetBombPosition ();
   
   float timeElapsed = ((GetWorldTime () - g_timeBombPlanted) / g_botVar[Variable_C4Timer]->GetFloat ()) * 100;
   float desiredRadius = 768.0;

   // start the manual calculations
   if (timeElapsed > 85.0)
      desiredRadius = 4096.0;
   else if (timeElapsed > 68.0)
      desiredRadius = 2048.0;
   else if (timeElapsed > 52.0)
      desiredRadius = 1280.0;
   else if (timeElapsed > 28.0)
      desiredRadius = 1024.0;

   // we hear bomb if length greater than radius
   if (desiredRadius < (pev->origin - bombOrigin).GetLength2D ())
      return bombOrigin;

   return Vector::GetNull ();
}

void Bot::MoveToVector (Vector to)
{
   if (to == Vector::GetNull ())
      return;

   FindPath (m_currentWaypointIndex, g_waypoint->FindNearest (to), 2);
}

void Bot::RunPlayerMovement (void)
{
   // the purpose of this function is to compute, according to the specified computation
   // method, the msec value which will be passed as an argument of pfnRunPlayerMove. This
   // function is called every frame for every bot, since the RunPlayerMove is the function
   // that tells the engine to put the bot character model in movement. This msec value
   // tells the engine how long should the movement of the model extend inside the current
   // frame. It is very important for it to be exact, else one can experience bizarre
   // problems, such as bots getting stuck into each others. That's because the model's
   // bounding boxes, which are the boxes the engine uses to compute and detect all the
   // collisions of the model, only exist, and are only valid, while in the duration of the
   // movement. That's why if you get a pfnRunPlayerMove for one bot that lasts a little too
   // short in comparison with the frame's duration, the remaining time until the frame
   // elapses, that bot will behave like a ghost : no movement, but bullets and players can
   // pass through it. Then, when the next frame will begin, the stucking problem will arise !

   int msecMethod = g_botVar[Variable_MsecMethod]->GetInt ();

   if (msecMethod < 1 || msecMethod > 4)
      msecMethod = 1;

   // this is a pierre-marie baty's method for calculation the msec value, thanks to jussi kivilinna for additional code
   if (msecMethod == 2)
   {
      m_msecVal = static_cast <byte> (m_frameInterval * 1000.0);
      m_msecDel += m_frameInterval * 1000.0 - m_msecVal;

      // remove effect of integer conversion
      if(m_msecDel > 1.5)
      {
         m_msecVal += 1;
         m_msecDel -= 1;
      }

      if (m_msecVal < 1) // don't allow msec to be less than 1...
         m_msecVal = 1;

      if (m_msecVal > 250) // ...or greater than 250
         m_msecVal = 250;
   }

   // this is rich whitehouse's method for calculating the msec value.
   else if (msecMethod == 3)
   {
      if (m_msecDel > GetWorldTime ())
         m_msecNum += 1.0;
      else
      {
         if (m_msecNum > 0.0)
            m_msecVal = static_cast <byte> (450.0 / m_msecNum);

         m_msecDel = GetWorldTime () + 0.5; // next check in half a second
         m_msecNum = 0.0;
      }
   }

   // this is tobias heimann's method for calculating the msec value.
   else if (msecMethod == 1)
   {
      if (((m_msecDel + (m_msecNum / 1000.0)) < GetWorldTime () - 0.51) || (m_msecDel > GetWorldTime ()))
      {
         m_msecDel = GetWorldTime () - 0.05;
         m_msecNum = 0.0;
      }

      m_msecVal = static_cast <byte> ((GetWorldTime () - m_msecDel) * 1000.0 - m_msecNum);
      m_msecNum = (GetWorldTime () - m_msecDel) * 1000.0;

      if (m_msecNum > 1000.0)
      {
         m_msecDel += m_msecNum / 1000.0;
         m_msecNum = 0.0;
      }
   }

   // this is leon hartwig's method for computing the msec value
   else if (msecMethod == 4)
   {
      m_msecVal = static_cast <byte> ((GetWorldTime () - m_msecInterval) * 1000.0);

      // update msec interval for last calculation method
      m_msecInterval = GetWorldTime ();
   }

   // validate range of the msec value
   if (m_msecVal > 200)
      m_msecVal = 200;
   else if (m_msecVal < 5)
      m_msecVal = 5;

   (*g_engfuncs.pfnRunPlayerMove) (GetEntity (), m_moveAngles, m_moveSpeed, m_strafeSpeed, 0.0, pev->button, pev->impulse, m_msecVal);
}

void Bot::CheckBurstMode (float distance)
{
   // this function checks burst mode, and switch it depending distance to to enemy.

   if (HasShield ())
      return; // no checking when shiled is active

   // if current weapon is glock, disable burstmode on long distances, enable it else
   if (m_currentWeapon == Weapon_Glock18 && distance < 300.0 && m_weaponBurstMode == BurstMode_Disabled)
      pev->button |= IN_ATTACK2;
   else if (m_currentWeapon == Weapon_Glock18 && distance >= 300 && m_weaponBurstMode == BurstMode_Enabled)
      pev->button |= IN_ATTACK2;

   // if current weapon is famas, disable burstmode on short distances, enable it else
   if (m_currentWeapon == Weapon_Famas && distance > 400.0 && m_weaponBurstMode == BurstMode_Disabled)
      pev->button |= IN_ATTACK2;
   else if (m_currentWeapon == Weapon_Famas && distance <= 400 && m_weaponBurstMode == BurstMode_Enabled)
      pev->button |= IN_ATTACK2;
}

void Bot::CheckSilencer (void)
{
   if (m_currentWeapon == Weapon_Usp || m_currentWeapon == Weapon_M4A1 && !HasShield ())
   {
      int iRandomNum = (m_personality == Personality_Rusher ? 35 : 65);

      // aggressive bots don't like the silencer
      if (g_randGen.Long (1, 100) <= (m_currentWeapon == Weapon_Usp ? iRandomNum / 3 : iRandomNum))
      {
         if (pev->weaponanim > 6) // is the silencer not attached...
            pev->button |= IN_ATTACK2; // attach the silencer
      }
      else
      {
         if (pev->weaponanim <= 6) // is the silencer attached...
            pev->button |= IN_ATTACK2; // detach the silencer
      }
   }
}

float Bot::GetBombTimeleft (void)
{
   if (!g_bombPlanted)
      return 0.0;

   float timeLeft = ((g_timeBombPlanted + g_botVar[Variable_C4Timer]->GetFloat ()) - GetWorldTime ());

   if (timeLeft < 0.0)
      return 0.0;

   return timeLeft;
}

float Bot::GetEstimatedReachTime (void)
{
   float estimatedTime = 5.0; // time to reach next waypoint

   // calculate 'real' time that we need to get from one waypoint to another
   if (m_currentWaypointIndex >= 0 && m_currentWaypointIndex < g_numWaypoints && m_prevWptIndex[0] >= 0 && m_prevWptIndex[0] < g_numWaypoints)
   {
      float distance = (g_waypoint->GetPath (m_prevWptIndex[0])->origin - g_waypoint->GetPath (m_currentWaypointIndex)->origin).GetLength ();

      // caclulate estimated time
      if (pev->maxspeed <= 0.0)
         estimatedTime = 5.0 * distance / 240.0;
      else
         estimatedTime = 5.0 * distance / pev->maxspeed;

      // check for special waypoints, that can slowdown our movement
      if ((g_waypoint->GetPath (m_currentWaypointIndex)->flags & PointFlag_Crouch) || (g_waypoint->GetPath (m_currentWaypointIndex)->flags & PointFlag_Ladder) || (pev->button & IN_DUCK))
         estimatedTime *= 3.0;

      // check for too low values
      if (estimatedTime < 3.0)
         estimatedTime = 3.0;

      // check for too high values
      if (estimatedTime > 8.0)
         estimatedTime = 8.0;
   }
   return estimatedTime;
}

bool Bot::OutOfBombTimer (void)
{
   if (m_currentWaypointIndex == -1 || ((g_mapType & MapType_De) && (m_hasProgressBar || GetCurrentTask ()->taskID == Task_EscapeFromBomb)))
      return false; // if CT bot already start defusing, or already escaping, return false

   // calculate left time
   float timeLeft = GetBombTimeleft ();

   // if time left greater than 13, no need to do other checks
   if (timeLeft > 13)
      return false;

   Vector bombOrigin = g_waypoint->GetBombPosition ();

   // for terrorist, if timer is lower than eleven seconds, return true
   if (static_cast <int> (timeLeft) < 12 && GetTeam (GetEntity ()) == Team_Terrorist && (bombOrigin - pev->origin).GetLength () < 1000)
      return true;

   bool hasTeammatesWithDefuserKit = false;

   // check if our teammates has defusal kit
   for (int i = 0; i < GetMaxClients (); i++)
   {
      Bot *bot = NULL; // temporaly pointer to bot

      // search players with defuse kit
      if ((bot = g_botManager->GetBot (i)) != NULL && GetTeam (bot->GetEntity ()) == Team_Counter && bot->m_hasDefuser && (bombOrigin - bot->pev->origin).GetLength () < 500)
      {
         hasTeammatesWithDefuserKit = true;
         break;
      }
   }

   // add reach time to left time
   float reachTime = g_waypoint->GetTravelTime (pev->maxspeed, g_waypoint->GetPath (m_currentWaypointIndex)->origin, bombOrigin);

   // for counter-terrorist check alos is we have time to reach position plus average defuse time
   if ((timeLeft < reachTime + 10.0 && !m_hasDefuser && !hasTeammatesWithDefuserKit) || (timeLeft < reachTime + 5.0 && m_hasDefuser))
      return true;

   if (m_hasProgressBar && IsOnFloor () && (m_hasDefuser ? 6.0 : 11.0 > GetBombTimeleft ()))
      return true;

   return false; // return false otherwise
}

void Bot::ReactOnSound (void)
{
   int ownIndex = GetIndex ();
   float ownSoundLast = 0.0;

   if (g_clients[ownIndex].timeSoundLasting > GetWorldTime ())
   {
      if (g_clients[ownIndex].maxTimeSoundLasting <= 0.0)
         g_clients[ownIndex].maxTimeSoundLasting = 0.5;

      ownSoundLast = (g_clients[ownIndex].hearingDistance * 0.2) * (g_clients[ownIndex].timeSoundLasting - GetWorldTime ()) / g_clients[ownIndex].maxTimeSoundLasting;
   }

   edict_t *player = NULL;

   float maxVolume = 0.0, volume = 0.0;
   int hearEnemyIndex = -1, hearEnemyDistance = 0;

   // loop through all enemy clients to check for hearable stuff
   for (int i = 0; i < GetMaxClients (); i++)
   {
      if (!(g_clients[i].flags & ClientFlag_Used) || !(g_clients[i].flags & ClientFlag_Alive) || g_clients[i].ent == GetEntity () || g_clients[i].timeSoundLasting < GetWorldTime ())
         continue;

      float distance = (g_clients[i].soundPosition - pev->origin).GetLength ();
      float hearingDistance = g_clients[i].hearingDistance;

      if (distance > hearingDistance)
         continue;

      if (g_clients[i].maxTimeSoundLasting <= 0.0)
         g_clients[i].maxTimeSoundLasting = 0.5;

      if (distance <= 0.5 * hearingDistance)
         volume = hearingDistance * (g_clients[i].timeSoundLasting - GetWorldTime ()) / g_clients[i].maxTimeSoundLasting;
      else
         volume = 2.0 * hearingDistance * (1.0 - distance / hearingDistance) * (g_clients[i].timeSoundLasting - GetWorldTime ()) / g_clients[i].maxTimeSoundLasting;

      if (g_clients[hearEnemyIndex].team == GetTeam (GetEntity ()) && g_botVar[Variable_Deathmatch]->GetInt () != 2)
         volume = 0.3 * volume;

      // we will care about the most hearable sound instead of the closest one - KWo
      if (volume < maxVolume)
         continue;

      maxVolume = volume;

      if (volume < ownSoundLast)
         continue;

      hearEnemyIndex = i;
      hearEnemyDistance = distance;
   }

   if (hearEnemyIndex >= 0)
   {
      if (g_clients[hearEnemyIndex].team != GetTeam (GetEntity ()) && g_botVar[Variable_Deathmatch]->GetInt () != 2)
         player = g_clients[hearEnemyIndex].ent;
   }

   // did the bot hear someone ?
   if (!FNullEnt (player))
   {
      // change to best weapon if heard something
      if (!(m_states & State_SeeingEnemy) && IsOnFloor () && m_currentWeapon != Weapon_BombC4 && m_currentWeapon != Weapon_Esplosive && m_currentWeapon != Weapon_Smoke && m_currentWeapon != Weapon_Flashbang && !g_botVar[Variable_KnifeMode]->GetBool ())
         SelectBestWeapon ();

      m_heardSoundTime = GetWorldTime () + 5.0;
      m_states |= State_HearingEnemy;

      if ((g_randGen.Long (0, 100) < 25) && FNullEnt (m_enemy) && FNullEnt (m_lastEnemy) && m_seeEnemyTime + 7.0 < GetWorldTime ())
         ChatterMessage (Chatter_HeardEnemy);

      m_aimFlags |= AimPos_LastEnemy; 

      // didn't bot already have an enemy ? take this one...
      if (m_lastEnemyOrigin == Vector::GetNull () || m_lastEnemy == NULL)
      {
         m_lastEnemy = player;
         m_lastEnemyOrigin = player->v.origin;
      }
      else // bot had an enemy, check if it's the heard one
      {
         if (player == m_lastEnemy)
         {
            // bot sees enemy ? then bail out !
            if (m_states & State_SeeingEnemy)
               return;

            m_lastEnemyOrigin = player->v.origin;
         }
         else
         {
            // if bot had an enemy but the heard one is nearer, take it instead
            float distance = (m_lastEnemyOrigin - pev->origin).GetLength ();

            if (distance > (player->v.origin - pev->origin).GetLength () && m_seeEnemyTime + 2.0 < GetWorldTime ())
            {
               m_lastEnemy = player;
               m_lastEnemyOrigin = player->v.origin;
            }
            else
               return;
         }
      }

      // check if heard enemy can be seen
      if (CheckVisibility (VARS (player), &m_lastEnemyOrigin, &m_visibility))
      {
         m_enemy = player;
         m_lastEnemy = player;
         m_enemyOrigin = m_lastEnemyOrigin;

         m_states |= State_SeeingEnemy;
         m_seeEnemyTime = GetWorldTime ();
      }
      else if (m_lastEnemy == player && m_seeEnemyTime + 4.0 > GetWorldTime () && g_botVar[Variable_ThruWalls]->GetBool () && (g_randGen.Long (1, 100) < g_skillTab[m_skill / 20].heardShootThruProb) && IsShootableThruObstacle (player->v.origin))
      {
         m_enemy = player;
         m_lastEnemy = player;
         m_lastEnemyOrigin = player->v.origin;

         m_states |= State_SeeingEnemy;
         m_seeEnemyTime = GetWorldTime ();
      }
   }
}

bool Bot::IsShootableBreakable (edict_t *ent)
{
   // this function is checking that pointed by ent pointer obstacle, can be destroyed.

   if (FClassnameIs (ent, "func_breakable") || (FClassnameIs (ent, "func_pushable") && (ent->v.spawnflags & SF_PUSH_BREAKABLE)))
      return (ent->v.takedamage != DAMAGE_NO) && ent->v.impulse <= 0 && !(ent->v.flags & FL_WORLDBRUSH) && !(ent->v.spawnflags & SF_BREAK_TRIGGER_ONLY) && ent->v.health < 500;

   return false;
}

void Bot::EquipInBuyzone (int iBuyCount)
{
   // this function is gets called when bot enters a buyzone, to allow bot to buy some stuff

   static float lastEquipTime = 0.0;

   // if bot is in buy zone, try to buy ammo for this weapon...
   if (lastEquipTime + 15.0 < GetWorldTime () && m_inBuyZone && g_timeRoundStart + g_randGen.Float (10.0, 20.0) + g_botVar[Variable_BuyTime]->GetFloat () < GetWorldTime () && !g_bombPlanted && m_moneyAmount > g_botBuyEconomyTable[0])
   {
      m_buyingFinished = false;
      m_buyState = iBuyCount;

      // push buy message
      PushMessageQueue (GameMessage_BuyStuff);

      m_nextBuyTime = GetWorldTime ();
      lastEquipTime = GetWorldTime ();
   }
}

bool Bot::IsBombDefusing (Vector bombOrigin)
{
   // this function finds if somebody currently defusing the bomb.

   // @todo: need to check progress bar for non-bots clients.
   bool defusingInProgress = false;

   for (int i = 0; i < GetMaxClients (); i++)
   {
      Bot *bot = g_botManager->GetBot (i);

      if (bot == NULL || bot == this)
         continue; // skip invalid bots

      if (GetTeam (GetEntity ()) != GetTeam (bot->GetEntity ()) || bot->GetCurrentTask ()->taskID == Task_EscapeFromBomb)
         continue; // skip other mess

      if ((bot->pev->origin - bombOrigin).GetLength () < 80 && (bot->GetCurrentTask ()->taskID == Task_DefuseBomb || bot->m_hasProgressBar))
      {
         defusingInProgress = true;
         break;
      }

      // take in account peoples too
      if (defusingInProgress || !(g_clients[i].flags & ClientFlag_Used) || !(g_clients[i].flags & ClientFlag_Alive) || g_clients[i].team != GetTeam (GetEntity ()) || IsValidBot (g_clients[i].ent))
         continue;

      if ((g_clients[i].ent->v.origin - bombOrigin).GetLength () < 80)
      {
         defusingInProgress = true;
         break;
      }
   }
   return defusingInProgress;
}